/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/asng/ascat.c                                             *
 * Created:     2006-08-20 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2006-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ascat.h"

#include <lib/applesingle.h>
#include <lib/getopt.h>


static const char *argv0 = NULL;

static unsigned long par_cat_entry = AS_ENTRY_DATA;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 0, "data", NULL, "Extract the data fork [default]" },
	{ 'e', 1, "entry", "int", "Extract a specific entry" },
	{ 'f', 0, "finder-info", NULL, "Extract finder info" },
	{ 'F', 0, "file-info", NULL, "Extract file info" },
	{ 'i', 1, "input", "string", "Set the input file name [none]" },
	{ 'n', 0, "name", NULL, "Extract the name" },
	{ 'r', 0, "resource", NULL, "Extract the resource fork" },
	{ 't', 0, "times", NULL, "Extract file times" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"ascat: extract entries from an AppleSingle file",
		"usage: ascat [options] [file]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"ascat version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2006-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int ascat_cat_entry (applesingle_t *as, FILE *fp, unsigned long id)
{
	unsigned long i, n, k;
	unsigned char buf[4096];

	n = as_get_entry_size (as, id);
	if (n == 0) {
		return (0);
	}

	i = 0;
	while (i < n) {
		k = n - i;
		if (k > 4096) {
			k = 4096;
		}

		if (as_read (as, buf, id, i, k)) {
			fprintf (stderr, "%s: read error\n", argv0);
			return (1);
		}

		if (fwrite (buf, 1, k, fp) != k) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		i += k;
	}

	fflush (fp);

	return (0);
}

static
int ascat_cat_fp (FILE *fp, FILE *out)
{
	applesingle_t as;

	as_init (&as, fp);

	if (as_read_header (&as)) {
		fprintf (stderr, "%s: error reading applesingle header\n",
			argv0
		);
		as_free (&as);
		return (1);
	}

	if (ascat_cat_entry (&as, out, par_cat_entry)) {
		as_free (&as);
		return (1);
	}

	as_free (&as);

	return (0);
}

static
int ascat_cat (const char *fname, FILE *out)
{
	int  r;
	FILE *fp;

	fp = fopen (fname, "rb");
	if (fp == NULL) {
		fprintf (stderr, "%s: error opening file (%s)\n",
			argv0, fname
		);
		return (1);
	}

	r = ascat_cat_fp (fp, out);

	fclose (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r, ret;
	char **optarg;

	argv0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'd':
			par_cat_entry = AS_ENTRY_DATA;
			break;

		case 'e':
			par_cat_entry = strtoul (optarg[0], NULL, 0);
			break;

		case 'f':
			par_cat_entry = AS_ENTRY_FINDER_INFO;
			break;

		case 'F':
			par_cat_entry = AS_ENTRY_FILE_INFO;
			break;

		case 'i':
			r |= ascat_cat (optarg[0], stdout);
			break;

		case 'n':
			par_cat_entry = AS_ENTRY_NAME;
			break;

		case 'r':
			par_cat_entry = AS_ENTRY_RSRC;
			break;

		case 't':
			par_cat_entry = AS_ENTRY_TIME;
			break;

		case 0:
			r |= ascat_cat (optarg[0], stdout);
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (ret) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
