/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/asng/asng.c                                              *
 * Created:     2009-10-04 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2009 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "asng.h"

#include <lib/applesingle.h>
#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static int           par_encode = 0;
static int           par_decode = 0;

static int           par_adouble = 0;

static int           par_type_ok = 0;
static unsigned long par_type = MAC_FOURCC_UNKN;

static int           par_creator_ok = 0;
static unsigned long par_creator = MAC_FOURCC_UNKN;

static int           par_macname_ok = 0;
static char          *par_macname = NULL;

static int           par_finder_flags_ok = 0;
static unsigned      par_finder_flags = 0;

static int           par_file_flags_ok = 0;
static unsigned      par_file_flags = 0;

static const char    *par_ctime = NULL;
static const char    *par_mtime = NULL;

static int           par_position_x_ok = 0;
static unsigned      par_position_x = 0;

static int           par_position_y_ok = 0;
static unsigned      par_position_y = 0;

static char          *par_fn_bin = NULL;
static FILE          *par_fp_bin = NULL;

static char          *par_fn_dat = NULL;
static FILE          *par_fp_dat = NULL;

static char          *par_fn_rsc = NULL;
static FILE          *par_fp_rsc = NULL;

static char          *par_fn_inf = NULL;
static FILE          *par_fp_inf = NULL;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ '2', 0, "appledouble", NULL, "Write an AppleDouble file [no]" },
	{ 'b', 1, "applesingle", "string", "Set the AppleSingle file name" },
	{ 'c', 1, "creator", "fourcc", "Set the creator [????]" },
	{ 'C', 1, "ctime", "string", "Set the creation time [0]" },
	{ 'd', 0, "decode", NULL, "Decode an AppleSingle file" },
	{ 'e', 0, "encode", NULL, "Encode an AppleSingle file" },
	{ 'f', 1, "data", "string", "Set the data fork file name [none]" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flags [0]" },
	{ 'L', 1, "file-flags", "int", "Set the file flags [0]" },
	{ 'm', 1, "mac-name", "string", "Set the mac file name [same as data file]" },
	{ 'M', 1, "mtime", "string", "Set the modification time [0]" },
	{ 'n', 1, "finf", "string", "Set the finf file name [none]" },
	{ 'r', 1, "rsrc", "string", "Set the resource fork file name [none]" },
	{ 't', 1, "type", "fourcc", "Set the type [????]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 1, "position-x", "int", "Set the x position [0]" },
	{ 'y', 1, "position-y", "int", "Set the y position [0]" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"asng: encode and decode AppleSingle/AppleDouble files",
		"usage: asng [options] [asng [data [resource [finf]]]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"asng version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int asng_decode_as (applesingle_t *as, macfile_t *mf)
{
	unsigned long i, k, n;
	unsigned char buf[4096];

	i = 0;
	n = mf_get_data_size (mf);

	while (i < n) {
		k = n - i;
		if (k > 4096) {
			k = 4096;
		}

		if (as_read (as, buf, AS_ENTRY_DATA, i, k)) {
			return (1);
		}

		if (par_fp_dat != NULL) {
			if (fwrite (buf, 1, k, par_fp_dat) != k) {
				return (1);
			}
		}

		i += k;
	}

	i = 0;
	n = mf_get_rsrc_size (mf);

	while (i < n) {
		k = n - i;
		if (k > 4096) {
			k = 4096;
		}

		if (as_read (as, buf, AS_ENTRY_RSRC, i, k)) {
			return (1);
		}

		if (par_fp_rsc != NULL) {
			if (fwrite (buf, 1, k, par_fp_rsc) != k) {
				return (1);
			}
		}

		i += k;
	}

	if (par_fp_inf != NULL) {
		mf_save (mf, par_fp_inf);
	}

	return (0);
}

static
int asng_decode (void)
{
	int           r;
	unsigned long tm;
	applesingle_t as;
	macfile_t     mf;

	par_fp_bin = mac_open_inp (par_fn_bin, "rb", 0);
	if (par_fp_bin == NULL) {
		fprintf (stderr, "error opening AppleSingle file\n");
		return (1);
	}

	as_init (&as, par_fp_bin);

	if (as_read_header (&as)) {
		return (1);
	}

	mf_init (&mf);

	tm = uf_get_mtime (par_fn_bin);

	mf_set_name_cstr (&mf, par_fn_dat);
	mf_set_ctime (&mf, tm);
	mf_set_mtime (&mf, tm);

	if (as_is_appledouble (&as)) {
		par_adouble = 1;
	}

	as_get_macfile (&as, &mf);

	if (par_adouble == 0) {
		par_fp_dat = mac_open_out (par_fn_dat, "wb", 0);
		if ((par_fn_dat != NULL) && (par_fp_dat == NULL)) {
			fprintf (stderr, "error opening data fork file\n");
			return (1);
		}
	}

	par_fp_rsc = mac_open_out (par_fn_rsc, "wb", 0);
	if ((par_fn_rsc != NULL) && (par_fp_rsc == NULL)) {
		fprintf (stderr, "error opening resource fork file\n");
		return (1);
	}

	par_fp_inf = mac_open_out (par_fn_inf, "w", 0);
	if ((par_fn_inf != NULL) && (par_fp_inf == NULL)) {
		fprintf (stderr, "error opening finf file\n");
		return (1);
	}

	r = asng_decode_as (&as, &mf);

	if (r) {
		return (1);
	}

	mac_close (par_fp_dat);
	mac_close (par_fp_rsc);

	par_fp_dat = NULL;
	par_fp_rsc = NULL;

	tm = mf_get_mtime (&mf);

	if (par_adouble == 0) {
		uf_set_mtime (par_fn_dat, tm);
	}

	uf_set_mtime (par_fn_rsc, tm);

	return (0);
}


static
int asng_encode_as (applesingle_t *as, macfile_t *mf)
{
	unsigned long       i, k, n;
	unsigned long       datn, rscn;
	const unsigned char *name;
	unsigned char       buf[4096];

	datn = mf_get_data_size (mf);
	rscn = mf_get_rsrc_size (mf);

	if ((datn > 0) && (par_fp_dat == NULL)) {
		return (1);
	}

	if ((rscn > 0) && (par_fp_rsc == NULL)) {
		return (1);
	}

	name = mf_get_name_pstr (mf);

	as_add_entry (as, AS_ENTRY_NAME, name[0], 0);
	as_add_entry (as, AS_ENTRY_TIME, 16, 0);
	as_add_entry (as, AS_ENTRY_FINDER_INFO, 32, 0);
	as_add_entry (as, AS_ENTRY_FILE_INFO, 4, 0);

	if (datn > 0) {
		as_add_entry (as, AS_ENTRY_DATA, datn, 0);
	}

	if (rscn > 0) {
		as_add_entry (as, AS_ENTRY_RSRC, rscn, 0);
	}

	as_fix_offsets (as);

	if (as_write_header (as, par_adouble)) {
		fprintf (stderr, "error writing AppleSingle header\n");
		return (1);
	}

	if (as_write (as, name + 1, AS_ENTRY_NAME, 0, name[0])) {
		return (1);
	}

	set_uint32_be (buf, 0, as_time_from_mac (mf_get_ctime (mf)));
	set_uint32_be (buf, 4, as_time_from_mac (mf_get_mtime (mf)));
	set_uint32_be (buf, 8, 0);
	set_uint32_be (buf, 12, 0);

	if (as_write (as, buf, AS_ENTRY_TIME, 0, 16)) {
		return (1);
	}

	set_uint32_be (buf, 0, mf_get_type (mf));
	set_uint32_be (buf, 4, mf_get_creator (mf));
	set_uint16_be (buf, 8, mf_get_finder_flags (mf));
	set_uint16_be (buf, 10, mf_get_position_y (mf));
	set_uint16_be (buf, 12, mf_get_position_x (mf));
	set_uint16_be (buf, 14, mf_get_folder_id (mf));

	memset (buf + 16, 0, 16);

	if (as_write (as, buf, AS_ENTRY_FINDER_INFO, 0, 32)) {
		return (1);
	}

	set_uint32_be (buf, 0, mf_get_file_flags (mf));

	if (as_write (as, buf, AS_ENTRY_FILE_INFO, 0, 4)) {
		return (1);
	}

	i = 0;
	n = datn;

	while (i < n) {
		k = n - i;
		if (k > 4096) {
			k = 4096;
		}

		if (fread (buf, 1, k, par_fp_dat) != k) {
			return (1);
		}

		if (as_write (as, buf, AS_ENTRY_DATA, i, k)) {
			return (1);
		}

		i += k;
	}

	i = 0;
	n = rscn;

	while (i < n) {
		k = n - i;
		if (k > 4096) {
			k = 4096;
		}

		if (fread (buf, 1, k, par_fp_rsc) != k) {
			return (1);
		}

		if (as_write (as, buf, AS_ENTRY_RSRC, i, k)) {
			return (1);
		}

		i += k;
	}

	return (0);
}

static
int asng_init_macfile (macfile_t *mf, const char *fdat, unsigned long time)
{
	mf_init (mf);

	if (par_fp_inf == NULL) {
		const char    *cstr;
		unsigned char *pstr;

		if (fdat == NULL) {
			cstr = "Noname";
		}
		else {
			cstr = mac_get_last_component (fdat, '/');
		}

		pstr = mac_pstring_encode (cstr);

		if (pstr != NULL) {
			mf_set_name_pstr (mf, pstr);
			free (pstr);
		}

		mf_set_type (mf, par_type);
		mf_set_creator (mf, par_creator);
		mf_set_finder_flags (mf, par_finder_flags);
		mf_set_file_flags (mf, par_file_flags);
		mf_set_ctime (mf, time);
		mf_set_mtime (mf, time);
		mf_set_position_x (mf, par_position_x);
		mf_set_position_y (mf, par_position_y);
	}
	else {
		if (mf_load (mf, par_fp_inf)) {
			fprintf (stderr, "error loading finf file\n");
			return (1);
		}
	}

	if (par_macname_ok) {
		if (mf_set_name_cstr (mf, par_macname)) {
			fprintf (stderr, "bad mac file name\n");
			return (1);
		}
	}

	if (par_type_ok) {
		mf_set_type (mf, par_type);
	}

	if (par_creator_ok) {
		mf_set_creator (mf, par_creator);
	}

	if (par_finder_flags_ok) {
		mf_set_finder_flags (mf, par_finder_flags);
	}

	if (par_file_flags_ok) {
		mf_set_file_flags (mf, par_file_flags);
	}

	if (par_ctime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mf_set_ctime (mf, time);
		}
		else {
			mf_set_ctime (mf, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mf_set_mtime (mf, time);
		}
		else {
			mf_set_mtime (mf, mac_time_from_string (par_mtime));
		}
	}

	if (par_position_x_ok) {
		mf_set_position_x (mf, par_position_x);
	}

	if (par_position_y_ok) {
		mf_set_position_y (mf, par_position_y);
	}

	return (0);
}

static
int asng_encode (void)
{
	int           r;
	unsigned long datn, rscn;
	unsigned long rsct, datt;
	applesingle_t as;
	macfile_t     mf;

	datn = 0;
	rscn = 0;

	datt = 0;
	rsct = 0;

	par_fp_bin = mac_open_out (par_fn_bin, "wb", 0);
	if (par_fp_bin == NULL) {
		fprintf (stderr, "error opening AppleSingle file\n");
		return (1);
	}

	if (par_adouble == 0) {
		par_fp_dat = mac_open_inp (par_fn_dat, "rb", 0);
		if ((par_fn_dat != NULL) && (par_fp_dat == NULL)) {
			fprintf (stderr, "error opening data fork file\n");
			return (1);
		}
		if (uf_stat (par_fn_dat, &datn, &datt)) {
			datn = 0;
			datt = 0;
		}
	}

	par_fp_rsc = mac_open_inp (par_fn_rsc, "rb", 0);
	if ((par_fn_rsc != NULL) && (par_fp_rsc == NULL)) {
		fprintf (stderr, "error opening resource fork file\n");
		return (1);
	}
	if (uf_stat (par_fn_rsc, &rscn, &rsct)) {
		rscn = 0;
		rsct = 0;
	}

	par_fp_inf = mac_open_inp (par_fn_inf, "r", 0);
	if ((par_fn_inf != NULL) && (par_fp_inf == NULL)) {
		fprintf (stderr, "error opening finf file\n");
		return (1);
	}

	if (datt == 0) {
		datt = rsct;
	}

	asng_init_macfile (&mf, par_fn_dat, datt);

	mf_set_data_size (&mf, datn);
	mf_set_rsrc_size (&mf, rscn);

	as_init (&as, par_fp_bin);

	r = asng_encode_as (&as, &mf);

	as_free (&as);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case '2':
			par_adouble = 1;
			break;

		case 'b':
			par_fn_bin = optarg[0];
			break;

		case 'c':
			if (str_get_fourcc (optarg[0], &par_creator)) {
				fprintf (stderr,
					"%s: bad creator fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_creator_ok = 1;
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'd':
			par_encode = 0;
			par_decode = 1;
			break;

		case 'e':
			par_encode = 1;
			par_decode = 0;
			break;

		case 'f':
			par_fn_dat = optarg[0];
			break;

		case 'l':
			par_finder_flags |= mf_finder_flag_from_string (optarg[0]);
			par_finder_flags_ok = 1;
			break;

		case 'L':
			par_file_flags |= mf_file_flag_from_string (optarg[0]);
			par_file_flags_ok = 1;
			break;

		case 'm':
			par_macname = optarg[0];
			par_macname_ok = 1;
			break;

		case 'M':
			par_mtime = optarg[0];
			break;

		case 'n':
			par_fn_inf = optarg[0];
			break;

		case 'r':
			par_fn_rsc = optarg[0];
			break;

		case 't':
			if (str_get_fourcc (optarg[0], &par_type)) {
				fprintf (stderr, "%s: bad type fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_type_ok = 1;
			break;

		case 'x':
			par_position_x = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_x_ok = 1;
			break;

		case 'y':
			par_position_y = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_y_ok = 1;
			break;

		case 0:
			if (par_fn_bin == NULL) {
				par_fn_bin = optarg[0];
			}
			else if (par_fn_dat == NULL) {
				par_fn_dat = optarg[0];
			}
			else if (par_fn_rsc == NULL) {
				par_fn_rsc = optarg[0];
			}
			else if (par_fn_inf == NULL) {
				par_fn_inf = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_decode) {
		r = asng_decode();
	}
	else if (par_encode) {
		r = asng_encode();
	}
	else {
		r = 1;
	}

	mac_close (par_fp_inf);
	mac_close (par_fp_rsc);
	mac_close (par_fp_dat);
	mac_close (par_fp_bin);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
