/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/binhex/binhex.c                                          *
 * Created:     2005-02-18 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "config.h"
#include "binhex.h"

#include <lib/crc.h>
#include <lib/hqx.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static const char    *argv0 = NULL;

static unsigned      par_op = OP_NONE;

static int           par_rle = 0;

static int           par_type_ok = 0;
static unsigned long par_type = MAC_FOURCC_UNKN;

static int           par_creator_ok = 0;
static unsigned long par_creator = MAC_FOURCC_UNKN;

static int           par_macname_ok = 0;
static char          *par_macname = NULL;

static int           par_finder_flags_ok = 0;
static unsigned      par_finder_flags = 0;

static char          *par_fn_hqx = NULL;
static FILE          *par_fp_hqx = NULL;

static char          *par_fn_dat = NULL;
static FILE          *par_fp_dat = NULL;

static char          *par_fn_rsc = NULL;
static FILE          *par_fp_rsc = NULL;

static char          *par_fn_inf = NULL;
static FILE          *par_fp_inf = NULL;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'b', 1, "binhex", "string", "Set the hqx file name" },
	{ 'c', 1, "creator", "fourcc", "Set the creator [????]" },
	{ 'd', 0, "decode", NULL, "Decode a binhexary file" },
	{ 'e', 0, "encode", NULL, "Encode a binhexary file" },
	{ 'f', 1, "data", "string", "Set the data fork file name" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flags" },
	{ 'm', 1, "mac-name", "string", "Set the mac file name [same as data file]" },
	{ 'r', 1, "rsrc", "string", "Set the resource fork file name" },
	{ 'R', 0, "rle", NULL, "Use RLE encoding [no]" },
	{ 't', 1, "type", "fourcc", "Set the type [????]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"binhex: encode and decode BinHex files",
		"usage: binhex [options] [hqx [data [resource [finf]]]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"binhex version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2005-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int binhex_decode_hqx (hqx_t *hqx, unsigned long mtime)
{
	unsigned long i, n;
	unsigned long datn, rscn;
	crc16_t       crc;
	unsigned char buf[4096];

	datn = hqx_get_data_size (hqx);
	rscn = hqx_get_rsrc_size (hqx);

	crc16_init (&crc, CRC_CRC16_CCITT, 0);

	n = datn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (hqx_read (hqx, buf, i)) {
			return (1);
		}

		crc16_calc (&crc, buf, i);

		if (par_fp_dat != NULL) {
			if (fwrite (buf, 1, i, par_fp_dat) != i) {
				return (1);
			}
		}

		n -= i;
	}

	crc16_done (&crc);

	if (hqx_read (hqx, buf, 2)) {
		return (1);
	}

	if (crc16_get_crc (&crc) != hqx_get_uint16 (buf, 0)) {
		fprintf (stderr, "%s: data fork crc error\n", argv0);
		return (1);
	}

	crc16_init (&crc, CRC_CRC16_CCITT, 0);

	n = rscn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (hqx_read (hqx, buf, i)) {
			return (1);
		}

		crc16_calc (&crc, buf, i);

		if (par_fp_rsc != NULL) {
			if (fwrite (buf, 1, i, par_fp_rsc) != i) {
				return (1);
			}
		}

		n -= i;
	}

	crc16_done (&crc);

	if (hqx_read (hqx, buf, 2)) {
		return (1);
	}

	if (crc16_get_crc (&crc) != hqx_get_uint16 (buf, 0)) {
		fprintf (stderr, "%s: resource fork crc error\n", argv0);
		return (1);
	}

	if (par_fp_inf != NULL) {
		macfile_t mf;

		mf_init (&mf);
		hqx_get_macfile (hqx, &mf);

		mf_set_ctime (&mf, mtime);
		mf_set_mtime (&mf, mtime);

		mf_save (&mf, par_fp_inf);
	}

	return (0);
}

static
int binhex_decode (void)
{
	unsigned long mtime;
	hqx_t         hqx;

	par_fp_hqx = mac_open_inp (par_fn_hqx, "r", 0);
	if (par_fp_hqx == NULL) {
		fprintf (stderr, "%s: error opening hqx file\n", argv0);
		return (1);
	}

	mtime = uf_get_mtime_fp (par_fp_hqx);

	if (hqx_init_read (&hqx, par_fp_hqx)) {
		return (1);
	}

	if (hqx_read_header (&hqx)) {
		fprintf (stderr, "%s: error reading hqx header\n", argv0);
		return (1);
	}

	if (hqx_get_crc (&hqx) != hqx_calc_crc (&hqx)) {
		fprintf (stderr, "%s: header crc error\n", argv0);
		return (1);
	}

	par_fp_dat = mac_open_out (par_fn_dat, "wb", 0);
	if ((par_fn_dat != NULL) && (par_fp_dat == NULL)) {
		fprintf (stderr, "error opening data fork file\n");
		return (1);
	}

	par_fp_rsc = mac_open_out (par_fn_rsc, "wb", 0);
	if ((par_fn_rsc != NULL) && (par_fp_rsc == NULL)) {
		fprintf (stderr, "error opening resource fork file\n");
		return (1);
	}

	par_fp_inf = mac_open_out (par_fn_inf, "w", 0);
	if ((par_fn_inf != NULL) && (par_fp_inf == NULL)) {
		fprintf (stderr, "error opening finf file\n");
		return (1);
	}

	if (binhex_decode_hqx (&hqx, mtime)) {
		return (1);
	}

	return (0);
}


static
int binhex_encode_hqx (hqx_t *hqx)
{
	unsigned long i, n;
	unsigned long datn, rscn;
	crc16_t       crc;
	unsigned char buf[4096];

	datn = hqx_get_data_size (hqx);
	rscn = hqx_get_rsrc_size (hqx);

	if ((datn > 0) && (par_fp_dat == NULL)) {
		return (1);
	}

	if ((rscn > 0) && (par_fp_rsc == NULL)) {
		return (1);
	}

	hqx_set_crc (hqx, hqx_calc_crc (hqx));

	hqx_write_header (hqx);

	crc16_init (&crc, CRC_CRC16_CCITT, 0);

	n = datn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (fread (buf, 1, i, par_fp_dat) != i) {
			return (1);
		}

		crc16_calc (&crc, buf, i);

		hqx_write (hqx, buf, i);

		n -= i;
	}

	crc16_done (&crc);

	hqx_set_uint16 (buf, 0, crc16_get_crc (&crc));
	hqx_write (hqx, buf, 2);

	crc16_init (&crc, CRC_CRC16_CCITT, 0);

	n = rscn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (fread (buf, 1, i, par_fp_rsc) != i) {
			return (1);
		}

		crc16_calc (&crc, buf, i);

		hqx_write (hqx, buf, i);

		n -= i;
	}

	crc16_done (&crc);

	hqx_set_uint16 (buf, 0, crc16_get_crc (&crc));
	hqx_write (hqx, buf, 2);

	hqx_flush (hqx);

	return (0);
}

static
int binhex_encode (void)
{
	int           r;
	unsigned long datn, rscn;
	unsigned long datt, rsct;
	hqx_t         hqx;

	datn = 0;
	rscn = 0;

	par_fp_hqx = mac_open_out (par_fn_hqx, "w", 0);
	if (par_fp_hqx == NULL) {
		fprintf (stderr, "error opening binhexary file\n");
		return (1);
	}

	par_fp_dat = mac_open_inp (par_fn_dat, "rb", 0);
	if ((par_fn_dat != NULL) && (par_fp_dat == NULL)) {
		fprintf (stderr, "error opening data fork file\n");
		return (1);
	}
	if (uf_stat (par_fn_dat, &datn, &datt)) {
		datn = 0;
		datt = 0;
	}

	par_fp_rsc = mac_open_inp (par_fn_rsc, "rb", 0);
	if ((par_fn_rsc != NULL) && (par_fp_rsc == NULL)) {
		fprintf (stderr, "error opening resource fork file\n");
		return (1);
	}
	if (uf_stat (par_fn_rsc, &rscn, &rsct)) {
		rscn = 0;
		rsct = 0;
	}

	par_fp_inf = mac_open_inp (par_fn_inf, "r", 0);
	if ((par_fn_inf != NULL) && (par_fp_inf == NULL)) {
		fprintf (stderr, "error opening finf file\n");
		return (1);
	}

	hqx_init_write (&hqx, par_fp_hqx, par_rle);

	if (par_fp_inf == NULL) {
		const char    *cstr;
		unsigned char *pstr;

		if (par_fn_dat == NULL) {
			cstr = "Noname";
		}
		else {
			cstr = mac_get_last_component (par_fn_dat, '/');
		}

		pstr = mac_pstring_encode (cstr);

		if (pstr != NULL) {
			hqx_set_name_pstr (&hqx, pstr);
			free (pstr);
		}

		hqx_set_type (&hqx, par_type);
		hqx_set_creator (&hqx, par_creator);
		hqx_set_finder_flags (&hqx, par_finder_flags);
	}
	else {
		macfile_t mf;

		if (mf_load (&mf, par_fp_inf)) {
			fprintf (stderr, "error loading finf file\n");
			return (1);
		}

		hqx_set_macfile (&hqx, &mf);
	}

	if (par_macname_ok) {
		if (hqx_set_name_cstr (&hqx, par_macname)) {
			fprintf (stderr, "bad mac file name\n");
			return (1);
		}
	}

	if (par_type_ok) {
		hqx_set_type (&hqx, par_type);
	}

	if (par_creator_ok) {
		hqx_set_creator (&hqx, par_creator);
	}

	if (par_finder_flags_ok) {
		hqx_set_finder_flags (&hqx, par_finder_flags);
	}

	hqx_set_version (&hqx, 0);

	hqx_set_data_size (&hqx, datn);
	hqx_set_rsrc_size (&hqx, rscn);
	hqx_set_crc (&hqx, hqx_calc_crc (&hqx));

	r = binhex_encode_hqx (&hqx);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;

	argv0 = argv[0];

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'b':
			par_fn_hqx = optarg[0];
			break;

		case 'c':
			if (str_get_fourcc (optarg[0], &par_creator)) {
				fprintf (stderr,
					"%s: bad creator fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_creator_ok = 1;
			break;

		case 'd':
			par_op = OP_DECODE;
			break;

		case 'e':
			par_op = OP_ENCODE;
			break;

		case 'f':
			par_fn_dat = optarg[0];
			break;

		case 'l':
			par_finder_flags = (unsigned) strtoul (optarg[0], NULL, 0);
			par_finder_flags_ok = 1;
			break;

		case 'm':
			par_macname = optarg[0];
			par_macname_ok = 1;
			break;

		case 'n':
			par_fn_inf = optarg[0];
			break;

		case 'r':
			par_fn_rsc = optarg[0];
			break;

		case 'R':
			par_rle = 1;
			break;

		case 't':
			if (str_get_fourcc (optarg[0], &par_type)) {
				fprintf (stderr, "%s: bad type fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_type_ok = 1;
			break;

		case 0:
			if (par_fn_hqx == NULL) {
				par_fn_hqx = optarg[0];
			}
			else if (par_fn_dat == NULL) {
				par_fn_dat = optarg[0];
			}
			else if (par_fn_rsc == NULL) {
				par_fn_rsc = optarg[0];
			}
			else if (par_fn_inf == NULL) {
				par_fn_inf = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_op == OP_NONE) {
		par_op = OP_DECODE;
	}

	if (par_op == OP_DECODE) {
		r = binhex_decode();
	}
	else if (par_op == OP_ENCODE) {
		r = binhex_encode();
	}
	else {
		r = 1;
	}

	mac_close (par_fp_inf);
	mac_close (par_fp_rsc);
	mac_close (par_fp_dat);
	mac_close (par_fp_hqx);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
