/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/dc42/dc42.c                                              *
 * Created:     2009-10-11 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2009-2013 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "dc42.h"

#include <lib/dc42.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


#define BUF_MAX 16384


enum {
	OP_DECODE,
	OP_ENCODE,
	OP_CHECK,
	OP_INFO
};


static const char    *argv0 = NULL;

static char          par_verbose = 0;
static unsigned      par_op = OP_INFO;
static char          par_first = 1;
static const char    *par_name = NULL;
static unsigned char par_buf[BUF_MAX];


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 0, "check", NULL, "Check an image file" },
	{ 'd', 0, "decode", NULL, "Decode an image file" },
	{ 'e', 0, "encode", NULL, "Encode an image file" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'I', 0, "info", NULL, "Print info about the disk image" },
	{ 'n', 1, "name", "string", "Set the image name for encoding [file name]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 't', 1, "tags", "string", "Set the tag file name [none]" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"dc42: encode and decode Disk Copy 4.2 image files",
		"usage: dc42 [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"dc42 version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2009-2012 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
unsigned long dc42_check_checksum (FILE *inp, unsigned long cnt)
{
	unsigned      n;
	unsigned long chk;

	chk = 0;

	while (cnt > 0) {
		n = (cnt < BUF_MAX) ? cnt : BUF_MAX;

		if (n > 1) {
			n &= ~1U;
		}

		if (fread (par_buf, 1, n, inp) != n) {
			break;
		}

		chk = dc42_calc_checksum (par_buf, n, chk);
		cnt -= n;
	}

	return (chk);
}

static
int dc42_check_fp (FILE *inp, FILE *out, const char *fname)
{
	int           r;
	unsigned char hdr[128];
	unsigned long chk1, chk2, cnt1, cnt2;

	r = 0;

	if (dc42_read (hdr, inp)) {
		fprintf (out, "%s: error reading header\n", fname);
		return (1);
	}

	chk1 = dc42_get_data_checksum (hdr);
	chk2 = dc42_check_checksum (inp, dc42_get_data_size (hdr));

	if (chk1 != chk2) {
		fprintf (out, "%s: data checksum error (%08lx / %08lx)\n",
			fname, chk1, chk2
		);
		r = 1;
	}

	cnt1 = dc42_get_tag_size (hdr);
	cnt2 = (cnt1 < 12) ? cnt1 : 12;

	chk1 = dc42_get_tag_checksum (hdr);
	chk2 = dc42_check_checksum (inp, cnt2);
	chk2 = dc42_check_checksum (inp, cnt1 - cnt2);

	if (chk1 != chk2) {
		fprintf (out, "%s: tag checksum error (%08lx / %08lx)\n",
			fname, chk1, chk2
		);
		r = 1;
	}

	if ((r == 0) && par_verbose) {
		fprintf (out, "%s: ok\n", fname);
	}

	return (r);
}

static
int dc42_check (const char *inp, const char *out)
{
	int  r;
	FILE *finp, *fout;

	finp = mac_open_inp (inp, "rb", 0);

	if (finp == NULL) {
		return (1);
	}

	fout = mac_open_out (out, "w", 1);

	if (fout == NULL) {
		mac_close (finp);
		return (1);
	}

	r = dc42_check_fp (finp, fout, inp);

	mac_close (fout);
	mac_close (finp);

	return (r);
}


static
int dc42_info_fp (FILE *inp, FILE *out, const char *fname)
{
	unsigned char hdr[128];
	char          *name;

	if (par_first) {
		par_first = 0;
	}
	else {
		fputs ("\n", out);
	}

	if (dc42_read (hdr, inp)) {
		fprintf (stderr, "%s: read error (header)\n", argv0);
		return (1);
	}

	name = dc42_get_name_cstr (hdr);

	if (name == NULL) {
		name = "<none>";
	}

	fprintf (out, "file name:     \"%s\"\n", fname);
	fprintf (out, "image name:    \"%s\"\n", name);
	fprintf (out, "data size:     %lu\n", dc42_get_data_size (hdr));
	fprintf (out, "tag size:      %lu\n", dc42_get_tag_size (hdr));
	fprintf (out, "data checksum: 0x%08lX\n", dc42_get_data_checksum (hdr));
	fprintf (out, "tag checksum:  0x%08lX\n", dc42_get_tag_checksum (hdr));

	free (name);

	return (0);
}

static
int dc42_info (const char *inp, const char *out)
{
	int  r;
	FILE *finp, *fout;

	finp = mac_open_inp (inp, "rb", 0);

	if (finp == NULL) {
		return (1);
	}

	fout = mac_open_out (out, "w", 1);

	if (fout == NULL) {
		mac_close (finp);
		return (1);
	}

	r = dc42_info_fp (finp, fout, inp);

	mac_close (fout);
	mac_close (finp);

	return (r);
}


static
int dc42_copy (FILE *inp, FILE *out, unsigned long cnt, unsigned long *chk)
{
	unsigned      n;
	unsigned long ck2;

	ck2 = 0;

	while (cnt > 0) {
		n = (cnt < BUF_MAX) ? cnt : BUF_MAX;

		if (n > 1) {
			n &= ~1U;
		}

		if (fread (par_buf, 1, n, inp) != n) {
			fprintf (stderr, "%s: read error\n", argv0);
			return (1);
		}

		ck2 = dc42_calc_checksum (par_buf, n, ck2);

		if (out != NULL) {
			if (fwrite (par_buf, 1, n, out) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		cnt -= n;
	}

	if (chk != NULL) {
		*chk = ck2;
	}

	return (0);
}

static
int dc42_decode_fp (FILE *inp, FILE *out, FILE *tag)
{
	int           r;
	unsigned long cnt1, cnt2, chk1, chk2;
	unsigned char hdr[128];

	if (dc42_read (hdr, inp)) {
		fprintf (stderr, "%s: read error (header)\n", argv0);
		return (1);
	}

	r = 0;

	cnt1 = dc42_get_data_size (hdr);
	chk1 = dc42_get_data_checksum (hdr);

	if (dc42_copy (inp, out, cnt1, &chk2)) {
		return (1);
	}

	if (chk1 != chk2) {
		fprintf (stderr, "%s: bad data checksum (%08lx / %08lx)\n",
			argv0, chk1, chk2
		);
		r = 1;
	}

	cnt1 = dc42_get_tag_size (hdr);
	cnt2 = (cnt1 < 12) ? cnt1 : 12;
	chk1 = dc42_get_tag_checksum (hdr);

	if (dc42_copy (inp, tag, cnt2, &chk2)) {
		return (1);
	}

	if (dc42_copy (inp, tag, cnt1 - cnt2, &chk2)) {
		return (1);
	}

	if (chk1 != chk2) {
		fprintf (stderr, "%s: bad tag checksum (%08lx / %08lx)\n",
			argv0, chk1, chk2
		);
		r = 1;
	}

	return (r);
}

static
int dc42_decode (const char *inp, const char *out, const char *tag)
{
	int  r;
	FILE *fp1, *fp2, *fp3;

	fp2 = NULL;
	fp3 = NULL;

	fp1 = mac_open_inp (inp, "rb", 1);

	if (fp1 == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			argv0, inp
		);
		return (1);
	}

	if (out != NULL) {
		fp2 = mac_open_out (out, "wb", 0);

		if (fp2 == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n",
				argv0, out
			);
			mac_close (fp1);
			return (1);
		}
	}

	if (tag != NULL) {
		fp3 = mac_open_out (tag, "wb", 0);

		if (fp3 == NULL) {
			fprintf (stderr, "%s: can't open tag file (%s)\n",
				argv0, out
			);
			mac_close (fp1);
			return (1);
		}
	}

	r = dc42_decode_fp (fp1, fp2, fp3);

	mac_close (fp3);
	mac_close (fp2);
	mac_close (fp1);

	return (r);
}

/*
 * Extract the volume name from the volume information.
 */
static
int dc42_get_name (FILE *inp, unsigned char *pstr, unsigned max)
{
	unsigned long       pos;
	const unsigned char *src;
	unsigned char       buf[256];

	mac_get_pos (inp, &pos);

	if (mac_read (inp, buf, 1024, 256)) {
		mac_set_pos (inp, pos);
		return (1);
	}

	mac_set_pos (inp, pos);

	if ((buf[0] == 'B') && (buf[1] == 'D')) {
		src = buf + 36;
	}
	else if ((buf[0] == 0xd2) && (buf[1] == 0xd7)) {
		src = buf + 36;
	}
	else {
		return (1);
	}

	if ((src[0] >= 128) || (src[0] >= max)) {
		return (1);
	}

	memcpy (pstr, src, src[0] + 1);

	return (0);
}

static
int dc42_encode_fp (FILE *inp, FILE *out, FILE *tag, const char *name, unsigned long dsize, unsigned long tsize)
{
	unsigned char hdr[128];
	unsigned char pname[64];
	unsigned long cnt, chk;

	dc42_init (hdr);

	if (mac_write (out, hdr, 0, 84)) {
		fprintf (stderr, "%s: write error (header)\n", argv0);
		return (1);
	}

	if (par_name != NULL) {
		dc42_set_name_cstr (hdr, par_name);
	}
	else if (dc42_get_name (inp, pname, 64) == 0) {
		dc42_set_name_pstr (hdr, pname);
	}
	else {
		dc42_set_name_cstr (hdr, "-not a Macintosh disk-");
	}

	dc42_set_data_size (hdr, dsize);
	dc42_set_tag_size (hdr, tsize);

	if (dc42_copy (inp, out, dsize, &chk)) {
		return (1);
	}

	dc42_set_data_checksum (hdr, chk);

	cnt = (tsize < 12) ? tsize : 12;

	if (dc42_copy (tag, out, cnt, &chk)) {
		return (1);
	}

	if (dc42_copy (tag, out, tsize - cnt, &chk)) {
		return (1);
	}

	dc42_set_tag_checksum (hdr, chk);

	if (mac_write (out, hdr, 0, 84)) {
		fprintf (stderr, "%s: write error (header)\n", argv0);
		return (1);
	}

	return (0);
}

static
int dc42_encode (const char *inp, const char *out, const char *tag)
{
	int           r;
	unsigned long dsize, mtime;
	unsigned long tsize;
	FILE          *finp, *fout, *ftag;

	if (uf_stat (tag, &tsize, &mtime)) {
		tsize = 0;
	}

	if (uf_stat (inp, &dsize, &mtime)) {
		dsize = 0;
	}

	finp = NULL;
	ftag = NULL;

	if (inp != NULL) {
		finp = mac_open_inp (inp, "rb", 0);

		if (finp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n",
				argv0, inp
			);
			return (1);
		}
	}

	if (tag != NULL) {
		ftag = mac_open_inp (tag, "rb", 0);

		if (ftag == NULL) {
			fprintf (stderr, "%s: can't open tag file (%s)\n",
				argv0, tag
			);
			mac_close (finp);
			return (1);
		}
	}

	fout = mac_open_out (out, "wb", 1);

	if (fout == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			argv0, out
		);
		mac_close (ftag);
		mac_close (finp);
		return (1);
	}

	r = dc42_encode_fp (finp, fout, ftag, inp, dsize, tsize);

	mac_close (ftag);
	mac_close (fout);
	mac_close (finp);

	return (r);
}

int main (int argc, char *argv[])
{
	int        r, ret;
	char       **optarg;
	const char *finp, *fout, *ftag;

	argv0 = argv[0];

	ret = 0;

	finp = NULL;
	fout = NULL;
	ftag = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'c':
			par_op = OP_CHECK;
			break;

		case 'd':
			par_op = OP_DECODE;
			break;

		case 'e':
			par_op = OP_ENCODE;
			break;

		case 'i':
			if (par_op == OP_CHECK) {
				ret |= dc42_check (optarg[0], fout);
			}
			else if (par_op == OP_INFO) {
				ret |= dc42_info (optarg[0], fout);
			}
			else {
				finp = optarg[0];
			}
			break;

		case 'I':
			par_op = OP_INFO;
			break;

		case 'n':
			par_name = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 't':
			ftag = optarg[0];
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (par_op == OP_CHECK) {
				ret |= dc42_check (optarg[0], fout);
			}
			else if (par_op == OP_INFO) {
				ret |= dc42_info (optarg[0], fout);
			}
			else {
				if (finp == NULL) {
					finp = optarg[0];
				}
				else if (fout == NULL) {
					fout = optarg[0];
				}
				else if (ftag == NULL) {
					ftag = optarg[0];
				}
				else {
					fprintf (stderr,
						"%s: too many files (%s)\n",
						argv0, optarg[0]
					);
					return (EXIT_FAILURE);
				}
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_op == OP_DECODE) {
		ret |= dc42_decode (finp, fout, ftag);
	}
	else if (par_op == OP_ENCODE) {
		ret |= dc42_encode (finp, fout, ftag);
	}
	else if (par_op == OP_CHECK) {
		;
	}
	else if (par_op == OP_INFO) {
		;
	}
	else {
		ret = 1;
	}

	if (ret) {
		return (EXIT_FAILURE);
	}

	return (0);
}
