/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/applesingle.c                                        *
 * Created:     2006-08-18 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2006-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/applesingle.h>
#include <lib/buffer.h>
#include <lib/macfile.h>

#define AS_MAGIC_AS 0x00051600
#define AS_MAGIC_AD 0x00051607


static
int asread (FILE *fp, void *buf, unsigned long ofs, unsigned long cnt)
{
	if (fseek (fp, ofs, SEEK_SET)) {
		return (1);
	}

	if (fread (buf, cnt, 1, fp) != 1) {
		return (1);
	}

	return (0);
}

static
int aswrite (FILE *fp, const void *buf, unsigned long ofs, unsigned long cnt)
{
	if (fseek (fp, ofs, SEEK_SET)) {
		return (1);
	}

	if (fwrite (buf, cnt, 1, fp) != 1) {
		return (1);
	}

	return (0);
}

void as_init (applesingle_t *as, FILE *fp)
{
	as->fp = fp;

	as->magic = 0;
	as->version = 0;

	as->entcnt = 0;
	as->entry = NULL;
}

void as_free (applesingle_t *as)
{
	if (as != NULL) {
		free (as->entry);
	}
}

applesingle_t *as_new (FILE *fp)
{
	applesingle_t *as;

	as = malloc (sizeof (applesingle_t));
	if (as == NULL) {
		return (NULL);
	}

	as_init (as, fp);

	return (as);
}

void as_del (applesingle_t *as)
{
	if (as != NULL) {
		as_free (as);
		free (as);
	}
}

int as_add_entry (applesingle_t *as, unsigned long id, unsigned long size, unsigned long ofs)
{
	unsigned i;
	as_entry_t *ent;

	for (i = 0; i < as->entcnt; i++) {
		if (as->entry[i].id == id) {
			return (1);
		}
	}

	ent = realloc (as->entry, (as->entcnt + 1) * sizeof (as_entry_t));
	if (ent == NULL) {
		return (1);
	}

	as->entry = ent;
	as->entcnt += 1;

	ent = as->entry + as->entcnt - 1;

	ent->id = id;
	ent->ofs = ofs;
	ent->size = size;

	return (0);
}

void as_fix_offsets (applesingle_t *as)
{
	unsigned      i;
	unsigned long ofs;

	ofs = 32;
	ofs = 26 + 12 * as->entcnt;

	for (i = 0; i < as->entcnt; i++) {
		as->entry[i].ofs = ofs;
		ofs += as->entry[i].size;
	}
}

static
as_entry_t *as_get_entry (applesingle_t *as, unsigned long id)
{
	unsigned i;

	for (i = 0; i < as->entcnt; i++) {
		if (as->entry[i].id == id) {
			return (&as->entry[i]);
		}
	}

	return (NULL);
}

unsigned as_get_entry_count (applesingle_t *as)
{
	return (as->entcnt);
}

unsigned long as_get_index_id (applesingle_t *as, unsigned idx)
{
	if (idx < as->entcnt) {
		return (as->entry[idx].id);
	}

	return (0);
}

unsigned long as_get_index_size (applesingle_t *as, unsigned idx)
{
	if (idx < as->entcnt) {
		return (as->entry[idx].size);
	}

	return (0);
}

unsigned long as_get_index_offset (applesingle_t *as, unsigned idx)
{
	if (idx < as->entcnt) {
		return (as->entry[idx].ofs);
	}

	return (0);
}

unsigned long as_get_entry_size (applesingle_t *as, unsigned long id)
{
	as_entry_t *ent;

	ent = as_get_entry (as, id);
	if (ent == NULL) {
		return (0);
	}

	return (ent->size);
}

unsigned long as_get_entry_offset (applesingle_t *as, unsigned long id)
{
	as_entry_t *ent;

	ent = as_get_entry (as, id);
	if (ent == NULL) {
		return (0);
	}

	return (ent->ofs);
}

unsigned long as_get_magic (applesingle_t *as)
{
	return (as->magic);
}

unsigned long as_get_version (applesingle_t *as)
{
	return (as->version);
}

int as_is_applesingle (const applesingle_t *as)
{
	return (as->magic == AS_MAGIC_AS);
}

int as_is_appledouble (const applesingle_t *as)
{
	return (as->magic == AS_MAGIC_AD);
}

int as_read_header (applesingle_t *as)
{
	unsigned      i, n;
	unsigned long id, ofs, size;
	unsigned char buf[32];

	if (as->entcnt > 0) {
		free (as->entry);
		as->entcnt = 0;
		as->entry = NULL;
	}

	if (asread (as->fp, buf, 0, 26)) {
		return (1);
	}

	as->magic = get_uint32_be (buf, 0);
	if ((as->magic != AS_MAGIC_AS) && (as->magic != AS_MAGIC_AD)) {
		return (1);
	}

	as->version = get_uint32_be (buf, 4);

	if (as->version == 0x00010000) {
		;
	}
	else if (as->version == 0x00020000) {
		;
	}
	else {
		return (1);
	}

	n = get_uint16_be (buf, 24);

	for (i = 0; i < n; i++) {
		if (asread (as->fp, buf, 26 + 12 * i, 12)) {
			return (1);
		}

		id = get_uint32_be (buf, 0);
		ofs = get_uint32_be (buf, 4);
		size = get_uint32_be (buf, 8);

		if (as_add_entry (as, id, size, ofs)) {
			return (1);
		}
	}

	return (0);
}

int as_write_header (applesingle_t *as, int ad)
{
	unsigned      i;
	unsigned char buf[32];

	for (i = 0; i < 26; i++) {
		buf[i] = 0;
	}

	if (ad) {
		set_uint32_be (buf, 0, AS_MAGIC_AD);
	}
	else {
		set_uint32_be (buf, 0, AS_MAGIC_AS);
	}

	set_uint32_be (buf, 4, 0x00020000);
	set_uint16_be (buf, 24, as->entcnt);

	if (aswrite (as->fp, buf, 0, 26)) {
		return (1);
	}

	for (i = 0; i < as->entcnt; i++) {
		set_uint32_be (buf, 0, as->entry[i].id);
		set_uint32_be (buf, 4, as->entry[i].ofs);
		set_uint32_be (buf, 8, as->entry[i].size);

		if (aswrite (as->fp, buf, 26 + 12 * i, 12)) {
			return (1);
		}
	}

	return (0);
}

int as_read (applesingle_t *as, void *buf, unsigned long id, unsigned long ofs, unsigned long cnt)
{
	as_entry_t *ent;

	ent = as_get_entry (as, id);
	if (ent == NULL) {
		return (1);
	}

	if ((ofs + cnt) > ent->size) {
		return (1);
	}

	if (asread (as->fp, buf, ent->ofs + ofs, cnt)) {
		return (1);
	}

	return (0);
}

int as_write (applesingle_t *as, const void *buf, unsigned long id, unsigned long ofs, unsigned long cnt)
{
	as_entry_t *ent;

	ent = as_get_entry (as, id);
	if (ent == NULL) {
		return (1);
	}

	if ((ofs + cnt) > ent->size) {
		return (1);
	}

	if (aswrite (as->fp, buf, ent->ofs + ofs, cnt)) {
		return (1);
	}

	return (0);
}

unsigned long as_time_to_mac (unsigned long val)
{
	if (val == 0x80000000) {
		return (0);
	}

	if (val & 0x80000000) {
		val = (~val + 1) & 0xffffffff;
		val = 3029529600UL - val;
	}
	else {
		val = 3029529600UL + val;
	}

	/* time zone! */

	return (val);
}

unsigned long as_time_from_mac (unsigned long val)
{
	if (val >= 3029529600UL) {
		val = val - 3029529600UL;
	}
	else {
		val = 3029529600UL - val;

		if (val & 0x80000000) {
			val = 0x80000000;
		}
		else {
			val = (~val + 1) & 0xffffffff;
		}

	}

	/* time zone! */

	return (val);
}

void as_get_macfile (applesingle_t *as, macfile_t *mf)
{
	unsigned long n;
	unsigned char buf[256];

	mf_set_data_size (mf, as_get_entry_size (as, AS_ENTRY_DATA));
	mf_set_rsrc_size (mf, as_get_entry_size (as, AS_ENTRY_RSRC));

	n = as_get_entry_size (as, AS_ENTRY_NAME);
	if (n > 0) {
		if (n > 255) {
			n = 255;
		}

		if (as_read (as, buf + 1, AS_ENTRY_NAME, 0, n) == 0) {
			buf[0] = n;
			mf_set_name_pstr (mf, buf);
		}
	}

	n = as_get_entry_size (as, AS_ENTRY_FINDER_INFO);
	if (n > 0) {
		n = (n < 32) ? n : 32;

		if (as_read (as, buf, AS_ENTRY_FINDER_INFO, 0, n) == 0) {
			if (n >= 4) {
				mf_set_type (mf, get_uint32_be (buf, 0));
			}
			if (n >= 8) {
				mf_set_creator (mf, get_uint32_be (buf, 4));
			}
			if (n >= 10) {
				mf_set_finder_flags (mf, get_uint16_be (buf, 8));
			}
			if (n >= 12) {
				mf_set_position_y (mf, get_uint16_be (buf, 10));
			}
			if (n >= 14) {
				mf_set_position_x (mf, get_uint16_be (buf, 12));
			}
			if (n >= 16) {
				mf_set_folder_id (mf, get_uint16_be (buf, 14));
			}
		}
	}

	n = as_get_entry_size (as, AS_ENTRY_TIME);
	if (n > 0) {
		n = (n < 16) ? n : 16;
		if (as_read (as, buf, AS_ENTRY_TIME, 0, n) == 0) {
			if (n >= 4) {
				mf_set_ctime (mf, as_time_to_mac (get_uint32_be (buf, 0)));
			}
			if (n >= 8) {
				mf_set_mtime (mf, as_time_to_mac (get_uint32_be (buf, 4)));
			}
		}
	}

	n = as_get_entry_size (as, AS_ENTRY_FILE_INFO);
	if (n >= 4) {
		if (as_read (as, buf, AS_ENTRY_FILE_INFO, 0, 4) == 0) {
			mf_set_file_flags (mf, get_uint32_be (buf, 0));
		}
	}
}
