/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/crc.h                                                *
 * Created:     2000-09-26 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#ifndef MAC_LIB_CRC_H
#define MAC_LIB_CRC_H 1


/* some standard polynomials */
#define CRC_CRC16        0x8005
#define CRC_CRC16_CCITT  0x1021
#define CRC_CRC32        0x04c11db7
#define CRC_CRC32_CDROM  0x8001801b
#define CRC_CRC32_CRC32C 0x1edc6f41

/* crc flags for crc_init() */
#define CRC_POSIX       1
#define CRC_REVERSE_INP 2
#define CRC_REVERSE_OUT 4
#define CRC_INV_V0      8
#define CRC_INV_V1      16


/*!***************************************************************************
 * @short The CRC-16 context
 *****************************************************************************/
typedef struct {
	unsigned       flags;
	unsigned short poly;
	unsigned short reg;
	unsigned short out;
	unsigned long  cnt[2];
	unsigned short tab[256];
} crc16_t;


/*!***************************************************************************
 * @short The CRC-32 context
 *****************************************************************************/
typedef struct {
	unsigned      flags;
	unsigned long poly;
	unsigned long reg;
	unsigned long out;
	unsigned long cnt[2];
	unsigned long tab[256];
} crc32_t;


/*!***************************************************************************
 * @short Reverse bit order
 * @param x    The value
 * @param size The number of bits to reverse
 *
 * Reverses the bit order of the low size bits of x.
 *****************************************************************************/
unsigned long crc_reverse (unsigned long x, unsigned size);

/*!***************************************************************************
 * @short Calculate a reciprocal polynomial
 * @param x    The polynomial
 * @param size The size of the polynomial - 1
 *****************************************************************************/
unsigned long crc_reverse_poly (unsigned long x, unsigned size);


/*!***************************************************************************
 * @short Initialize a CRC-16 context
 * @param crc   The CRC-16 context
 * @param poly  The generator polynomial
 * @param flags The CRC flags
 *
 * Initializes the context and creates the crc table.
 *****************************************************************************/
void crc16_init (crc16_t *crc, unsigned poly, unsigned flags);

void crc16_reset (crc16_t *crc, unsigned flags);

void crc16_set_params (crc16_t *crc, unsigned flags, unsigned v0, unsigned v1);

/*!***************************************************************************
 * @short Copy a CRC-16 context
 *****************************************************************************/
void crc16_copy (crc16_t *dst, const crc16_t *src);

/*!***************************************************************************
 * @short  Get the CRC register
 * @return The CRC16 checksum
 *
 * crc16_done() must be called before this.
 *****************************************************************************/
unsigned short crc16_get_crc (const crc16_t *crc);

/*!***************************************************************************
 * @short Calculate a CRC-16 checksum for a byte buffer.
 * @param crc The CRC-16 context
 * @param buf The buffer
 * @param cnt The buffer size
 *
 * This function can be called multiple times.
 *****************************************************************************/
void crc16_calc (crc16_t *crc, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short Finish calculating a CRC-16 checksum
 * @param crc The CRC-16 context
 *
 * This function should be called after the last call to crc16_calc() and
 * before crc16_get_crc().
 *****************************************************************************/
void crc16_done (crc16_t *crc);

unsigned short crc16_crc (const void *buf, unsigned long cnt, unsigned short gen, unsigned flags);


/*!***************************************************************************
 * @short Initialize a CRC-32 context
 * @param poly The CRC polynomial
 *****************************************************************************/
void crc32_init (crc32_t *crc, unsigned long poly, unsigned flags);

/*!***************************************************************************
 * @short Reset a CRC-32 context
 *****************************************************************************/
void crc32_reset (crc32_t *crc, unsigned flags);

void crc32_set_params (crc16_t *crc, unsigned flags, unsigned long v0, unsigned long v1);

/*!***************************************************************************
 * @short Copy a CRC-32 context
 *****************************************************************************/
void crc32_copy (crc32_t *dst, const crc32_t *src);

/*!***************************************************************************
 * @short Get the CRC register.
 *****************************************************************************/
unsigned long crc32_get_crc (const crc32_t *crc);

/*!***************************************************************************
 * @short Calculate a CRC-32 checksum for a byte buffer.
 * @param crc The CRC-32 context
 * @param buf The buffer
 * @param cnt The buffer size
 *
 * This function can be called multiple times.
 *****************************************************************************/
void crc32_calc (crc32_t *crc, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short Finish calculating a CRC-32 checksum
 * @param crc The CRC-32 context
 *
 * This function should be called after the last call to crc32_calc() and
 * before crc32_get_crc().
 *****************************************************************************/
void crc32_done (crc32_t *crc);

unsigned long crc32_crc (const void *buf, unsigned long cnt, unsigned long gen, unsigned flags);


#endif
