/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/dir.c                                                *
 * Created:     2007-12-02 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2016 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <string.h>

#include <dirent.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "dir.h"
#include "strings.h"


static
int ms_stat (mac_stat_t *ms, char *name)
{
	unsigned    mode_mask;
	struct stat st;

#ifdef HAVE_LSTAT
	if (lstat (name, &st)) {
		return (1);
	}

	ms->is_link = (S_ISLNK (st.st_mode) != 0);
#else
	if (stat (name, &st)) {
		return (1);
	}

	ms->is_link = 0;
#endif

	ms->name = name;

	mode_mask = S_IRWXU;

#ifdef S_IRWXG
	mode_mask |= S_IRWXG;
#endif

#ifdef S_IRWXO
	mode_mask |= S_IRWXO;
#endif

	ms->size = st.st_size;
	ms->mode = st.st_mode & mode_mask;
	ms->mtime = st.st_mtime;

	ms->is_file = (S_ISREG (st.st_mode) != 0);
	ms->is_dir = (S_ISDIR (st.st_mode) != 0);

	return (0);
}

static
int ms_compare (const mac_stat_t *ms1, const mac_stat_t *ms2)
{
	if (ms1->is_dir != ms2->is_dir) {
		if (ms1->is_dir) {
			return (-1);
		}
		else {
			return (1);
		}
	}

	return (strcasecmp (ms1->name, ms2->name));
}

static
void msl_del (mac_stat_t *msl, unsigned long cnt)
{
	unsigned long i;

	for (i = 0; i < cnt; i++) {
		free (msl[i].name);
	}

	free (msl);
}

static
void msl_sort (mac_stat_t *msl, unsigned long cnt)
{
	unsigned long i, j;
	mac_stat_t    tmp;

	for (i = 1; i < cnt; i++) {
		if (ms_compare (msl + i, msl + i - 1) >= 0) {
			continue;
		}

		tmp = msl[i];
		msl[i] = msl[i - 1];

		j = i - 1;

		while ((j > 0) && (ms_compare (&tmp, msl + j - 1) < 0)) {
			msl[j] = msl[j - 1];
			j -= 1;
		}

		msl[j] = tmp;
	}
}

static
int msl_read (const char *name, mac_stat_t **msl, unsigned long *cnt)
{
	unsigned long max;
	DIR           *dir;
	struct dirent *de;
	char          *str;
	mac_stat_t    *tmp;

	*msl = NULL;
	*cnt = 0;
	max = 0;

	dir = opendir (name);

	if (dir == NULL) {
		return (1);
	}

	while (1) {
		de = readdir (dir);

		if (de == NULL) {
			break;
		}

		if (strcmp (de->d_name, ".") == 0) {
			continue;
		}

		if (strcmp (de->d_name, "..") == 0) {
			continue;
		}

		if (*cnt >= max) {
			max += (max < 16) ? 16 : max;

			tmp = realloc (*msl, max * sizeof (mac_stat_t));

			if (tmp == NULL) {
				break;
			}

			*msl = tmp;
		}

		str = mac_combine_path (name, de->d_name);

		if (str == NULL) {
			break;
		}

		if (ms_stat (*msl + *cnt, str)) {
			free (str);
			break;
		}

		*cnt += 1;
	}

	closedir (dir);

	if (de != NULL) {
		return (1);
	}

	return (0);
}

static
int list_enter_dir (mac_stat_t *ms, int rec, void *ext, mac_list_f fct)
{
	if (fct (ext, MAC_LIST_ENTER, ms)) {
		return (1);
	}

	if (rec) {
		if (mac_dir_list_dir (ms->name, rec, ext, fct)) {
			return (1);
		}
	}

	if (fct (ext, MAC_LIST_LEAVE, ms)) {
		return (1);
	}

	return (0);
}

/*
 * List the contents of directory name.
 */
int mac_dir_list_dir (const char *name, int rec, void *ext, mac_list_f fct)
{
	int           r;
	unsigned long i, n;
	mac_stat_t    *msl;

	if (msl_read (name, &msl, &n)) {
		msl_del (msl, n);
		return (1);
	}

	msl_sort (msl, n);

	r = 0;

	for (i = 0; i < n; i++) {
		if (fct (ext, MAC_LIST_LIST, msl + i)) {
			r = 1;
			break;
		}

		if (msl[i].is_dir) {
			if (list_enter_dir (msl + i, rec, ext, fct)) {
				r = 1;
				break;
			}
		}
	}

	msl_del (msl, n);

	return (r);
}

/*
 * List base and, if it is a directory, enter it.
 */
int mac_dir_list (const char *base, int rec, void *ext, mac_list_f fct)
{
	char       *str;
	mac_stat_t ms;

	str = mac_combine_path (base, "");

	if (ms_stat (&ms, str)) {
		free (str);
		return (1);
	}

	if (fct (ext, MAC_LIST_LIST, &ms)) {
		free (str);
		return (1);
	}

	if (ms.is_dir) {
		if (list_enter_dir (&ms, rec, ext, fct)) {
			free (str);
			return (1);
		}
	}

	free (str);

	return (0);
}
