/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/macfile.c                                            *
 * Created:     2006-08-19 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2006-2021 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/fourcc.h>
#include <lib/macfile.h>
#include <lib/mactime.h>
#include <lib/strings.h>


#define MACFILE_FOURCC_BINA 0x42494e41
#define MACFILE_FOURCC_UNKN 0x3f3f3f3f


typedef struct {
	unsigned short flag;
	const char     *name;
} mf_finder_flag_t;


static mf_finder_flag_t mf_finder_flags[] = {
	{ MF_FINDER_ALIAS, "ALIAS" },
	{ MF_FINDER_LOCKED, "LOCKED" },
	{ MF_FINDER_INVISIBLE, "INVISIBLE" },
	{ MF_FINDER_BUNDLE, "BUNDLE" },
	{ MF_FINDER_NAME_LOCKED, "NLOCKED" },
	{ MF_FINDER_SYSTEM, "SYSTEM" },
	{ MF_FINDER_STATIONARY, "STATIONARY" },
	{ MF_FINDER_BOZO, "BOZO" },
	{ MF_FINDER_CUSTOM_ICON, "ICON" },
	{ MF_FINDER_BUSY, "BUSY" },
	{ MF_FINDER_CHANGED, "CHANGED" },
	{ MF_FINDER_INITED, "INITED" },
	{ MF_FINDER_NOINITS, "NOINITS" },
	{ MF_FINDER_SHARED, "SHARED" },
	{ MF_FINDER_COLOR4, "COLOR4" },
	{ MF_FINDER_COLOR3, "COLOR3" },
	{ MF_FINDER_COLOR2, "COLOR2" },
	{ MF_FINDER_COLOR1, "COLOR1" },
	{ MF_FINDER_ONDESK, "ONDESK" },
	{ 0, NULL }
};


void mf_init (macfile_t *mf)
{
	mf_set_name_cstr (mf, "Noname");

	mf->data_size = 0;
	mf->rsrc_size = 0;
	mf->comment_size = 0;

	mf->file_type = MACFILE_FOURCC_BINA;
	mf->file_creator = MACFILE_FOURCC_UNKN;

	mf->finder_flags = 0;
	mf->file_flags = 0;

	mf->position_x = 0;
	mf->position_y = 0;

	mf->folder_id = 0;

	mf->scroll_x = 0;
	mf->scroll_y = 0;

	mf->mtime = 0;
	mf->ctime = 0;
}


const unsigned char *mf_get_name_pstr (const macfile_t *mf)
{
	return (mf->name);
}

void mf_set_name_pstr (macfile_t *mf, const void *str)
{
	const unsigned char *tmp = str;

	memcpy (mf->name, tmp, tmp[0] + 1);
}

char *mf_get_name_cstr (const macfile_t *mf)
{
	return (mac_pstring_decode (mf->name, "/"));
}

int mf_set_name_cstr (macfile_t *mf, const char *val)
{
	return (mac_pstring_encode_buf (mf->name, 256, val));
}

unsigned long mf_get_data_size (const macfile_t *mf)
{
	return (mf->data_size);
}

void mf_set_data_size (macfile_t *mf, unsigned long val)
{
	mf->data_size = val;
}


unsigned long mf_get_rsrc_size (const macfile_t *mf)
{
	return (mf->rsrc_size);
}

void mf_set_rsrc_size (macfile_t *mf, unsigned long val)
{
	mf->rsrc_size = val;
}


unsigned long mf_get_comment_size (const macfile_t *mf)
{
	return (mf->comment_size);
}

void mf_set_comment_size (macfile_t *mf, unsigned long val)
{
	mf->comment_size = val;
}


unsigned long mf_get_type (const macfile_t *mf)
{
	return (mf->file_type);
}

void mf_set_type (macfile_t *mf, unsigned long val)
{
	mf->file_type = val;
}


unsigned long mf_get_creator (const macfile_t *mf)
{
	return (mf->file_creator);
}

void mf_set_creator (macfile_t *mf, unsigned long val)
{
	mf->file_creator = val;
}


unsigned short mf_get_window_y1 (const macfile_t *mf)
{
	return ((mf->file_type >> 16) & 0xffff);
}

void mf_set_window_y1 (macfile_t *mf, unsigned short val)
{
	mf->file_type &= 0x0000ffff;
	mf->file_type |= (unsigned long) (val & 0xffff) << 16;
}


unsigned short mf_get_window_x1 (const macfile_t *mf)
{
	return (mf->file_type & 0xffff);
}

void mf_set_window_x1 (macfile_t *mf, unsigned short val)
{
	mf->file_type &= 0xffff0000;
	mf->file_type |= val & 0xffff;
}


unsigned short mf_get_window_y2 (const macfile_t *mf)
{
	return ((mf->file_creator >> 16) & 0xffff);
}

void mf_set_window_y2 (macfile_t *mf, unsigned short val)
{
	mf->file_creator &= 0x0000ffff;
	mf->file_creator |= (unsigned long) (val & 0xffff) << 16;
}


unsigned short mf_get_window_x2 (const macfile_t *mf)
{
	return (mf->file_creator & 0xffff);
}

void mf_set_window_x2 (macfile_t *mf, unsigned short val)
{
	mf->file_creator &= 0xffff0000;
	mf->file_creator |= val & 0xffff;
}


unsigned short mf_get_finder_flags (const macfile_t *mf)
{
	return (mf->finder_flags);
}

void mf_set_finder_flags (macfile_t *mf, unsigned short val)
{
	mf->finder_flags = val;
}


unsigned long mf_get_file_flags (const macfile_t *mf)
{
	return (mf->file_flags);
}

void mf_set_file_flags (macfile_t *mf, unsigned long val)
{
	mf->file_flags = val;
}


unsigned short mf_get_position_x (const macfile_t *mf)
{
	return (mf->position_x);
}

void mf_set_position_x (macfile_t *mf, unsigned short val)
{
	mf->position_x = val;
}


unsigned short mf_get_position_y (const macfile_t *mf)
{
	return (mf->position_y);
}

void mf_set_position_y (macfile_t *mf, unsigned short val)
{
	mf->position_y = val;
}


unsigned short mf_get_folder_id (const macfile_t *mf)
{
	return (mf->folder_id);
}

void mf_set_folder_id (macfile_t *mf, unsigned short val)
{
	mf->folder_id = val;
}


unsigned short mf_get_window_view (const macfile_t *mf)
{
	return (mf->folder_id);
}

void mf_set_window_view (macfile_t *mf, unsigned short val)
{
	mf->folder_id = val;
}


unsigned short mf_get_scroll_x (const macfile_t *mf)
{
	return (mf->scroll_x);
}

void mf_set_scroll_x (macfile_t *mf, unsigned short val)
{
	mf->scroll_x = val;
}


unsigned short mf_get_scroll_y (const macfile_t *mf)
{
	return (mf->scroll_y);
}

void mf_set_scroll_y (macfile_t *mf, unsigned short val)
{
	mf->scroll_y = val;
}


unsigned long mf_get_mtime (const macfile_t *mf)
{
	return (mf->mtime);
}

void mf_set_mtime (macfile_t *mf, unsigned long val)
{
	mf->mtime = val;
}


unsigned long mf_get_ctime (const macfile_t *mf)
{
	return (mf->ctime);
}

void mf_set_ctime (macfile_t *mf, unsigned long val)
{
	mf->ctime = val;
}

void mf_add_times (macfile_t *mf, long val)
{
	unsigned long tmp;

	if (val < 0) {
		tmp = -val;

		mf->mtime = (mf->mtime > tmp) ? (mf->mtime - tmp) : 0;
		mf->ctime = (mf->ctime > tmp) ? (mf->ctime - tmp) : 0;
	}
	else {
		tmp = val;

		mf->mtime += tmp;
		mf->ctime += tmp;
	}
}

const char *mf_finder_flag_to_string (unsigned short v)
{
	unsigned    i;
	static char buf[16];

	i = 0;
	while (mf_finder_flags[i].name != NULL) {
		if (v & mf_finder_flags[i].flag) {
			return (mf_finder_flags[i].name);
		}

		i += 1;
	}

	sprintf (buf, "0x%04X", v);

	return (buf);
}

unsigned short mf_finder_flag_from_string (const char *name)
{
	unsigned i;

	i = 0;
	while (mf_finder_flags[i].name != NULL) {
		if (strcasecmp (name, mf_finder_flags[i].name) == 0) {
			return (mf_finder_flags[i].flag);
		}

		i += 1;
	}

	return ((unsigned short) strtoul (name, NULL, 0));
}

const char *mf_file_flag_to_string (unsigned long v)
{
	static char buf[64];

	if (v & MF_FILE_LOCKED) {
		return ("LOCKED");
	}

	if (v & MF_FILE_PROTECTED) {
		return ("PROTECTED");
	}

	if (v & MF_FILE_ISDIR) {
		return ("ISDIR");
	}

	sprintf (buf, "0x%02lX", v);

	return (buf);
}

unsigned long mf_file_flag_from_string (const char *name)
{
	if (strcasecmp (name, "LOCKED") == 0) {
		return (MF_FILE_LOCKED);
	}

	if (strcasecmp (name, "PROTECTED") == 0) {
		return (MF_FILE_PROTECTED);
	}

	if (strcasecmp (name, "ISDIR") == 0) {
		return (MF_FILE_ISDIR);
	}

	return (strtoul (name, NULL, 0));
}

/*
 * Convert val to a two's complemement 16-bit unsigned integer
 */
static
unsigned mf_get_uint16 (int val)
{
	unsigned ret;

	if (val < 0) {
		ret = (unsigned) -val;
		ret = (~ret + 1) & 0xffff;
	}
	else {
		ret = (unsigned) val;
	}

	return (ret);
}

static
int mf_get_sint16 (unsigned val)
{
	int ret;

	if (val & 0x8000) {
		ret = -(int) ((~val + 1) & 0xffff);
	}
	else {
		ret = (int) val;
	}

	return (ret);
}

void mf_save (macfile_t *mf, FILE *fp)
{
	unsigned long v;
	int           isdir;
	unsigned      y, m, d, hrs, min, sec;
	char          *name;
	char          buf[256];

	isdir = (mf_get_file_flags (mf) & MF_FILE_ISDIR) != 0;

	name = mac_pstring_decode (mf_get_name_pstr (mf), "/\"");

	if (name != NULL) {
		fprintf (fp, "name:\t\t\"%s\"\n", name);
		free (name);
	}

	if (isdir) {
		fprintf (fp, "window-x1:\t%d\n", mf_get_sint16 (mf_get_window_x1 (mf)));
		fprintf (fp, "window-y1:\t%d\n", mf_get_sint16 (mf_get_window_y1 (mf)));
		fprintf (fp, "window-x2:\t%d\n", mf_get_sint16 (mf_get_window_x2 (mf)));
		fprintf (fp, "window-y2:\t%d\n", mf_get_sint16 (mf_get_window_y2 (mf)));

		fprintf (fp, "scroll-x:\t%d\n", mf_get_sint16 (mf_get_scroll_x (mf)));
		fprintf (fp, "scroll-y:\t%d\n", mf_get_sint16 (mf_get_scroll_y (mf)));
	}
	else {
		v = mf_get_type (mf);
		str_set_fourcc (buf, 256, v);
		fprintf (fp, "type:\t\t\"%s\"\n", buf);

		v = mf_get_creator (mf);
		str_set_fourcc (buf, 256, v);
		fprintf (fp, "creator:\t\"%s\"\n", buf);

		fprintf (fp, "data-size:\t%lu\n", mf_get_data_size (mf));

		fprintf (fp, "rsrc-size:\t%lu\n", mf_get_rsrc_size (mf));
	}

	v = mf_get_file_flags (mf);
	fprintf (fp, "file-flags:\t0x%02lX\n", v);

	v = mf_get_finder_flags (mf);
	fprintf (fp, "finder-flags:\t0x%04lX\n", v);

	v = mf_get_position_x (mf);
	fprintf (fp, "position-x:\t%d\n", mf_get_sint16 (v));

	v = mf_get_position_y (mf);
	fprintf (fp, "position-y:\t%d\n", mf_get_sint16 (v));

	if (isdir) {
		v = mf_get_window_view (mf);
		fprintf (fp, "window-view:\t0x%04lx\n", v);
	}
	else {
		v = mf_get_folder_id (mf);
		fprintf (fp, "folder-id:\t0x%04lx\n", v);
	}

	v = mf_get_ctime (mf);
	mac_time_to_ymd (v, &y, &m, &d);
	mac_time_to_hms (v, &hrs, &min, &sec);
	fprintf (fp, "ctime:\t\t\"%04u-%02u-%02u %02u:%02u:%02u\"\n",
		y, m + 1, d + 1, hrs, min, sec
	);

	v = mf_get_mtime (mf);
	mac_time_to_ymd (v, &y, &m, &d);
	mac_time_to_hms (v, &hrs, &min, &sec);
	fprintf (fp, "mtime:\t\t\"%04u-%02u-%02u %02u:%02u:%02u\"\n",
		y, m + 1, d + 1, hrs, min, sec
	);
}

static
void mf_skip_space (FILE *fp)
{
	int c;

	while (1) {
		c = fgetc (fp);

		if (c == EOF) {
			return;
		}
		else if (c == '#') {
			while ((c != 0x0d) && (c != 0x0a)) {
				c = fgetc (fp);
				if (c == EOF) {
					return;
				}
			}
		}
		else if ((c == ' ') || (c == '\t')) {
			;
		}
		else if ((c == 0x0d) || (c == 0x0a)) {
			;
		}
		else {
			ungetc (c, fp);
			return;
		}
	}
}

static
int mf_get_key (FILE *fp, char *key, unsigned max)
{
	int      c;
	unsigned i;

	mf_skip_space (fp);

	i = 0;
	while (i < max) {
		c = fgetc (fp);
		if (c == EOF) {
			break;
		}
		if ((c == ' ') || (c == '\t')) {
			break;
		}
		if (c == ':') {
			break;
		}

		key[i++] = c;
	}

	if (i >= max) {
		return (1);
	}

	key[i] = 0;

	return (0);
}

static
int mf_get_val (FILE *fp, char *val, unsigned max)
{
	int      c;
	int      q;
	unsigned i;

	mf_skip_space (fp);

	i = 0;
	q = 0;
	while (i < max) {
		c = fgetc (fp);
		if (c == EOF) {
			break;
		}

		if (c == '"') {
			if (q) {
				break;
			}
			else if (i == 0) {
				q = 1;
				continue;
			}
		}

		if ((c == 0x0d) || (c == 0x0a)) {
			break;
		}

		val[i++] = c;
	}

	if (i >= max) {
		return (1);
	}

	val[i] = 0;

	return (0);
}

static
int mf_get_key_val (FILE *fp, char *key, unsigned kmax, char *val, unsigned vmax)
{
	if (mf_get_key (fp, key, kmax)) {
		return (1);
	}

	if (mf_get_val (fp, val, vmax)) {
		return (1);
	}

	return (0);
}

int mf_load (macfile_t *mf, FILE *fp)
{
	char          key[64];
	char          val[256];
	unsigned long vu;
	long          vs;

	mf_init (mf);

	while (mf_get_key_val (fp, key, 64, val, 256) == 0) {
		if (key[0] == 0) {
			return (0);
		}

		if (strcmp (key, "name") == 0) {
			if (mf_set_name_cstr (mf, val)) {
				return (1);
			}
		}
		else if (strcmp (key, "type") == 0) {
			if (str_get_fourcc (val, &vu)) {
				return (1);
			}

			mf_set_type (mf, vu);
		}
		else if (strcmp (key, "creator") == 0) {
			if (str_get_fourcc (val, &vu)) {
				return (1);
			}

			mf_set_creator (mf, vu);
		}
		else if (strcmp (key, "window-y1") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_window_y1 (mf, vu);
		}
		else if (strcmp (key, "window-x1") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_window_x1 (mf, vu);
		}
		else if (strcmp (key, "window-y2") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_window_y2 (mf, vu);
		}
		else if (strcmp (key, "window-x2") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_window_x2 (mf, vu);
		}
		else if (strcmp (key, "data-size") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_data_size (mf, vu);
		}
		else if (strcmp (key, "rsrc-size") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_rsrc_size (mf, vu);
		}
		else if (strcmp (key, "scroll-x") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_scroll_x (mf, vu);
		}
		else if (strcmp (key, "scroll-y") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_scroll_y (mf, vu);
		}
		else if (strcmp (key, "file-flags") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_file_flags (mf, vu);
		}
		else if (strcmp (key, "finder-flags") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_finder_flags (mf, vu);
		}
		else if (strcmp (key, "position-x") == 0) {
			vs = strtol (val, NULL, 0);
			mf_set_position_x (mf, mf_get_uint16 (vs));
		}
		else if (strcmp (key, "position-y") == 0) {
			vs = strtol (val, NULL, 0);
			mf_set_position_y (mf, mf_get_uint16 (vs));
		}
		else if (strcmp (key, "folder-id") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_folder_id (mf, vu);
		}
		else if (strcmp (key, "window-view") == 0) {
			vu = strtoul (val, NULL, 0);
			mf_set_window_view (mf, vu);
		}
		else if (strcmp (key, "ctime") == 0) {
			vu = mac_time_from_string (val);
			mf_set_ctime (mf, vu);
		}
		else if (strcmp (key, "mtime") == 0) {
			vu = mac_time_from_string (val);
			mf_set_mtime (mf, vu);
		}
		else {
			fprintf (stderr, "warning: unknown key (%s)\n", key);
		}
	}

	return (0);
}

void mf_print_info (macfile_t *mf, FILE *fp)
{
	unsigned long v;
	unsigned      y, m, d, hrs, min, sec;
	char          *name;
	char          buf[256];

	name = mf_get_name_cstr (mf);
	if (name != NULL) {
		fprintf (fp, "file name:          '%s'\n", name);
		free (name);
	}

	v = mf_get_type (mf);
	str_set_fourcc (buf, 256, v);
	fprintf (fp, "type:               \"%s\" (0x%08lX)\n", buf, v);

	v = mf_get_creator (mf);
	str_set_fourcc (buf, 256, v);
	fprintf (fp, "creator:            \"%s\" (0x%08lX)\n", buf, v);

	v = mf_get_data_size (mf);
	fprintf (fp, "data fork size:     %lu\n", v);

	v = mf_get_rsrc_size (mf);
	fprintf (fp, "rsrc fork size:     %lu\n", v);

	v = mf_get_finder_flags (mf);
	fprintf (fp, "finder flags:       0x%04lX\n", v);

	v = mf_get_file_flags (mf);
	fprintf (fp, "file flags:         0x%02lX\n", v);

	v = mf_get_position_x (mf);
	fprintf (fp, "position x:         %d\n", mf_get_sint16 (v));

	v = mf_get_position_y (mf);
	fprintf (fp, "position y:         %d\n", mf_get_sint16 (v));

	v = mf_get_folder_id (mf);
	fprintf (fp, "folder id:          0x%04lX\n", v);

	v = mf_get_ctime (mf);
	mac_time_to_ymd (v, &y, &m, &d);
	mac_time_to_hms (v, &hrs, &min, &sec);
	fprintf (fp, "creation time:      "
		"%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
		y, m + 1, d + 1, hrs, min, sec, v
	);

	v = mf_get_mtime (mf);
	mac_time_to_ymd (v, &y, &m, &d);
	mac_time_to_hms (v, &hrs, &min, &sec);
	fprintf (fp, "modification time:  "
		"%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
		y, m + 1, d + 1, hrs, min, sec, v
	);

	fflush (fp);
}
