/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/rsrc.c                                               *
 * Created:     2004-03-21 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/rsrc.h>


rsrc_t *rsc_new (unsigned long type, unsigned id)
{
	rsrc_t *r;

	r = malloc (sizeof (rsrc_t));
	if (r == NULL) {
		return (NULL);
	}

	r->type = type;
	r->id = id;
	r->size = 0;

	r->name = NULL;
	r->name_cnt = 0;

	r->data = NULL;

	r->fp = NULL;
	r->offset = 0;

	return (r);
}

void rsc_del (rsrc_t *r)
{
	if (r != NULL) {
		free (r->data);
		free (r->name);
		free (r);
	}
}

unsigned long rsc_get_type (rsrc_t *r)
{
	return (r->type);
}

void rsc_set_type (rsrc_t *r, unsigned long type)
{
	r->type = type & 0xffffffff;
}

unsigned rsc_get_id (rsrc_t *r)
{
	return (r->id);
}

void rsc_set_id (rsrc_t *r, unsigned id)
{
	r->id = id & 0xffff;
}

unsigned rsc_get_name_cnt (rsrc_t *r)
{
	return (r->name_cnt);
}

unsigned char *rsc_get_name (rsrc_t *r)
{
	return (r->name);
}

int rsc_set_name (rsrc_t *r, const unsigned char *str, unsigned cnt)
{
	unsigned char *tmp;

	tmp = realloc (r->name, (cnt == 0) ? 1 : cnt);
	if (tmp == NULL) {
		return (1);
	}

	if (cnt > 0) {
		memcpy (tmp, str, cnt);
	}

	r->name = tmp;
	r->name_cnt = cnt;

	return (0);
}

unsigned long rsc_get_size (rsrc_t *r)
{
	return (r->size);
}

unsigned char *rsc_get_data (rsrc_t *r)
{
	if (r->data != NULL) {
		return (r->data);
	}

	if (r->size == 0) {
		return (NULL);
	}

	r->data = malloc (r->size);
	if (r->data == NULL) {
		return (NULL);
	}

	if (mac_read (r->fp, r->data, r->offset, r->size)) {
		free (r->data);
		r->data = NULL;
		return (NULL);
	}

	return (r->data);
}

void rsc_set_data (rsrc_t *r, void *buf, unsigned long cnt)
{
	free (r->data);

	r->data = buf;
	r->size = cnt;
}


rsrc_type_t *rt_new (unsigned long type)
{
	rsrc_type_t *rt;

	rt = malloc (sizeof (rsrc_type_t));
	if (rt == NULL) {
		return (NULL);
	}

	rt->type = type;

	rt->dupcnt = 0;

	rt->rsrc_cnt = 0;
	rt->rsrc = NULL;

	return (rt);
}

void rt_del (rsrc_type_t *rt)
{
	unsigned i;

	if (rt != NULL) {
		for (i = 0; i < rt->rsrc_cnt; i++) {
			rsc_del (rt->rsrc[i]);
		}

		free (rt->rsrc);
		free (rt);
	}
}

rsrc_t *rt_get_resource (rsrc_type_t *rt, unsigned id)
{
	unsigned i;

	for (i = 0; i < rt->rsrc_cnt; i++) {
		if (rt->rsrc[i]->id == id) {
			return (rt->rsrc[i]);
		}
	}

	return (NULL);
}

rsrc_t *rt_add_resource (rsrc_type_t *rt, rsrc_t *r)
{
	rsrc_t   **tmp;

	if (rt_get_resource (rt, r->id) != NULL) {
		rt->dupcnt += 1;
	}

	tmp = realloc (rt->rsrc, (rt->rsrc_cnt + 1) * sizeof (rsrc_t *));
	if (tmp == NULL) {
		return (NULL);
	}

	r->type = rt->type;

	rt->rsrc = tmp;
	rt->rsrc[rt->rsrc_cnt] = r;
	rt->rsrc_cnt += 1;

	return (r);
}

rsrc_t *rt_new_resource (rsrc_type_t *rt, unsigned id)
{
	rsrc_t *r;

	r = rsc_new (rt->type, id);
	if (r == NULL) {
		return (NULL);
	}

	if (rt_add_resource (rt, r) == NULL) {
		rsc_del (r);
		return (NULL);
	}

	return (r);
}

void rt_sort (rsrc_type_t *rt, int revrsc)
{
	unsigned i, j;
	rsrc_t   **r, *t;

	r = rt->rsrc;

	for (i = 1; i < rt->rsrc_cnt; i++) {
		t = r[i];
		j = i;

		if (revrsc) {
			while ((j > 0) && (t->id >= r[j - 1]->id)) {
				r[j] = r[j - 1];
				j -= 1;
			}
		}
		else {
			while ((j > 0) && (t->id < r[j - 1]->id)) {
				r[j] = r[j - 1];
				j -= 1;
			}
		}

		r[j] = t;
	}
}

unsigned rt_remove_duplicates (rsrc_type_t *rt)
{
	unsigned i, j, k, n;
	rsrc_t   **r;

	r = rt->rsrc;

	n = 0;

	i = 0;
	while (i < rt->rsrc_cnt) {
		j = i + 1;
		while (j < rt->rsrc_cnt) {
			if (r[i]->id == r[j]->id) {
				rsc_del (r[j]);
				for (k = j + 1; k < rt->rsrc_cnt; k++) {
					r[k - 1] = r[k];
				}

				n += 1;
				rt->rsrc_cnt -= 1;
			}

			j += 1;
		}

		i += 1;
	}

	rt->dupcnt = 0;

	return (n);
}


rsrc_file_t *rf_new (void)
{
	rsrc_file_t *rf;

	rf = malloc (sizeof (rsrc_file_t));
	if (rf == NULL) {
		return (NULL);
	}

	rf->attr = 0;
	rf->dupcnt = 0;

	rf->type_cnt = 0;
	rf->type = NULL;

	return (rf);
}

void rf_del (rsrc_file_t *rf)
{
	unsigned i;

	if (rf != NULL) {
		for (i = 0; i < rf->type_cnt; i++) {
			rt_del (rf->type[i]);
		}

		free (rf);
	}
}

rsrc_type_t *rf_get_type (rsrc_file_t *rf, unsigned long type)
{
	unsigned i;

	for (i = 0; i < rf->type_cnt; i++) {
		if (rf->type[i]->type == type) {
			return (rf->type[i]);
		}
	}

	return (NULL);
}

rsrc_type_t *rf_add_type (rsrc_file_t *rf, rsrc_type_t *rt)
{
	rsrc_type_t **tmp;

	if (rf_get_type (rf, rt->type) != NULL) {
		return (NULL);
	}

	tmp = realloc (rf->type, (rf->type_cnt + 1) * sizeof (rsrc_type_t *));
	if (tmp == NULL) {
		return (NULL);
	}

	rf->type = tmp;
	rf->type[rf->type_cnt] = rt;
	rf->type_cnt += 1;

	return (rt);
}

rsrc_type_t *rf_new_type (rsrc_file_t *rf, unsigned long type)
{
	rsrc_type_t *rt;

	if (rf_get_type (rf, type) != NULL) {
		return (NULL);
	}

	rt = rt_new (type);
	if (rt == NULL) {
		return (NULL);
	}

	if (rf_add_type (rf, rt) == NULL) {
		rt_del (rt);
		return (NULL);
	}

	return (rt);
}

rsrc_t *rf_add_resource (rsrc_file_t *rf, rsrc_t *r)
{
	unsigned    i;
	rsrc_type_t *rt;

	for (i = 0; i < rf->type_cnt; i++) {
		if (rf->type[i]->type == r->type) {
			return (rt_add_resource (rf->type[i], r));
		}
	}

	rt = rf_new_type (rf, r->type);
	if (rt == NULL) {
		return (NULL);
	}

	return (rt_add_resource (rt, r));
}

void rf_sort (rsrc_file_t *rf, int revtyp, int revrsc)
{
	unsigned    i, j;
	rsrc_type_t **rt, *t;

	if (rf->type_cnt == 0) {
		return;
	}

	rt_sort (rf->type[0], revrsc);

	rt = rf->type;

	for (i = 1; i < rf->type_cnt; i++) {
		t = rt[i];

		rt_sort (t, revrsc);

		j = i;

		if (revtyp) {
			while ((j > 0) && (t->type >= rt[j - 1]->type)) {
				rt[j] = rt[j - 1];
				j -= 1;
			}
		}
		else {
			while ((j > 0) && (t->type < rt[j - 1]->type)) {
				rt[j] = rt[j - 1];
				j -= 1;
			}
		}

		rt[j] = t;
	}
}

unsigned long rf_remove_duplicates (rsrc_file_t *rf)
{
	unsigned      i;
	unsigned long n;

	n = 0;

	for (i = 0; i < rf->type_cnt; i++) {
		n += rt_remove_duplicates (rf->type[i]);
	}

	rf->dupcnt = 0;

	return (n);
}

unsigned long rf_get_resource_count (rsrc_file_t *rf)
{
	unsigned      i;
	unsigned long n;

	n = 0;

	for (i = 0; i < rf->type_cnt; i++) {
		n += rf->type[i]->rsrc_cnt;
	}

	return (n);
}


static
unsigned long rf_get_name_size (rsrc_file_t *rf)
{
	unsigned      i, j;
	unsigned long n;
	rsrc_type_t   *rt;
	rsrc_t        *r;

	n = 0;

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		for (j = 0; j < rt->rsrc_cnt; j++) {
			r = rt->rsrc[j];

			if (r->name_cnt > 0) {
				n += r->name_cnt + 1;
			}
		}
	}

	return (n);
}

static
unsigned long rf_get_resource_size (rsrc_file_t *rf)
{
	unsigned      i, j;
	unsigned long n;
	rsrc_type_t   *rt;

	n = 0;

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		for (j = 0; j < rt->rsrc_cnt; j++) {
			n += rt->rsrc[j]->size + 4;
		}
	}

	return (n);
}


static
int rt_read_resources (rsrc_type_t *rt, FILE *fp,
	unsigned long refofs, unsigned long rnlofs, unsigned long datofs,
	unsigned rcnt)
{
	unsigned      i;
	unsigned      nofs, dofs;
	unsigned      id;
	unsigned char buf[12];
	rsrc_t        *r;

	for (i = 0; i < rcnt; i++) {
		if (mac_read (fp, buf, refofs + 12 * i, 12)) {
			return (1);
		}

		id = get_uint16_be (buf, 0);

		r = rt_new_resource (rt, id);
		if (r == NULL) {
			return (1);
		}

		nofs = get_uint16_be (buf, 2);
		dofs = get_uint32_be (buf, 4) & 0x00ffffff;

		r->attr = get_uint8_be (buf, 4);

		r->fp = fp;
		r->offset = datofs + dofs + 4;

		if (mac_read (fp, buf, datofs + dofs, 4)) {
			return (1);
		}

		r->size = get_uint32_be (buf, 0);

		if (nofs != 0xffff) {
			if (mac_read (fp, buf, rnlofs + nofs, 1)) {
				return (1);
			}

			r->name_cnt = get_uint8_be (buf, 0);

			if (r->name_cnt > 0) {
				r->name = malloc (r->name_cnt);
				if (r->name == NULL) {
					r->name_cnt = 0;
					return (1);
				}

				if (mac_read (fp, r->name, rnlofs + nofs + 1, r->name_cnt)) {
					return (1);
				}
			}
		}
	}

	return (0);
}

static
int rf_read_types (rsrc_file_t *rf, FILE *fp,
	unsigned long rtlofs, unsigned long rnlofs, unsigned long datofs,
	unsigned tcnt)
{
	unsigned      i;
	unsigned long type;
	unsigned      rcnt;
	unsigned      refofs;
	unsigned char buf[8];
	rsrc_type_t   *rt;

	for (i = 0; i < tcnt; i++) {
		if (mac_read (fp, buf, rtlofs + 2 + 8 * i, 8)) {
			return (1);
		}

		type = get_uint32_be (buf, 0);

		rt = rf_new_type (rf, type);
		if (rt == NULL) {
			return (1);
		}

		rcnt = get_uint16_be (buf, 4) + 1;
		refofs = get_uint16_be (buf, 6);

		if (rt_read_resources (rt, fp, rtlofs + refofs, rnlofs, datofs, rcnt)) {
			return (1);
		}

		rf->dupcnt += rt->dupcnt;
	}

	return (0);
}

static
int rf_read_rmap (rsrc_file_t *rf, FILE *fp,
	unsigned long mapofs, unsigned long datofs)
{
	unsigned      rtlofs;
	unsigned      rnlofs;
	unsigned      tcnt;
	unsigned char buf[30];

	if (mac_read (fp, buf, mapofs, 30)) {
		return (1);
	}

	rf->attr = get_uint16_be (buf, 22);

	rtlofs = get_uint16_be (buf, 24);
	rnlofs = get_uint16_be (buf, 26);
	tcnt = (get_uint16_be (buf, 28) + 1) & 0xffff;

	if (tcnt == 0) {
		return (0);
	}

	if (rf_read_types (rf, fp, mapofs + rtlofs, mapofs + rnlofs, datofs, tcnt)) {
		return (1);
	}

	return (0);
}

static
int rf_read_header (rsrc_file_t *rf, FILE *fp, unsigned long ofs)
{
	unsigned long datofs;
	unsigned long mapofs;
	unsigned char buf[16];
	/*
	unsigned long datcnt;
	unsigned long mapcnt;
	*/

	if (mac_read (fp, buf, ofs, 16)) {
		return (1);
	}

	datofs = get_uint32_be (buf, 0);
	mapofs = get_uint32_be (buf, 4);

	/*
	datcnt = get_uint32_be (buf, 8);
	mapcnt = get_uint32_be (buf, 12);
	*/

	if (rf_read_rmap (rf, fp, ofs + mapofs, ofs + datofs)) {
		return (1);
	}

	return (0);
}

rsrc_file_t *rf_read (FILE *fp)
{
	unsigned long ofs;
	rsrc_file_t   *rf;

	ofs = ftell (fp);

	rf = rf_new();
	if (rf == NULL) {
		return (NULL);
	}

	if (rf_read_header (rf, fp, ofs)) {
		rf_del (rf);
		return (NULL);
	}

	return (rf);
}


static
int rf_set_offset (FILE *fp, unsigned long cur, unsigned long ofs)
{
	if (cur > ofs) {
		return (1);
	}

	if (cur == ofs) {
		return (0);
	}

	while (cur < ofs) {
		fputc (0, fp);
		cur += 1;
	}

	return (0);
}

static
int rf_write_data (rsrc_file_t *rf, FILE *fp, unsigned long datofs)
{
	unsigned      i, j;
	unsigned char buf[4];
	rsrc_type_t   *rt;
	rsrc_t        *r;

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		for (j = 0; j < rt->rsrc_cnt; j++) {
			r = rt->rsrc[j];

			set_uint32_be (buf, 0, r->size);

			if (fwrite (buf, 4, 1, fp) != 1) {
				return (1);
			}

			if (r->size > 0) {
				unsigned char *dat;

				dat = rsc_get_data (r);
				if (dat == NULL) {
					return (1);
				}

				if (fwrite (dat, r->size, 1, fp) != 1) {
					return (1);
				}
			}
		}
	}

	return (0);
}

static
int rf_write_rmap (rsrc_file_t *rf, FILE *fp,
	unsigned long mapofs, unsigned long rtlofs, unsigned long rnlofs)
{
	unsigned char buf[28];

	set_uint32_be (buf, 0, 0);
	set_uint32_be (buf, 4, 0);
	set_uint32_be (buf, 8, 0);
	set_uint32_be (buf, 12, 0);
	set_uint32_be (buf, 16, 0);
	set_uint16_be (buf, 20, 0);
	set_uint16_be (buf, 22, rf->attr);
	set_uint16_be (buf, 24, rtlofs - mapofs);
	set_uint16_be (buf, 26, rnlofs - mapofs);

	if (fwrite (buf, 28, 1, fp) != 1) {
		return (1);
	}

	return (0);
}

static
int rf_write_rtl (rsrc_file_t *rf, FILE *fp,
	unsigned long rtlofs, unsigned long refofs)
{
	unsigned      i;
	unsigned long refidx;
	unsigned char buf[8];
	rsrc_type_t   *rt;

	set_uint16_be (buf, 0, (rf->type_cnt - 1) & 0xffff);

	if (fwrite (buf, 2, 1, fp) != 1) {
		return (1);
	}

	refidx = 0;

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		if (rt->rsrc_cnt > 0) {
			set_uint32_be (buf, 0, rt->type);
			set_uint16_be (buf, 4, rt->rsrc_cnt - 1);
			set_uint16_be (buf, 6, refofs - rtlofs + refidx);

			refidx += 12 * rt->rsrc_cnt;

			if (fwrite (buf, 8, 1, fp) != 1) {
				return (1);
			}
		}
	}

	return (0);
}

static
int rf_write_ref (rsrc_file_t *rf, FILE *fp, unsigned long refofs)
{
	unsigned      i, j;
	unsigned long rnlidx;
	unsigned long datidx;
	unsigned char buf[12];
	rsrc_type_t   *rt;
	rsrc_t        *r;

	rnlidx = 0;
	datidx = 0;

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		for (j = 0; j < rt->rsrc_cnt; j++) {
			r = rt->rsrc[j];

			set_uint16_be (buf, 0, r->id);
			if (r->name_cnt == 0) {
				set_uint16_be (buf, 2, 0xffff);
			}
			else {
				set_uint16_be (buf, 2, rnlidx);
				rnlidx += r->name_cnt + 1;
			}

			set_uint32_be (buf, 4, datidx);
			datidx += r->size + 4;

			set_uint8_be (buf, 4, r->attr);
			set_uint32_be (buf, 8, 0);

			if (fwrite (buf, 12, 1, fp) != 1) {
				return (1);
			}
		}
	}

	return (0);
}

static
int rf_write_rnl (rsrc_file_t *rf, FILE *fp, unsigned long rnlofs)
{
	unsigned      i, j;
	unsigned char buf[4];
	rsrc_type_t   *rt;
	rsrc_t        *r;

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		for (j = 0; j < rt->rsrc_cnt; j++) {
			r = rt->rsrc[j];

			if (r->name_cnt > 0) {
				set_uint8_be (buf, 0, r->name_cnt);

				if (fwrite (buf, 1, 1, fp) != 1) {
					return (1);
				}

				if (fwrite (r->name, r->name_cnt, 1, fp) != 1) {
					return (1);
				}
			}
		}
	}

	return (0);
}

int rf_write (rsrc_file_t *rf, FILE *fp)
{
	unsigned long tcnt, rcnt;
	unsigned long rtlofs, rtlcnt;
	unsigned long mapofs, mapcnt;
	unsigned long refofs, refcnt;
	unsigned long rnlofs, rnlcnt;
	unsigned long datofs, datcnt;
	unsigned char buf[16];

	tcnt = rf->type_cnt;
	rcnt = rf_get_resource_count (rf);

	datofs = 256;
	datcnt = rf_get_resource_size (rf);

	mapofs = datofs + datcnt;
	rtlofs = mapofs + 28;
	rtlcnt = 8 * tcnt + 2;
	refofs = rtlofs + rtlcnt;
	refcnt = 12 * rcnt;
	rnlofs = refofs + refcnt;
	rnlcnt = rf_get_name_size (rf);
	mapcnt = 28 + rtlcnt + refcnt + rnlcnt;

	set_uint32_be (buf, 0, datofs);
	set_uint32_be (buf, 4, mapofs);
	set_uint32_be (buf, 8, datcnt);
	set_uint32_be (buf, 12, mapcnt);

	if (fwrite (buf, 16, 1, fp) != 1) {
		return (1);
	}

	if (rf_set_offset (fp, 16, datofs)) {
		return (1);
	}

	if (rf_write_data (rf, fp, datofs)) {
		return (1);
	}

	if (rf_set_offset (fp, datofs + datcnt, mapofs)) {
		return (1);
	}

	if (rf_write_rmap (rf, fp, mapofs, rtlofs, rnlofs)) {
		return (1);
	}

	if (rf_write_rtl (rf, fp, rtlofs, refofs)) {
		return (1);
	}

	if (rf_write_ref (rf, fp, refofs)) {
		return (1);
	}

	if (rf_write_rnl (rf, fp, rnlofs)) {
		return (1);
	}

	return (0);
}
