/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/strings.c                                            *
 * Created:     2007-12-02 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "strings.h"


/*
 * Check if c is to be quoted
 */
static
int chr_check_quote (unsigned char c, const char *quote)
{
	if ((c < 32) || (c > 126)) {
		return (1);
	}

	if (c == '=') {
		return (1);
	}

	if (quote != NULL) {
		while (*quote != 0) {
			if (*quote == c) {
				return (1);
			}

			quote += 1;
		}
	}

	return (0);
}

static
char chr_get_hex_digit (unsigned char c)
{
	c &= 0x0f;

	if (c <= 9) {
		return ('0' + c);
	}

	return ('a' + (c - 10));
}

static
void chr_quote (char *dst, unsigned char c)
{
	dst[0] = '=';
	dst[1] = chr_get_hex_digit (c >> 4);
	dst[2] = chr_get_hex_digit (c);
}

static
int chr_set_hex_digit (unsigned char *d, char s)
{
	if ((s >= '0') && (s <= '9')) {
		*d = s - '0';
	}
	else if ((s >= 'A') && (s <= 'F')) {
		*d = s - 'A' + 10;
	}
	else if ((s >= 'a') && (s <= 'f')) {
		*d = s - 'a' + 10;
	}
	else {
		return (1);
	}

	return (0);
}

static
int chr_unquote (unsigned char *buf, const char *src)
{
	unsigned char d1, d2;

	if (chr_set_hex_digit (&d1, src[0])) {
		return (1);
	}

	if (chr_set_hex_digit (&d2, src[1])) {
		return (1);
	}

	*buf = (d1 << 4) | d2;

	return (0);
}

int mac_string_decode_buf (char *dst, unsigned max,
	const unsigned char *src, unsigned cnt, const char *qc)
{
	unsigned i, j;

	j = 0;
	for (i = 0; i < cnt; i++) {
		if (chr_check_quote (src[i], qc)) {
			if ((j + 3) >= max) {
				return (1);
			}

			chr_quote (dst + j, src[i]);

			j += 3;
		}
		else {
			if ((j + 1) >= max) {
				return (1);
			}

			dst[j] = (char) src[i];

			j += 1;
		}
	}

	dst[j] = 0;

	return (0);
}

int mac_pstring_decode_buf (char *dst, unsigned max,
	const unsigned char *src, const char *qc)
{
	return (mac_string_decode_buf (dst, max, src + 1, src[0], qc));
}

char *mac_string_decode (const unsigned char *src, unsigned cnt, const char *qc)
{
	unsigned i, n;
	char     *ret;

	n = 0;

	for (i = 0; i < cnt; i++) {
		if (chr_check_quote (src[i], qc)) {
			n += 3;
		}
		else {
			n += 1;
		}
	}

	ret = malloc (n + 1);
	if (ret == NULL) {
		return (NULL);
	}

	if (mac_string_decode_buf (ret, n + 1, src, cnt, qc)) {
		free (ret);
		return (NULL);
	}

	return (ret);
}

char *mac_pstring_decode (const unsigned char *src, const char *qc)
{
	return (mac_string_decode (src + 1, src[0], qc));
}


int mac_string_encode_buf (unsigned char *dst, unsigned max,
	const char *src, unsigned *cnt)
{
	unsigned i, j;

	i = 0;
	j = 0;
	while ((j < max) && (src[i] != 0)) {
		if (src[i] == '=') {
			if (src[i + 1] == '=') {
				dst[j] = '=';
				i += 2;
			}
			else {
				if (chr_unquote (dst + j, src + i + 1)) {
					break;
				}

				i += 3;
			}
		}
		else {
			dst[j] = (unsigned char) src[i];

			i += 1;
		}

		j += 1;
	}

	*cnt = j;

	if (src[i] != 0) {
		return (1);
	}

	return (0);
}

int mac_pstring_encode_buf (unsigned char *dst, unsigned max, const char *src)
{
	unsigned cnt;

	if (max < 1) {
		return (1);
	}

	if (mac_string_encode_buf (dst + 1, max - 1, src, &cnt)) {
		return (1);
	}

	if (cnt > 255) {
		return (1);
	}

	dst[0] = cnt;

	return (0);
}

static
unsigned char *mac_string_encode_alloc (const char *src, unsigned *cnt, unsigned res)
{
	unsigned      i, n;
	unsigned char *ret;

	n = res;

	i = 0;
	while (src[i] != 0) {
		if (src[i] == '=') {
			if (src[i + 1] == '=') {
				i += 2;
			}
			else {
				i += 3;
			}
		}
		else {
			i += 1;
		}

		n += 1;
	}

	ret = malloc ((n == 0) ? 1 : n);
	if (ret == NULL) {
		return (NULL);
	}

	if (mac_string_encode_buf (ret + res, n - res, src, cnt)) {
		free (ret);
		return (NULL);
	}

	return (ret);
}

unsigned char *mac_string_encode (const char *src, unsigned *cnt)
{
	return (mac_string_encode_alloc (src, cnt, 0));
}

unsigned char *mac_pstring_encode (const char *src)
{
	unsigned      cnt;
	unsigned char *ret;

	ret = mac_string_encode_alloc (src, &cnt, 1);

	if (ret == NULL) {
		return (NULL);
	}

	if (cnt > 255) {
		free (ret);
		return (NULL);
	}

	ret[0] = cnt;

	return (ret);
}

/*
 * get str1 + "/" + str2
 */
char *mac_combine_path (const char *str1, const char *str2)
{
	unsigned long n1, n2;
	char          *ret;

	while (*str2 == '/') {
		str2 += 1;
	}

	n1 = strlen (str1);
	n2 = strlen (str2);

	while ((n1 > 1) && (str1[n1 - 1] == '/')) {
		n1 -= 1;
	}

	while ((n2 > 0) && (str2[n2 - 1] == '/')) {
		n2 -= 1;
	}

	ret = malloc (n1 + n2 + 2);

	if (n1 > 0) {
		memcpy (ret, str1, n1);

		if ((ret[n1 - 1] != '/') && (n2 > 0)) {
			ret[n1++] = '/';
		}
	}

	if (n2 > 0) {
		memcpy (ret + n1, str2, n2);
	}

	ret[n1 + n2] = 0;

	return (ret);
}

const char *mac_get_last_component (const char *str, char sep)
{
	unsigned i;

	i = strlen (str);

	while ((i > 0) && (str[i - 1] == sep)) {
		i -= 1;
	}

	while ((i > 0) && (str[i - 1] != sep)) {
		i -= 1;
	}

	return (str + i);
}
