/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_inp_as.c                                       *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/applesingle.h>
#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/macfile.h>
#include <lib/unixfile.h>


typedef struct {
	macc_inp_t    inp;

	applesingle_t as;

	unsigned long data_offset;
	unsigned long rsrc_offset;

	char          *data_name;
	FILE          *data_fp;
} macc_inp_as_t;


static
void macc_as_close (macc_inp_t *minp)
{
	macc_inp_as_t *as;

	as = minp->ext;

	mac_close (as->data_fp);
	free (as->data_name);

	mac_close (minp->fp);
	free (minp->ext);
}

static
int macc_as_read_header (macc_inp_t *minp, macfile_t *mf)
{
	unsigned long  mtime;
	macc_inp_as_t *as;

	as = minp->ext;

	as_init (&as->as, minp->fp);

	if (as_read_header (&as->as)) {
		fprintf (stderr, "%s: as header error\n", arg0);
		return (1);
	}

	mf_init (mf);

	mtime = uf_get_mtime_fp ((as->data_fp != NULL) ? as->data_fp : minp->fp);

	mf_set_mtime (mf, mtime);
	mf_set_ctime (mf, mtime);

	as_get_macfile (&as->as, mf);

	if (as->data_fp != NULL) {
		mf_set_data_size (mf, uf_get_fsize_fp (as->data_fp));
	}

	return (0);
}

static
int macc_as_read_data_start (macc_inp_t *minp)
{
	macc_inp_as_t *as;

	as = minp->ext;

	as->data_offset = 0;

	return (0);
}

static
int macc_as_read_data (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	macc_inp_as_t *as;

	as = minp->ext;

	if (as->data_fp != NULL) {
		if (fread (buf, 1, cnt, as->data_fp) != cnt) {
			fprintf (stderr, "%s: as data read error\n", arg0);
			return (1);
		}
	}
	else {
		if (as_read (&as->as, buf, AS_ENTRY_DATA, as->data_offset, cnt)) {
			fprintf (stderr, "%s: as data read error\n", arg0);
			return (1);
		}
	}

	as->data_offset += cnt;

	return (0);
}

static
int macc_as_read_data_done (macc_inp_t *minp)
{
	return (0);
}

static
int macc_as_read_rsrc_start (macc_inp_t *minp)
{
	macc_inp_as_t *as;

	as = minp->ext;

	as->rsrc_offset = 0;

	return (0);
}

static
int macc_as_read_rsrc (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	macc_inp_as_t *as;

	as = minp->ext;

	if (as_read (&as->as, buf, AS_ENTRY_RSRC, as->rsrc_offset, cnt)) {
		fprintf (stderr, "%s: as resource read error\n", arg0);
		return (1);
	}

	as->rsrc_offset += cnt;

	return (0);
}

static
int macc_as_read_rsrc_done (macc_inp_t *minp)
{
	return (0);
}

static
int macc_as_read_comm_start (macc_inp_t *minp)
{
	return (0);
}

static
int macc_as_read_comm (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	return (cnt > 0);
}

static
int macc_as_read_comm_done (macc_inp_t *minp)
{
	return (0);
}

macc_inp_t *macc_inp_as_open (const char *fname, const char *dname)
{
	macc_inp_t     *minp;
	macc_inp_as_t *as;

	as = malloc (sizeof (macc_inp_as_t));

	if (as == NULL) {
		return (NULL);
	}

	minp = &as->inp;

	minp->ext = as;

	minp->close = macc_as_close;

	minp->read_header = macc_as_read_header;

	minp->read_data_start = macc_as_read_data_start;
	minp->read_data = macc_as_read_data;
	minp->read_data_done = macc_as_read_data_done;

	minp->read_rsrc_start = macc_as_read_rsrc_start;
	minp->read_rsrc = macc_as_read_rsrc;
	minp->read_rsrc_done = macc_as_read_rsrc_done;

	minp->read_comm_start = macc_as_read_comm_start;
	minp->read_comm = macc_as_read_comm;
	minp->read_comm_done = macc_as_read_comm_done;

	minp->fp = mac_open_inp (fname, "rb", 1);

	if (minp->fp == NULL) {
		free (as);
		return (NULL);
	}

	if (dname == NULL) {
		as->data_name = NULL;
		as->data_fp = NULL;
	}
	else {
		as->data_fp = mac_open_inp (dname, "rb", 1);

		if (as->data_fp == NULL) {
			mac_close (minp->fp);
			free (as);
			return (NULL);
		}

		as->data_name = strdup (dname);
	}

	return (minp);
}
