/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_inp_hcx.c                                      *
 * Created:     2013-05-27 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2013 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/macfile.h>
#include <lib/unixfile.h>


typedef struct {
	macc_inp_t     inp;

	char           fname[64];

	unsigned long  data_ofs;
	unsigned long  data_cnt;

	unsigned long  rsrc_ofs;
	unsigned long  rsrc_cnt;

	unsigned short buf_idx;
	unsigned short buf_cnt;
	unsigned char  buf[256];

	char           compressed;

	unsigned       check1;
	unsigned       check2;
} macc_inp_hcx_t;


static
void macc_hcx_close (macc_inp_t *minp)
{
	mac_close (minp->fp);
	free (minp->ext);
}

static
int hcx_get_hex (const char *src, unsigned cnt, unsigned *val)
{
	unsigned i, d;

	*val = 0;

	for (i = 0; i < cnt; i++) {
		d = src[i];

		if ((d >= '0') && (d <= '9')) {
			d = d - '0';
		}
		else if ((d >= 'A') && (d <= 'F')) {
			d = d - 'A' + 10;
		}
		else if ((d >= 'a') && (d <= 'f')) {
			d = d - 'a' + 10;
		}
		else {
			return (1);
		}

		*val = (*val << 4) | d;
	}

	return (0);
}

static
int hcx_read_line (FILE *fp, char *buf, unsigned max, unsigned *cnt)
{
	int      c;
	unsigned i;

	i = 0;

	while ((c = fgetc (fp)) != EOF) {
		if ((c == 0x0d) || (c == 0x0a)) {
			*cnt = i;
			return (0);
		}

		if (i < max) {
			buf[i++] = c;
		}
	}

	return (i == 0);
}

static
int hcx_decode_line_hex (macc_inp_hcx_t *hcx, const char *src, unsigned cnt)
{
	unsigned i, v;

	cnt = cnt / 2;

	if (cnt > sizeof (hcx->buf)) {
		return (1);
	}

	for (i = 0; i < cnt; i++) {
		if (hcx_get_hex (src, 2, &v)) {
			return (1);
		}

		hcx->buf[i] = v;

		hcx->check2 = (hcx->check2 + v) & 0xff;

		src += 2;
	}

	hcx->buf_idx = 0;
	hcx->buf_cnt = cnt;

	return (0);
}

static
int hcx_decode_line_hcx (macc_inp_hcx_t *hcx, const char *src, unsigned cnt)
{
	unsigned i, n, b, v;

	if (cnt < 1) {
		return (1);
	}

	v = (cnt < 2) ? 32 : src[1];

	if ((src[0] < 32) || (src[0] > 95) || (v < 32) || (v > 95)) {
		return (1);
	}

	v = (v - 32) & 63;
	n = (((src[0] - 32) & 63) << 2) | (v >> 4);

	if (n > sizeof (hcx->buf)) {
		return (1);
	}

	b = 4;

	src += (cnt < 2) ? 1 : 2;
	cnt -= (cnt < 2) ? 1 : 2;

	for (i = 0; i < n; i++) {
		while (b < 8) {
			v <<= 6;
			b += 6;

			if (cnt > 0) {
				if ((*src < 32) || (*src > 95)) {
					return (1);
				}

				v |= (*(src++) - 32) & 63;

				cnt -= 1;
			}
		}

		b -= 8;
		hcx->buf[i] = (v >> b) & 0xff;

		hcx->check2 = (hcx->check2 + hcx->buf[i]) & 0xffff;
		hcx->check2 = ((hcx->check2 << 3) | (hcx->check2 >> 13)) & 0xffff;
	}

	hcx->buf_idx = 0;
	hcx->buf_cnt = n;

	return (0);
}

static
int hcx_decode_line (macc_inp_hcx_t *hcx, const char *src, unsigned cnt)
{
	if (hcx->compressed) {
		return (hcx_decode_line_hcx (hcx, src, cnt));
	}
	else {
		return (hcx_decode_line_hex (hcx, src, cnt));
	}
}

static
int hcx_match_line (const char *buf, unsigned cnt, const char *str)
{
	while ((*str != 0) && (cnt > 0)) {
		if (toupper (*buf) != toupper (*str)) {
			return (0);
		}

		buf += 1;
		str += 1;
		cnt -= 1;
	}

	return (*str == 0);
}

static
int macc_hcx_read_header (macc_inp_t *minp, macfile_t *mf)
{
	int            part;
	unsigned long  type, crtr, mtime;
	unsigned       flags;
	unsigned       cnt;
	char           buf[256];
	macc_inp_hcx_t *hcx;

	hcx = minp->ext;

	hcx->compressed = 0;

	hcx->data_ofs = 0;
	hcx->data_cnt = 0;
	hcx->rsrc_ofs = 0;
	hcx->rsrc_cnt = 0;

	hcx->check1 = 0;
	hcx->check2 = 0;

	part = 0;
	type = 0x3f3f3f3f;
	crtr = 0x3f3f3f3f;
	flags = 0;

	while (hcx_read_line (minp->fp, buf, sizeof (buf), &cnt) == 0) {
		if (cnt == 0) {
			continue;
		}

		if (cnt >= sizeof (buf)) {
			fprintf (stderr, "%s: malformed hcx (long line)\n", arg0);
			return (1);
		}

		if (part == 0) {
			if (hcx_match_line (buf, cnt, "(This file must be converted")) {
				part = 1;
			}
		}
		else if (part == 1) {
			if (hcx_match_line (buf, cnt, "***COMPRESSED")) {
				hcx->compressed = 1;
			}
			else if (hcx_match_line (buf, cnt, "***RESOURCE")) {
				part = 3;
				hcx->rsrc_ofs = ftell (minp->fp);
			}
			else if (hcx_match_line (buf, cnt, "***DATA")) {
				part = 2;
				hcx->data_ofs = ftell (minp->fp);
			}
			else if (hcx_match_line (buf, cnt, "***CRC:")) {
				if (cnt < 11) {
					fprintf (stderr,
						"%s: malformed hcx (crc)\n",
						arg0
					);
					return (1);
				}

				if (hcx_get_hex (buf + 7, 4, &hcx->check1)) {
					fprintf (stderr,
						"%s: malformed hcx (crc)\n",
						arg0
					);
					return (1);
				}
				break;
			}
			else if (hcx_match_line (buf, cnt, "***CHECKSUM:")) {
				if (cnt < 14) {
					fprintf (stderr,
						"%s: malformed hcx (checksum)\n",
						arg0
					);
					return (1);
				}

				if (hcx_get_hex (buf + 12, 2, &hcx->check1)) {
					fprintf (stderr,
						"%s: malformed hcx (checksum)\n",
						arg0
					);
					return (1);
				}
				break;
			}
			else if (hcx_match_line (buf, cnt, "#")) {
				unsigned idx;

				if ((cnt >= 14) && (buf[9] == '$')) {
					type = get_uint32_be (buf, 1);
					crtr = get_uint32_be (buf, 5);
					idx = 10;
				}
				else if ((cnt >= 10) && (buf[5] == '$')) {
					type = get_uint32_be (buf, 1);
					crtr = 0x3f3f3f3f;
					idx = 6;
				}
				else if ((cnt >= 6) && (buf[1] == '$')) {
					type = 0x3f3f3f3f;
					crtr = 0x3f3f3f3f;
					idx = 2;
				}
				else {
					fprintf (stderr,
						"%s: malformed hcx (meta)\n",
						arg0
					);
					return (1);
				}

				if (hcx_get_hex (buf + idx, 4, &flags)) {
					fprintf (stderr,
						"%s: malformed hcx (meta)\n",
						arg0
					);
					return (1);
				}
			}
		}
		else if ((part == 2) || (part == 3)) {
			if (hcx_match_line (buf, cnt, "***END")) {
				part = 1;
			}
			else {
				if (hcx_decode_line (hcx, buf, cnt)) {
					fprintf (stderr,
						"%s: malformed hcx (data line)\n",
						arg0
					);
					return (1);
				}

				if (part == 2) {
					hcx->data_cnt += hcx->buf_cnt;
				}
				else {
					hcx->rsrc_cnt += hcx->buf_cnt;
				}
			}
		}
	}

	if (hcx->check1 != hcx->check2) {
		fprintf (stderr, "%s: crc error (%04X / %04X)\n",
			arg0, hcx->check1, hcx->check2
		);
		return (1);
	}

	mf_init (mf);

	mf_set_name_cstr (mf, hcx->fname);

	mtime = uf_get_mtime_fp (minp->fp);

	mf_set_mtime (mf, mtime);
	mf_set_ctime (mf, mtime);

	mf_set_type (mf, type);
	mf_set_creator (mf, crtr);

	mf_set_data_size (mf, hcx->data_cnt);
	mf_set_rsrc_size (mf, hcx->rsrc_cnt);

	mf_set_finder_flags (mf, flags);

	return (0);
}

static
int macc_hcx_read_fork_start (macc_inp_t *minp, unsigned long ofs)
{
	macc_inp_hcx_t *hcx = minp->ext;

	if (fseek (minp->fp, ofs, SEEK_SET)) {
		return (1);
	}

	hcx->buf_idx = 0;
	hcx->buf_cnt = 0;

	return (0);
}

static
int macc_hcx_read_fork (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	unsigned       n;
	unsigned char  *dst;
	char           src[256];
	macc_inp_hcx_t *hcx;

	hcx = minp->ext;

	dst = buf;

	while (cnt > 0) {
		while (hcx->buf_cnt == 0) {
			if (hcx_read_line (minp->fp, src, sizeof (src), &n)) {
				fprintf (stderr, "%s: malformed hcx (fork)\n",
					arg0
				);
				return (1);
			}

			if (n > 0) {
				if (hcx_decode_line (hcx, src, n)) {
					fprintf (stderr,
						"%s: malformed hcx (fork)\n",
						arg0
					);
					return (1);
				}
			}
		}

		n = (hcx->buf_cnt < cnt) ? hcx->buf_cnt : cnt;

		memcpy (dst, hcx->buf + hcx->buf_idx, n);

		hcx->buf_idx += n;
		hcx->buf_cnt -= n;

		dst += n;
		cnt -= n;
	}

	return (0);
}

static
int macc_hcx_read_fork_done (macc_inp_t *minp)
{
	return (0);
}

static
int macc_hcx_read_data_start (macc_inp_t *minp)
{
	macc_inp_hcx_t *hcx = minp->ext;

	return (macc_hcx_read_fork_start (minp, hcx->data_ofs));
}

static
int macc_hcx_read_rsrc_start (macc_inp_t *minp)
{
	macc_inp_hcx_t *hcx = minp->ext;

	return (macc_hcx_read_fork_start (minp, hcx->rsrc_ofs));
}

static
int macc_hcx_read_comm_start (macc_inp_t *minp)
{
	return (0);
}

static
int macc_hcx_read_comm (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	return (cnt > 0);
}

static
int macc_hcx_read_comm_done (macc_inp_t *minp)
{
	return (0);
}

static
void hcx_set_fname (macc_inp_hcx_t *hcx, const char *str)
{
	unsigned i, n, p1, p2;

	i = 0;
	p1 = 0;
	p2 = 0;

	while (str[i] != 0) {
		if (str[i] == '/') {
			p1 = i + 1;
			p2 = 0;
		}
		else if (str[i] == '.') {
			p2 = i;
		}

		i += 1;
	}

	n = sizeof (hcx->fname);

	for (i = 0; i < n; i++) {
		if (p1 == p2) {
			hcx->fname[i] = 0;
			return;
		}

		if ((hcx->fname[i] = str[p1++]) == 0) {
			return;
		}
	}

	hcx->fname[n - 1] = 0;
}

macc_inp_t *macc_inp_hcx_open (const char *fname)
{
	macc_inp_t     *minp;
	macc_inp_hcx_t *hcx;

	if ((hcx = malloc (sizeof (macc_inp_hcx_t))) == NULL) {
		return (NULL);
	}

	minp = &hcx->inp;
	minp->ext = hcx;

	minp->close = macc_hcx_close;

	minp->read_header = macc_hcx_read_header;

	minp->read_data_start = macc_hcx_read_data_start;
	minp->read_data = macc_hcx_read_fork;
	minp->read_data_done = macc_hcx_read_fork_done;

	minp->read_rsrc_start = macc_hcx_read_rsrc_start;
	minp->read_rsrc = macc_hcx_read_fork;
	minp->read_rsrc_done = macc_hcx_read_fork_done;

	minp->read_comm_start = macc_hcx_read_comm_start;
	minp->read_comm = macc_hcx_read_comm;
	minp->read_comm_done = macc_hcx_read_comm_done;

	minp->fp = mac_open_inp (fname, "r", 1);

	if (minp->fp == NULL) {
		free (hcx);
		return (NULL);
	}

	hcx_set_fname (hcx, fname);

	return (minp);
}
