/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_inp_hqx.c                                      *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/file.h>
#include <lib/hqx.h>
#include <lib/macfile.h>
#include <lib/unixfile.h>


typedef struct {
	macc_inp_t    inp;

	hqx_t         hqx;

	crc16_t       crc;
} macc_inp_hqx_t;


static
void macc_hqx_close (macc_inp_t *minp)
{
	mac_close (minp->fp);
	free (minp->ext);
}

static
int macc_hqx_read_header (macc_inp_t *minp, macfile_t *mf)
{
	unsigned long  mtime;
	macc_inp_hqx_t *hqx;

	hqx = minp->ext;

	if (hqx_init_read (&hqx->hqx, minp->fp)) {
		fprintf (stderr, "%s: hqx header read error\n", arg0);
		return (1);
	}

	if (hqx_read_header (&hqx->hqx)) {
		fprintf (stderr, "%s: hqx header error\n", arg0);
		return (1);
	}

	if (hqx_check_crc (&hqx->hqx)) {
		fprintf (stderr, "%s: hqx header bad crc\n", arg0);
		return (1);
	}

	mf_init (mf);

	mtime = uf_get_mtime_fp (minp->fp);

	mf_set_mtime (mf, mtime);
	mf_set_ctime (mf, mtime);

	hqx_get_macfile (&hqx->hqx, mf);

	mf_set_finder_flags (mf, mf_get_finder_flags (mf) & ~MF_FINDER_INITED);

	return (0);
}

static
int macc_hqx_read_block_start (macc_inp_t *minp)
{
	macc_inp_hqx_t *hqx;

	hqx = minp->ext;

	crc16_reset (&hqx->crc, 0);

	return (0);
}

static
int macc_hqx_read_block (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	macc_inp_hqx_t *hqx;

	hqx = minp->ext;

	if (hqx_read (&hqx->hqx, buf, cnt)) {
		fprintf (stderr, "%s: hqx read error\n", arg0);
		return (1);
	}

	crc16_calc (&hqx->crc, buf, cnt);

	return (0);
}

static
int macc_hqx_read_block_done (macc_inp_t *minp)
{
	unsigned char  crc_buf[2];
	macc_inp_hqx_t *hqx;

	hqx = minp->ext;

	crc16_done (&hqx->crc);

	if (hqx_read (&hqx->hqx, crc_buf, 2)) {
		fprintf (stderr, "%s: hqx read error\n", arg0);
		return (1);
	}

	if (get_uint16_be (crc_buf, 0) != crc16_get_crc (&hqx->crc)) {
		fprintf (stderr, "%s: hqx crc error\n", arg0);
		return (1);
	}

	return (0);
}

static
int macc_hqx_read_comm_start (macc_inp_t *minp)
{
	return (0);
}

static
int macc_hqx_read_comm (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	return (cnt > 0);
}

static
int macc_hqx_read_comm_done (macc_inp_t *minp)
{
	return (0);
}

macc_inp_t *macc_inp_hqx_open (const char *fname)
{
	macc_inp_t     *minp;
	macc_inp_hqx_t *hqx;

	hqx = malloc (sizeof (macc_inp_hqx_t));

	if (hqx == NULL) {
		return (NULL);
	}

	minp = &hqx->inp;

	minp->ext = hqx;

	minp->close = macc_hqx_close;

	minp->read_header = macc_hqx_read_header;

	minp->read_data_start = macc_hqx_read_block_start;
	minp->read_data = macc_hqx_read_block;
	minp->read_data_done = macc_hqx_read_block_done;

	minp->read_rsrc_start = macc_hqx_read_block_start;
	minp->read_rsrc = macc_hqx_read_block;
	minp->read_rsrc_done = macc_hqx_read_block_done;

	minp->read_comm_start = macc_hqx_read_comm_start;
	minp->read_comm = macc_hqx_read_comm;
	minp->read_comm_done = macc_hqx_read_comm_done;

	minp->fp = mac_open_inp (fname, "r", 1);

	if (minp->fp == NULL) {
		free (hqx);
		return (NULL);
	}

	crc16_init (&hqx->crc, CRC_CRC16_CCITT, 0);

	return (minp);
}
