/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_out_hcx.c                                      *
 * Created:     2013-05-27 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2013 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/macfile.h>


typedef struct {
	macc_out_t    out;

	char          compressed;

	unsigned      cur_fork;

	unsigned      buf_cnt;
	unsigned char buf[64];

	unsigned long data_idx;
	unsigned long data_cnt;
	unsigned long rsrc_cnt;
	unsigned char *data;

	unsigned      check;
} macc_out_hcx_t;


static
void macc_hcx_close (macc_out_t *mout)
{
	macc_out_hcx_t *hcx = mout->ext;

	if (hcx->compressed) {
		fprintf (mout->fp, "***CRC:%04X\n", hcx->check & 0xffff);
	}
	else {
		fprintf (mout->fp, "***CHECKSUM:%02X\n", hcx->check & 0xff);
	}

	free (hcx->data);

	mac_close (mout->fp);

	free (mout->ext);
}

static
void hcx_write_oscode (macc_out_hcx_t *hcx, unsigned long val)
{
	int      c;
	unsigned i;

	for (i = 0; i < 4; i++) {
		c = (val >> 24) & 0xff;

		if ((c < 32) || (c > 126)) {
			c = '-';
		}

		fputc (c, hcx->out.fp);

		val <<= 8;
	}
}

static
int macc_hcx_write_header (macc_out_t *mout, macfile_t *mf)
{
	macc_out_hcx_t *hcx = mout->ext;

	fputs ("(This file must be converted with BinHex.Hex)\n", mout->fp);

	fputc ('#', mout->fp);
	hcx_write_oscode (hcx, mf_get_type (mf));
	hcx_write_oscode (hcx, mf_get_creator (mf));
	fprintf (mout->fp, "$%04X\n", mf_get_finder_flags (mf) & 0xffff);

	if (hcx->compressed) {
		fputs ("***COMPRESSED\n", mout->fp);
	}

	hcx->cur_fork = 0;

	hcx->data_cnt = mf_get_data_size (mf);
	hcx->rsrc_cnt = mf_get_rsrc_size (mf);
	hcx->data = NULL;

	hcx->check = 0;

	return (0);
}

static
int macc_hcx_write_fork_start (macc_out_t *mout, int data)
{
	macc_out_hcx_t *hcx = mout->ext;

	hcx->buf_cnt = 0;

	hcx->cur_fork = 0;

	if (data) {
		if (hcx->data_cnt == 0) {
			return (0);
		}

		hcx->cur_fork = 1;

		if (hcx->rsrc_cnt == 0) {
			fputs ("***DATA FORK\n", mout->fp);
		}
		else {
			hcx->data_idx = 0;

			if ((hcx->data = malloc (hcx->data_cnt)) == NULL) {
				fprintf (stderr, "%s: malloc failed\n", arg0);
				return (1);
			}
		}
	}
	else if (hcx->rsrc_cnt > 0) {
		hcx->cur_fork = 2;
		fputs ("***RESOURCE FORK\n", mout->fp);
	}

	return (0);
}

static
int macc_hcx_write_data_start (macc_out_t *mout)
{
	return (macc_hcx_write_fork_start (mout, 1));
}

static
int macc_hcx_write_rsrc_start (macc_out_t *mout)
{
	return (macc_hcx_write_fork_start (mout, 0));
}

static
void hcx_calc_check (macc_out_hcx_t *hcx, unsigned val)
{
	if (hcx->compressed) {
		hcx->check = (hcx->check + val) & 0xffff;
		hcx->check = ((hcx->check << 3) | (hcx->check >> 13)) & 0xffff;
	}
	else {
		hcx->check = (hcx->check + val) & 0xff;
	}
}

static
void hcx_write_line (macc_out_hcx_t *hcx)
{
	unsigned i, b, v;

	if (hcx->compressed) {
		v = hcx->buf_cnt;
		b = 8;

		for (i = 0; i < hcx->buf_cnt; i++) {
			v = (v << 8) | hcx->buf[i];
			b += 8;

			while (b >= 6) {
				b -= 6;
				fputc (((v >> b) & 63) + 32, hcx->out.fp);
			}
		}

		while ((b > 0) && (b < 6)) {
			v = v << 8;
			b += 2;
			fputc (((v >> b) & 63) + 32, hcx->out.fp);
		}
	}
	else {
		for (i = 0; i < hcx->buf_cnt; i++) {
			fprintf (hcx->out.fp, "%02X", hcx->buf[i]);
		}
	}

	fputc ('\n', hcx->out.fp);

	hcx->buf_cnt = 0;
}

static
int hcx_write_fork (macc_out_hcx_t *hcx, const void *buf, unsigned long cnt)
{
	const unsigned char *src;

	src = buf;

	while (cnt > 0) {
		hcx_calc_check (hcx, *src);

		hcx->buf[hcx->buf_cnt++] = *(src++);
		cnt -= 1;

		if (hcx->buf_cnt >= 32) {
			hcx_write_line (hcx);
		}
	}

	return (0);
}

static
int macc_hcx_write_fork (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	unsigned long  max;
	macc_out_hcx_t *hcx;

	hcx = mout->ext;

	if (hcx->cur_fork == 0) {
		return (1);
	}

	if ((hcx->cur_fork == 1) && (hcx->data != NULL)) {
		max = hcx->data_cnt - hcx->data_idx;

		if (cnt > max) {
			return (1);
		}

		memcpy (hcx->data + hcx->data_idx, buf, cnt);

		hcx->data_idx += cnt;
	}
	else {
		if (hcx_write_fork (hcx, buf, cnt)) {
			return (1);
		}
	}


	return (0);
}

static
int hcx_write_fork_done (macc_out_hcx_t *hcx)
{
	if (hcx->buf_cnt > 0) {
		hcx_write_line (hcx);
	}

	fputs ("***END OF DATA\n", hcx->out.fp);

	return (0);
}

static
int macc_hcx_write_fork_done (macc_out_t *mout)
{
	macc_out_hcx_t *hcx = mout->ext;

	if (hcx->cur_fork == 1) {
		if (hcx->data == NULL) {
			if (hcx_write_fork_done (hcx)) {
				return (1);
			}
		}
	}
	else if (hcx->cur_fork == 2) {
		if (hcx_write_fork_done (hcx)) {
			return (1);
		}

		if (hcx->data != NULL) {
			fputs ("***DATA FORK\n", mout->fp);

			if (hcx_write_fork (hcx, hcx->data, hcx->data_cnt)) {
				return (1);
			}

			free (hcx->data);
			hcx->data = NULL;

			if (hcx_write_fork_done (hcx)) {
				return (1);
			}
		}
	}

	hcx->cur_fork = 0;

	return (0);
}

static
int macc_hcx_write_comm_start (macc_out_t *mout)
{
	return (0);
}

static
int macc_hcx_write_comm (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	return (0);
}

static
int macc_hcx_write_comm_done (macc_out_t *mout)
{
	return (0);
}

macc_out_t *macc_out_hcx_open (const char *fname, int compressed)
{
	macc_out_t     *mout;
	macc_out_hcx_t *hcx;

	if ((hcx = malloc (sizeof (macc_out_hcx_t))) == NULL) {
		return (NULL);
	}

	mout = &hcx->out;
	mout->ext = hcx;

	mout->close = macc_hcx_close;

	mout->write_header = macc_hcx_write_header;

	mout->write_data_start = macc_hcx_write_data_start;
	mout->write_data = macc_hcx_write_fork;
	mout->write_data_done = macc_hcx_write_fork_done;

	mout->write_rsrc_start = macc_hcx_write_rsrc_start;
	mout->write_rsrc = macc_hcx_write_fork;
	mout->write_rsrc_done = macc_hcx_write_fork_done;

	mout->write_comm_start = macc_hcx_write_comm_start;
	mout->write_comm = macc_hcx_write_comm;
	mout->write_comm_done = macc_hcx_write_comm_done;

	if ((mout->fp = mac_open_out (fname, "w", 1)) == NULL) {
		free (hcx);
		return (NULL);

	}

	hcx->compressed = (compressed != 0);

	return (mout);
}
