/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_out_mb.c                                       *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/macbinary.h>
#include <lib/macfile.h>


typedef struct {
	macc_out_t    out;

	unsigned char mb[128];

	unsigned long data_size;
	unsigned long rsrc_size;
	unsigned long comm_size;
} macc_out_mb_t;


static
void macc_mb_close (macc_out_t *mout)
{
	mac_close (mout->fp);
	free (mout->ext);
}

static
int macc_mb_write_header (macc_out_t *mout, macfile_t *mf)
{
	macc_out_mb_t *mb;

	mb = mout->ext;

	mb_init (mb->mb);

	if (mf_get_file_flags (mf) & MF_FILE_ISDIR) {
		mb_set_id (mb->mb, 1);
	}
	else {
		mb_set_id (mb->mb, 0);
	}

	mb_set_macfile (mb->mb, mf);

	mb_set_crc (mb->mb, mb_calc_crc (mb->mb));

	if (mb_write (mb->mb, mout->fp)) {
		fprintf (stderr, "%s: mb header write error\n", arg0);
		return (1);
	}

	mb->data_size = mf_get_data_size (mf);
	mb->rsrc_size = mf_get_rsrc_size (mf);
	mb->comm_size = mf_get_comment_size (mf);

	return (0);
}

static
int macc_mb_write_block_start (macc_out_t *mout)
{
	return (0);
}

static
int macc_mb_write_block (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	if (fwrite (buf, 1, cnt, mout->fp) != cnt) {
		fprintf (stderr, "%s: mb write error\n", arg0);
		return (1);
	}

	return (0);
}

static
int macc_mb_write_block_done (macc_out_t *mout, unsigned long size)
{
	size = (128 - (size & 127)) & 127;

	while (size > 0) {
		fputc (0, mout->fp);
		size -= 1;
	}

	return (0);
}

static
int macc_mb_write_data_done (macc_out_t *mout)
{
	macc_out_mb_t *mb;

	mb = mout->ext;

	return (macc_mb_write_block_done (mout, mb->data_size));
}

static
int macc_mb_write_rsrc_done (macc_out_t *mout)
{
	macc_out_mb_t *mb;

	mb = mout->ext;

	return (macc_mb_write_block_done (mout, mb->rsrc_size));
}

static
int macc_mb_write_comm_done (macc_out_t *mout)
{
	macc_out_mb_t *mb;

	mb = mout->ext;

	return (macc_mb_write_block_done (mout, mb->comm_size));
}

macc_out_t *macc_out_mb_open (const char *fname)
{
	macc_out_t    *mout;
	macc_out_mb_t *mb;

	mb = malloc (sizeof (macc_out_mb_t));

	if (mb == NULL) {
		return (NULL);
	}

	mout = &mb->out;

	mout->ext = mb;

	mout->close = macc_mb_close;

	mout->write_header = macc_mb_write_header;

	mout->write_data_start = macc_mb_write_block_start;
	mout->write_data = macc_mb_write_block;
	mout->write_data_done = macc_mb_write_data_done;

	mout->write_rsrc_start = macc_mb_write_block_start;
	mout->write_rsrc = macc_mb_write_block;
	mout->write_rsrc_done = macc_mb_write_rsrc_done;

	mout->write_comm_start = macc_mb_write_block_start;
	mout->write_comm = macc_mb_write_block;
	mout->write_comm_done = macc_mb_write_comm_done;

	mout->fp = mac_open_out (fname, "wb", 1);

	if (mout->fp == NULL) {
		free (mb);
		return (NULL);
	}

	return (mout);
}
