/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macrsc/macrsc.c                                          *
 * Created:     2005-02-17 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2020 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "macrsc.h"
#include "parse.h"

#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/rsrc.h>
#include <lib/strings.h>


#define OP_LIST    0
#define OP_WRITE   1
#define OP_DECODE  2
#define OP_ENCODE  3


static char          *argv0 = NULL;

static unsigned      par_op = OP_LIST;

static int           par_sort = 0;
static int           par_sort_revtyp = 0;
static int           par_sort_revrsc = 0;

static int           par_purge = 0;
static int           par_binary = 0;

static int           par_rsrc_type_ok = 0;
static unsigned long par_rsrc_type = 0;
static int           par_rsrc_id_ok = 0;
static unsigned      par_rsrc_id = 0;
static int           par_rsrc_attr_ok = 0;
static unsigned      par_rsrc_attr = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'A', 1, "attributes", "int", "Set resource fork attributes [0]" },
	{ 'b', 0, "binary", NULL, "Dump resources as binary [no]" },
	{ 'd', 0, "decode", NULL, "Dump resources" },
	{ 'e', 0, "encode", NULL, "Undump reources" },
	{ 'i', 1, "input", "string", "Set the input file name" },
	{ 'l', 0, "list", NULL, "List resources [default]" },
	{ 'n', 1, "id", "int", "Set the resource id [all]" },
	{ 'o', 1, "output", "string", "Set the output file name" },
	{ 'p', 0, "remove-duplicates", NULL, "Remove duplicate resources [no]" },
	{ 'r', 0, "reverse-types", NULL, "Sort types in reverse order [no]" },
	{ 'R', 0, "reverse-resources", NULL, "Sort resources in reverse order [no]" },
	{ 's', 0, "sort", NULL, "Sort resources [no]" },
	{ 't', 1, "type", "string", "Set the resource type [all]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'w', 0, "write", NULL, "Copy resources" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"macrsc: encode and decode macintosh resource files",
		"usage: macrsc [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"macrsc version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2005-2020 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int check_type (unsigned long type)
{
	if (par_rsrc_type_ok == 0) {
		return (1);
	}

	if (par_rsrc_type == type) {
		return (1);
	}

	return (0);
}

static
int check_id (unsigned id)
{
	if (par_rsrc_id_ok == 0) {
		return (1);
	}

	if (par_rsrc_id == id) {
		return (1);
	}

	return (0);
}

static
int macrsc_open (const char *inp, FILE **finp, const char *minp,
	const char *out, FILE **fout, const char *mout)
{
	*finp = mac_open_inp (inp, minp, 1);

	if (*finp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			argv0, inp
		);
		return (1);
	}

	*fout = mac_open_out (out, mout, 1);

	if (*fout == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			argv0, inp
		);

		mac_close (*finp);

		return (1);
	}

	return (0);
}

static
int macrsc_rf_write (rsrc_file_t *rf, FILE *fp)
{
	if (par_sort) {
		rf_sort (rf, par_sort_revtyp, par_sort_revrsc);
	}

	if (par_rsrc_attr_ok) {
		rf->attr = par_rsrc_attr;
	}

	if (rf_write (rf, fp)) {
		fprintf (stderr, "%s: resource file write error\n", argv0);
		return (1);
	}

	return (0);
}

static
int macrsc_list_fp (FILE *inp, FILE *fp)
{
	unsigned    i, j;
	rsrc_file_t *rf;
	rsrc_type_t *rt;
	rsrc_t      *r;
	char        str[256];

	rf = rf_read (inp);
	if (rf == NULL) {
		fprintf (stderr, "%s: resource file read error\n", argv0);
		return (1);
	}

	if (par_purge) {
		rf_remove_duplicates (rf);
	}

	if (par_sort) {
		rf_sort (rf, par_sort_revtyp, par_sort_revrsc);
	}

	fprintf (fp, "file\t0x%04x\t%u\t%lu\n",
		rf->attr, rf->type_cnt, rf_get_resource_count (rf)
	);

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		if (check_type (rt->type) == 0) {
			continue;
		}

		str_set_fourcc (str, 256, rt->type);

		fprintf (fp, "type\t\"%s\"\t%u\n", str, rt->rsrc_cnt);

		for (j = 0; j < rt->rsrc_cnt; j++) {
			r = rt->rsrc[j];

			if (check_id (r->id) == 0) {
				continue;
			}

			fprintf (fp, "rsrc\t\"%s\"\t0x%04x\t0x%02x\t%lu\t%lu",
				str, r->id, r->attr, r->offset, r->size
			);

			if (r->name_cnt > 0) {
				char *tmp;

				tmp = mac_string_decode (r->name, r->name_cnt, "\"");
				if (tmp != NULL) {
					fprintf (fp, "\t\"%s\"", tmp);
					free (tmp);
				}
			}

			fputs ("\n", fp);
		}
	}

	if (rf->dupcnt > 0) {
		fprintf (fp, "warning: %lu duplicate resources\n", rf->dupcnt);
	}

	rf_del (rf);

	return (0);
}

static
int macrsc_list (const char *inp, const char *out)
{
	int  r;
	FILE *finp, *fout;

	if (macrsc_open (inp, &finp, "rb", out, &fout, "w")) {
		return (1);
	}

	r = macrsc_list_fp (finp, fout);

	mac_close (fout);
	mac_close (finp);

	return (r);
}


static
int macrsc_write_fp (FILE *inp, FILE *out)
{
	int         r;
	rsrc_file_t *rf;

	rf = rf_read (inp);
	if (rf == NULL) {
		fprintf (stderr, "%s: resource file read error\n", argv0);
		return (1);
	}

	r = macrsc_rf_write (rf, out);

	rf_del (rf);

	return (r);
}

static
int macrsc_write (const char *inp, const char *out)
{
	int  r;
	FILE *finp, *fout;

	if (macrsc_open (inp, &finp, "rb", out, &fout, "wb")) {
		return (1);
	}

	r = macrsc_write_fp (finp, fout);

	mac_close (fout);
	mac_close (finp);

	return (r);
}


static
int macrsc_decode_rsc (FILE *fp, rsrc_t *r)
{
	unsigned long i;
	unsigned char *data;
	unsigned long size;
	char          type[16];

	size = rsc_get_size (r);

	if (par_binary) {
		if (size == 0) {
			return (0);
		}

		data = rsc_get_data (r);
		if (data == NULL) {
			fprintf (stderr, "%s: resource data read error\n",
				argv0
			);
			return (1);
		}

		if (fwrite (data, size, 1, fp) != 1) {
			return (1);
		}

		return (0);
	}

	str_set_fourcc (type, 16, r->type);

	fprintf (fp, "resource \"%s\" 0x%04x 0x%02x",
		type, r->id, r->attr
	);

	if (r->name_cnt > 0) {
		char *tmp;

		tmp = mac_string_decode (r->name, r->name_cnt, "\"");
		if (tmp == NULL) {
			return (1);
		}

		fprintf (fp, " \"%s\"", tmp);

		free (tmp);
	}

	fputs ("\n", fp);

	if (size > 0) {
		data = rsc_get_data (r);
		if (data == NULL) {
			fprintf (stderr, "%s: resource data read error\n",
				argv0
			);
			return (1);
		}

		for (i = 0; i < size; i++) {
			fputs (((i & 15) == 0) ? "\t" : " ", fp);
			fprintf (fp, "%02X", data[i]);
			if ((i & 15) == 15) {
				fputs ("\n", fp);
			}
		}

		if (size & 15) {
			fputs ("\n", fp);
		}
	}

	fputs ("end\n", fp);

	return (0);
}

static
int macrsc_decode_fp (FILE *inp, FILE *out)
{
	unsigned      i, j;
	rsrc_file_t   *rf;
	rsrc_type_t   *rt;
	rsrc_t        *rs;

	rf = rf_read (inp);
	if (rf == NULL) {
		fprintf (stderr, "%s: resource file read error\n", argv0);
		return (1);
	}

	if (par_purge) {
		rf_remove_duplicates (rf);
	}

	if (par_sort) {
		rf_sort (rf, par_sort_revtyp, par_sort_revrsc);
	}

	if (par_binary == 0) {
		if (rf->dupcnt > 0) {
			fprintf (out,
				"# warning: %lu duplicate resources\n\n",
				rf->dupcnt
			);
		}

		fprintf (out, "file 0x%04x\n", rf->attr);
	}

	for (i = 0; i < rf->type_cnt; i++) {
		rt = rf->type[i];

		if (check_type (rt->type) == 0) {
			continue;
		}

		for (j = 0; j < rt->rsrc_cnt; j++) {
			rs = rt->rsrc[j];

			if (check_id (rs->id) == 0) {
				continue;
			}

			if (par_binary == 0) {
				fputs ("\n", out);
			}

			if (macrsc_decode_rsc (out, rs)) {
				rf_del (rf);
				return (1);
			}
		}
	}

	rf_del (rf);

	return (0);
}

static
int macrsc_decode (const char *inp, const char *out)
{
	int  r;
	FILE *finp, *fout;

	if (macrsc_open (inp, &finp, "rb", out, &fout, "w")) {
		return (1);
	}

	r = macrsc_decode_fp (finp, fout);

	mac_close (fout);
	mac_close (finp);

	return (r);
}


static
int macrsc_read_line (FILE *fp, char *buf, unsigned max)
{
	unsigned i;
	int      c;

	i = 0;

	while (i < max) {
		c = fgetc (fp);

		if (c == EOF) {
			buf[i] = 0;
			return (i == 0);
		}

		if ((c == 0x0a) || (c == 0x0d)) {
			buf[i] = 0;
			return (0);
		}

		buf[i++] = c;
	}

	return (1);
}

static
int macrsc_data_add_byte (unsigned char **data, unsigned long *cnt,
	unsigned long *max, unsigned char val)
{
	unsigned long newmax;
	unsigned char *tmp;

	if (*cnt >= *max) {
		newmax = *cnt;

		while (newmax <= *max) {
			newmax = (newmax == 0) ? 16 : (2 * newmax);
		}

		tmp = realloc (*data, newmax);
		if (tmp == NULL) {
			return (1);
		}

		*data = tmp;
		*max = newmax;
	}

	(*data)[*cnt] = val;
	*cnt += 1;

	return (0);
}

static
int macrsc_data_add_string (unsigned char **data, unsigned long *cnt,
	unsigned long *max, const char *str)
{
	unsigned      i, n;
	unsigned char *tmp;
	unsigned      len;

	tmp = mac_string_encode (str, &n);
	if (tmp == NULL) {
		return (1);
	}

	len = 0;

	i = 0;
	while (i < n) {
		if (((i + 1) < n) && (tmp[i] == '\\') && (tmp[i + 1] == 'p')) {
			i += 2;
		}
		else {
			i += 1;
			len += 1;
		}
	}

	i = 0;
	while (i < n) {
		if (((i + 1) < n) && (tmp[i] == '\\') && (tmp[i + 1] == 'p')) {
			if (macrsc_data_add_byte (data, cnt, max, len)) {
				return (1);
			}
			i += 2;
		}
		else {
			if (macrsc_data_add_byte (data, cnt, max, tmp[i])) {
				return (1);
			}
			i += 1;
		}
	}

	free (tmp);

	return (0);
}

static
rsrc_t *macrsc_encode_rsc (FILE *fp, const char *str)
{
	int           err;
	rsrc_t        *r;
	unsigned char val;
	unsigned long type;
	unsigned long id;
	unsigned long attr;
	char          buf[768];
	const char    *tmp;
	unsigned char *data;
	unsigned long cnt, max;

	if (match_text (&str, "resource") == 0) {
		return (NULL);
	}

	if (match_string (&str, buf, 256) == 0) {
		return (NULL);
	}

	if (str_get_fourcc (buf, &type)) {
		return (NULL);
	}

	if (match_uint32 (&str, &id) == 0) {
		return (NULL);
	}

	r = rsc_new (type, id & 0xffff);
	if (r == NULL) {
		return (NULL);
	}

	if (match_uint32 (&str, &attr) == 0) {
		attr = 0;
	}

	r->attr = attr;

	if (match_string (&str, buf, 768)) {
		unsigned char *tmp;

		tmp = mac_pstring_encode (buf);
		if (tmp == NULL) {
			rsc_del (r);
			return (NULL);
		}

		rsc_set_name (r, tmp + 1, tmp[0]);

		free (tmp);
	}

	if (match_eol (&str) == 0) {
		rsc_del (r);
		return (NULL);
	}

	data = NULL;
	cnt = 0;
	max = 0;

	while (1) {
		if (macrsc_read_line (fp, buf, 768)) {
			break;
		}

		tmp = buf;

		if (match_text (&tmp, "end")) {
			if (match_eol (&tmp) == 0) {
				break;
			}

			rsc_set_data (r, data, cnt);

			return (r);
		}

		err = 1;

		while (1) {
			char str[768];

			if (match_uhex8 (&tmp, &val)) {
				if (macrsc_data_add_byte (&data, &cnt, &max, val)) {
					break;
				}
			}
			else if (match_string (&tmp, str, 768)) {
				if (macrsc_data_add_string (&data, &cnt, &max, str)) {
					break;
				}
			}
			else {
				err = 0;
				break;
			}
		}

		if (err) {
			break;
		}

		if (match_eol (&tmp) == 0) {
			break;
		}
	}

	free (data);
	rsc_del (r);

	return (NULL);
}

static
int macrsc_encode_rec (FILE *inp, rsrc_file_t *rf)
{
	rsrc_t        *r;
	char          buf[256];
	const char    *tmp;
	unsigned long type;
	unsigned      id;

	type = 0;
	id = 0;

	while (1) {
		if (macrsc_read_line (inp, buf, 256)) {
			return (0);
		}

		tmp = buf;

		if (match_text (&tmp, "resource")) {
			r = macrsc_encode_rsc (inp, buf);

			if (r == NULL) {
				fprintf (stderr,
					"%s: parse error after %08lX %04X\n",
					argv0, type, id
				);
			}
			else {
				type = r->type;
				id = r->id;
				if (rf_add_resource (rf, r) == NULL) {
					fprintf (stderr,
						"%s: adding resource %08lX %04X failed\n",
						argv0, r->type, r->id
					);
				}
			}
		}
		else if (match_text (&tmp, "file")) {
			unsigned long val;

			if (match_uint32 (&tmp, &val)) {
				rf->attr = val & 0xffff;
			}

			if (match_eol (&tmp) == 0) {
				break;
			}
		}
		else if (match_text (&tmp, "include")) {
			FILE *fp;

			if (match_string (&tmp, buf, 256) == 0) {
				break;
			}

			fp = mac_open_inp (buf, "r", 0);
			if (fp == NULL) {
				fprintf (stderr,
					"%s: can't open include file (%s)\n",
					argv0, buf
				);
				return (1);
			}

			if (macrsc_encode_rec (fp, rf)) {
				mac_close (fp);
				return (1);
			}

			mac_close (fp);
		}
		else if (match_eol (&tmp) == 0) {
			break;
		}
	}

	fprintf (stderr, "%s: parse error after %08lX %04X\n", argv0, type, id);

	return (1);
}

static
int macrsc_encode_fp (FILE *inp, FILE *out)
{
	rsrc_file_t *rf;

	rf = rf_new();

	if (macrsc_encode_rec (inp, rf)) {
		rf_del (rf);
		return (1);
	}

	if (par_purge) {
		rf_remove_duplicates (rf);
	}

	if (par_sort) {
		rf_sort (rf, par_sort_revtyp, par_sort_revrsc);
	}

	macrsc_rf_write (rf, out);

	rf_del (rf);

	return (0);
}

static
int macrsc_encode (const char *inp, const char *out)
{
	int  r;
	FILE *finp, *fout;

	if (macrsc_open (inp, &finp, "r", out, &fout, "wb")) {
		return (1);
	}

	r = macrsc_encode_fp (finp, fout);

	mac_close (fout);
	mac_close (finp);

	return (r);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'A':
			par_rsrc_attr = (unsigned) strtoul (optarg[0], NULL, 0);
			par_rsrc_attr &= 0xffff;
			par_rsrc_attr_ok = 1;
			break;

		case 'b':
			par_binary = 1;
			break;

		case 'd':
			par_op = OP_DECODE;
			break;

		case 'e':
			par_op = OP_ENCODE;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'l':
			par_op = OP_LIST;
			break;

		case 'n':
			par_rsrc_id = (unsigned) strtoul (optarg[0], NULL, 0);
			par_rsrc_id &= 0xffff;
			par_rsrc_id_ok = 1;
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_purge = 1;
			break;

		case 'r':
			par_sort_revtyp = 1;
			break;

		case 'R':
			par_sort_revrsc = 1;
			break;

		case 's':
			par_sort = 1;
			break;

		case 't':
			if (str_get_fourcc (optarg[0], &par_rsrc_type)) {
				fprintf (stderr,
					"%s: bad resource type (%s)\n",
					argv0, optarg[0]
				);
				return (1);
			}
			par_rsrc_type_ok = 1;
			break;

		case 'w':
			par_op = OP_WRITE;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr,
					"%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	switch (par_op) {
	case OP_LIST:
		r = macrsc_list (finp, fout);
		break;

	case OP_WRITE:
		r = macrsc_write (finp, fout);
		break;

	case OP_DECODE:
		r = macrsc_decode (finp, fout);
		break;

	case OP_ENCODE:
		r = macrsc_encode (finp, fout);
		break;

	default:
		r = 1;
	}

	return (r);
}
