/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mactobdf/mactobdf.c                                      *
 * Created:     2001-01-06 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "mactobdf.h"

#include <lib/buffer.h>
#include <lib/getopt.h>


static char *argv0 = NULL;

static FontInfo finfo = {
	"testfont",
	12
};


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'f', 1, "font-name", "string", "Set the font name" },
	{ 'i', 1, "input", "string", "Set the input file name" },
	{ 'o', 1, "output", "string", "Set the output file name" },
	{ 's', 1, "font-size", "int", "Set the font size" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mactobdf: convert macintosh font resources to BDF",
		"usage: mactobdf [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mactobdf version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2001-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

int get_int8 (const void *buf, unsigned i)
{
	unsigned            val;
	int                 ret;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	val = tmp[i];

	if (val & 0x80) {
		ret = -(int)((~val + 1) & 0x7f);
	}
	else {
		ret = (int) val;
	}

	return (ret);
}

int get_int16 (const void *buf, unsigned i)
{
	unsigned            val;
	int                 ret;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];

	if (val & 0x8000) {
		ret = -(int)((~val + 1) & 0x7fff);
	}
	else {
		ret = (int) val;
	}

	return (ret);
}


void prt_fontrec (FILE *file, FontRec *fr)
{
	fprintf (file, "fontType:    %x\n", fr->fontType & 0xffff);
	fprintf (file, "firstChar:   %d\n", fr->firstChar);
	fprintf (file, "lastChar:    %d\n", fr->lastChar);
	fprintf (file, "widMax:      %d\n", fr->widMax);
	fprintf (file, "kernMax:     %d\n", fr->kernMax);
	fprintf (file, "nDescent:    %d\n", fr->nDescent);
	fprintf (file, "fRectWidth:  %d\n", fr->fRectWidth);
	fprintf (file, "fRectHeight: %d\n", fr->fRectHeight);
	fprintf (file, "owTLoc:      %d\n", fr->owTLoc);
	fprintf (file, "ascent:      %d\n", fr->ascent);
	fprintf (file, "descent:     %d\n", fr->descent);
	fprintf (file, "leading:     %d\n", fr->leading);
	fprintf (file, "rowWords:    %d\n", fr->rowWords);
	fflush (file);
}

void prt_chartab (FILE *file, CharTab *ct)
{
	unsigned i;

	fprintf (file, "Characters: %u\n", ct->chrcnt);

	for (i = 0; i < ct->chrcnt; i++) {
		fprintf (file, "chr %u: bitx: %u  bitw: %u  ofs: %d  w: %d  p: %d\n",
			i, ct->chr[i].bi_x, ct->chr[i].bi_w,
			ct->chr[i].ofs, ct->chr[i].width,
			(int)ct->chr[i].present
		);
	}

	fflush (file);
}

BitImage *bi_new (unsigned w, unsigned h, unsigned wcnt)
{
	unsigned i;
	BitImage *bi;

	bi = malloc (sizeof (BitImage));
	if (bi == NULL) {
		return (NULL);
	}

	bi->w = w;
	bi->h = h;
	bi->wcnt = wcnt;

	bi->row[0] = malloc (2 * wcnt * h);
	if (bi->row[0] == NULL) {
		free (bi);
		return (NULL);
	}

	for (i = 1; i < h; i++) {
		bi->row[i] = bi->row[i - 1] + 2 * wcnt;
	}

	return (bi);
}

void bi_del (BitImage *bi)
{
	if (bi != NULL) {
		free (bi->row[0]);
		free (bi);
	}
}

int get_fontrec (FILE *fp, FontRec *fr)
{
	unsigned char buf[32];

	if (fread (buf, 1, 26, fp) != 26) {
		return (1);
	}

	fr->fontType = get_int16 (buf, 0);
	fr->firstChar = get_int16 (buf, 2);
	fr->lastChar = get_int16 (buf, 4);
	fr->widMax = get_int16 (buf, 6);
	fr->kernMax = get_int16 (buf, 8);
	fr->nDescent = get_int16 (buf, 10);
	fr->fRectWidth = get_int16 (buf, 12);
	fr->fRectHeight = get_int16 (buf, 14);
	fr->owTLoc = 2 * get_int16 (buf, 16) + 16;
	fr->ascent = get_int16 (buf, 18);
	fr->descent = get_int16 (buf, 20);
	fr->leading = get_int16 (buf, 22);
	fr->rowWords = get_int16 (buf, 24);

	return (0);
}

int get_bitimage (FILE *fp, BitImage *bi)
{
	unsigned i;

	for (i = 0; i < bi->h; i++) {
		if (fread (bi->row[i], 2, bi->wcnt, fp) != bi->wcnt) {
			return (1);
		}
	}

	return (0);
}

int get_loctab (FILE *file, CharTab *ct)
{
	unsigned      i;
	unsigned      pos0, pos1;
	unsigned char buf[2];

	if (fread (buf, 2, 1, file) != 1) {
		return (1);
	}

	pos0 = get_uint16_be (buf, 0);

	for (i = 0; i < ct->chrcnt; i++) {
		if (fread (buf, 2, 1, file) != 1) {
			return (1);
		}

		pos1 = get_uint16_be (buf, 0);

		ct->chr[i].bi_x = pos0;
		ct->chr[i].bi_w = pos1 - pos0;

		pos0 = pos1;
	}

	if (fread (buf, 2, 1, file) != 1) {
		return (1);
	}

	return (0);
}

int get_wotab (FILE *file, CharTab *ct)
{
	unsigned i;
	unsigned char buf[2];

	for (i = 0; i < ct->chrcnt; i++) {
		if (fread (buf, 2, 1, file) != 1) {
			return (1);
		}

		ct->chr[i].ofs = get_int8 (buf, 0);
		ct->chr[i].width = get_uint8_be (buf, 1);

		if (get_uint16_be (buf, 0) == 0xffff) {
			ct->chr[i].present = 0;
		}
		else {
			ct->chr[i].present = 1;
		}
	}

	if (fread (buf, 2, 1, file) != 1) {
		return (1);
	}

	return (0);
}

int set_bitmap (FILE *file, unsigned i, FontRec *fr, BitImage *bi, CharTab *ct)
{
	unsigned char buf[256];
	unsigned      x, y;
	unsigned      w;
	unsigned      n;
	unsigned      bi_x;

	w = ct->chr[i].bi_w;
	if (w == 0) {
		fprintf (file, "%02x\n", 0);
		return (0);
	}

	n = (w + 7) / 8;

	for (y = 0; y < fr->fRectHeight; y++) {
		bi_x = ct->chr[i].bi_x;

		for (x = 0; x < n; x++) {
			buf[x] = 0;
		}

		for (x = 0; x < w; x++) {
			if (bi->row[y][bi_x >> 3] & (0x80 >> (bi_x & 7))) {
				buf[x >> 3] |= 0x80 >> (x & 7);
			}

			bi_x += 1;
		}

		for (x = 0; x < n; x++) {
			fprintf (file, "%02x", buf[x]);
		}

		fprintf (file, "\n");
	}

	return (0);
}

int set_char (FILE *file, unsigned i, FontRec *fr, BitImage *bi, CharTab *ct)
{
	unsigned chr_w;

	chr_w = ct->chr[i].width;

	if (chr_w == 255) {
		chr_w = 0;
	}

	fprintf (file, "STARTCHAR C%04x\n", fr->firstChar + i);
	fprintf (file, "ENCODING %u\n", fr->firstChar + i);
	fprintf (file, "SWIDTH %u %u\n", 1000, 0);
	fprintf (file, "DWIDTH %u %u\n", chr_w, 0);

	if (ct->chr[i].bi_w == 0) {
		fprintf (file, "BBX %d %d %d %d\n",
			1, 1, 0, 0);
	}
	else {
		fprintf (file, "BBX %d %d %d %d\n",
			ct->chr[i].bi_w, fr->fRectHeight, fr->kernMax + ct->chr[i].ofs, -fr->descent);
	}

	fprintf (file, "BITMAP\n");

	if (set_bitmap (file, i, fr, bi, ct)) {
		return (1);
	}

	fprintf (file, "ENDCHAR\n");

	return (0);
}

unsigned ct_cnt_present (CharTab *ct)
{
	unsigned i, n;

	n = 0;

	for (i = 0; i < ct->chrcnt; i++) {
		if (ct->chr[i].present) {
			n += 1;
		}
	}

	return (n);
}

int set_bdf (FILE *fp, FontInfo *fi, FontRec *fr, BitImage *bi, CharTab *ct)
{
	unsigned i;

	fprintf (fp, "STARTFONT 2.1\n");
	fprintf (fp, "COMMENT Created by mactobdf by Hampa Hug\n");

	fprintf (fp, "FONT %s\n", fi->name);

	fprintf (fp, "SIZE %u 72 72\n", fi->size);

	fprintf (fp, "FONTBOUNDINGBOX %d %d %d %d\n",
		fr->fRectWidth, fr->fRectHeight, fr->kernMax, -fr->descent);

	fprintf (fp, "STARTPROPERTIES %u\n", 2);

	fprintf (fp, "FONT_ASCENT %d\n", fr->ascent);
	fprintf (fp, "FONT_DESCENT %d\n", fr->descent);

	fprintf (fp, "ENDPROPERTIES\n");

	fprintf (fp, "CHARS %u\n", ct_cnt_present (ct));

	for (i = 0; i < ct->chrcnt; i++) {
		if (ct->chr[i].present) {
			if (set_char (fp, i, fr, bi, ct)) {
				return (1);
			}
		}
	}

	fprintf (fp, "ENDFONT\n");

	return (0);
}

int mactobdf (FILE *inp, FILE *out)
{
	FontRec  fr;
	BitImage *bi;
	CharTab  ct;

	if (get_fontrec (inp, &fr)) {
		return (1);
	}

/*  prt_fontrec (stderr, &fr); */

	bi = bi_new (16 * fr.rowWords, fr.fRectHeight, fr.rowWords);
	if (bi == NULL) {
		return (1);
	}

	if (get_bitimage (inp, bi)) {
		bi_del (bi);
		return (1);
	}

	ct.chrcnt = fr.lastChar - fr.firstChar + 1;

	if (get_loctab (inp, &ct)) {
		bi_del (bi);
		return (1);
	}

	if (get_wotab (inp, &ct)) {
		bi_del (bi);
		return (1);
	}

	set_bdf (out, &finfo, &fr, bi, &ct);

	bi_del (bi);

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'f':
			strncpy (finfo.name, optarg[0], 255);
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 's':
			finfo.size = (unsigned) strtoul (optarg[0], NULL, 0);
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr,
					"%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "w");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	r = mactobdf (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	return (r);
}
