/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar.c                                                *
 * Created:     2011-06-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2020 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <lib/crc.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/mar.h>


static const char *arg0 = NULL;

crc32_t        par_crc;
unsigned char  par_buf[MAR_BUF_MAX];


char           par_creator_ok = 0;
char           par_file_flags_ok = 0;
char           par_finder_flags_ok = 0;
char           par_mac_name_ok = 0;
char           par_mar_id_ok = 0;
char           par_position_x_ok = 0;
char           par_position_y_ok = 0;
char           par_scroll_x_ok = 0;
char           par_scroll_y_ok = 0;
char           par_type_ok = 0;
char           par_window_view_ok = 0;
char           par_window_x1_ok = 0;
char           par_window_y1_ok = 0;
char           par_window_x2_ok = 0;
char           par_window_y2_ok = 0;

unsigned long  par_creator = 0;
const char     *par_ctime = NULL;
unsigned long  par_file_flags_msk = 0;
unsigned long  par_file_flags_val = 0;
unsigned short par_finder_flags_msk = 0;
unsigned short par_finder_flags_val = 0;
const char     *par_mac_name = NULL;
unsigned long  par_mar_id = 0;
const char     *par_mtime = NULL;
unsigned       par_position_x = 0;
unsigned       par_position_y = 0;
unsigned       par_scroll_x = 0;
unsigned       par_scroll_y = 0;
unsigned long  par_type = 0;
unsigned short par_window_view = 0;
unsigned       par_window_x1 = 0;
unsigned       par_window_y1 = 0;
unsigned       par_window_x2 = 0;
unsigned       par_window_y2 = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 0, "create", NULL, "Create a mar archive" },
	{ 'd', 0, "decode", NULL, "Decode a mar file" },
	{ 'e', 0, "encode", NULL, "Encode a mar file" },
	{ 'f', 0, "fix", NULL, "Set mar header fields" },
	{ 'i', 0, "info", NULL, "Print info about a mar file" },
	{ 'I', 0, "info-mf", NULL, "Print info about a mar file" },
	{ 'l', 0, "list", NULL, "List the contents of a mar archive" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 0, "extract", NULL, "Extract a mar archive" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar: encode and decode mar files",
		"usage: mar [options]",
		opts
	);

	fflush (stdout);
}

void print_help_properties (void)
{
	fputs (
		"\nproperties are:\n"
		"  creator ctime file-flags finder-flags mac-name mar-id\n"
		"  mtime position-x position-y scroll-x scroll-y type\n"
		"  window-view window-x1 window-y1 window-x2 window-y2\n",
		stdout
	);
}

void print_version (void)
{
	fputs (
		"mar version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2011-2020 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int set_file_flags (const char *str)
{
	unsigned long v;

	if (str[0] == '+') {
		v = mf_file_flag_from_string (str + 1);
		par_file_flags_msk |= v;
		par_file_flags_val |= v;
	}
	else if (str[0] == '-') {
		v = mf_file_flag_from_string (str + 1);
		par_file_flags_msk |= v;
		par_file_flags_val &= ~v;
	}
	else {
		v = mf_file_flag_from_string (str);
		par_file_flags_msk = 0xffffffff;
		par_file_flags_val = v;
	}

	if (v == 0) {
		return (1);
	}

	par_file_flags_ok = 1;

	return (0);
}

static
int set_finder_flags (const char *str)
{
	unsigned v;

	if (str[0] == '+') {
		v = mf_finder_flag_from_string (str + 1);
		par_finder_flags_msk |= v;
		par_finder_flags_val |= v;
	}
	else if (str[0] == '-') {
		v = mf_finder_flag_from_string (str + 1);
		par_finder_flags_msk |= v;
		par_finder_flags_val &= ~v;
	}
	else {
		v = mf_finder_flag_from_string (str);
		par_finder_flags_msk = 0xffff;
		par_finder_flags_val = v;
	}

	if (v == 0) {
		return (1);
	}

	par_finder_flags_ok = 1;

	return (0);
}

int mar_set_property (const char *key, const char *val)
{
	if (strcmp (key, "creator") == 0) {
		if (str_get_fourcc (val, &par_creator)) {
			fprintf (stderr, "%s: bad creator fourcc (%s)\n",
				arg0, val
			);
			return (1);
		}
		par_creator_ok = 1;
	}
	else if (strcmp (key, "ctime") == 0) {
		par_ctime = val;
	}
	else if (strcmp (key, "finder-flags") == 0) {
		if (set_finder_flags (val)) {
			return (1);
		}
	}
	else if (strcmp (key, "file-flags") == 0) {
		if (set_file_flags (val)) {
			return (1);
		}
	}
	else if (strcmp (key, "mar-id") == 0) {
		if (str_get_fourcc (val, &par_mar_id)) {
			fprintf (stderr, "%s: bad mar id fourcc (%s)\n",
				arg0, val
			);
			return (1);
		}
		par_mar_id_ok = 1;
	}
	else if (strcmp (key, "mac-name") == 0) {
		par_mac_name = val;
		par_mac_name_ok = 1;
	}
	else if (strcmp (key, "mtime") == 0) {
		par_mtime = val;
	}
	else if ((strcmp (key, "position-x") == 0) || (strcmp (key, "x") == 0)) {
		par_position_x = (unsigned) strtoul (val, NULL, 0);
		par_position_x_ok = 1;
	}
	else if ((strcmp (key, "position-y") == 0) || (strcmp (key, "y") == 0)) {
		par_position_y = (unsigned) strtoul (val, NULL, 0);
		par_position_y_ok = 1;
	}
	else if (strcmp (key, "scroll-x") == 0) {
		par_scroll_x = (unsigned) strtoul (val, NULL, 0);
		par_scroll_x_ok = 1;
	}
	else if (strcmp (key, "scroll-y") == 0) {
		par_scroll_y = (unsigned) strtoul (val, NULL, 0);
		par_scroll_y_ok = 1;
	}
	else if (strcmp (key, "type") == 0) {
		if (str_get_fourcc (val, &par_type)) {
			fprintf (stderr, "%s: bad type fourcc (%s)\n",
				arg0, val
			);
			return (1);
		}
		par_type_ok = 1;
	}
	else if ((strcmp (key, "window-view") == 0) || (strcmp (key, "view") == 0)) {
		par_window_view = (unsigned) strtoul (val, NULL, 0);
		par_window_view_ok = 1;
	}
	else if (strcmp (key, "window-x1") == 0) {
		par_window_x1 = (unsigned) strtoul (val, NULL, 0);
		par_window_x1_ok = 1;
	}
	else if (strcmp (key, "window-y1") == 0) {
		par_window_y1 = (unsigned) strtoul (val, NULL, 0);
		par_window_y1_ok = 1;
	}
	else if (strcmp (key, "window-x2") == 0) {
		par_window_x2 = (unsigned) strtoul (val, NULL, 0);
		par_window_x2_ok = 1;
	}
	else if (strcmp (key, "window-y2") == 0) {
		par_window_y2 = (unsigned) strtoul (val, NULL, 0);
		par_window_y2_ok = 1;
	}
	else {
		return (1);
	}

	return (0);
}

/*
 * Copy cnt bytes from inp to out and calculate their crc.
 */
int mar_copy_crc (FILE *inp, FILE *out, unsigned long cnt, unsigned long *crc)
{
	unsigned long i, n;

	if (cnt == 0) {
		return (0);
	}

	crc32_reset (&par_crc, 0);

	n = cnt;

	while (n > 0) {
		i = (n < MAR_BUF_MAX) ? n : MAR_BUF_MAX;

		if (fread (par_buf, 1, i, inp) != i) {
			return (1);
		}

		crc32_calc (&par_crc, par_buf, i);

		if (out != NULL) {
			if (fwrite (par_buf, 1, i, out) != i) {
				return (1);
			}
		}

		n -= i;
	}

	crc32_done (&par_crc);

	*crc = crc32_get_crc (&par_crc);

	return (0);
}

int main (int argc, char **argv)
{
	int  r, ret;
	char **optarg;

	arg0 = argv[0];

	crc32_init (&par_crc, MAR_CRC_POLY, 0);

	r = mac_getopt (argc, argv, &optarg, opts);

	if (r == GETOPT_DONE) {
		print_help();
		return (0);
	}

	if (r < 0) {
		return (1);
	}

	ret = 1;

	switch (r) {
	case '?':
		print_help();
		return (0);

	case 'V':
		print_version();
		return (0);

	case 'c':
		ret = mar_create_main (argc, argv);
		break;

	case 'd':
		ret = mar_decode_main (argc, argv);
		break;

	case 'e':
		ret = mar_encode_main (argc, argv);
		break;

	case 'f':
		ret = mar_fix_main (argc, argv);
		break;

	case 'i':
		ret = mar_info_main (argc, argv);
		break;

	case 'I':
		ret = mar_info_mf_main (argc, argv);
		break;

	case 'l':
		ret = mar_list_main (argc, argv);
		break;

	case 'x':
		ret = mar_extract_main (argc, argv);
		break;

	case 0:
	default:
		fprintf (stderr, "%s: unknown command (%s)\n", argv[0], optarg[0]);
		return (1);
	}

	return (ret);
}
