/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_decode.c                                         *
 * Created:     2011-06-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2019 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static const char *arg0 = NULL;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 1, "data-file", "string", "Set the data fork file name [none]" },
	{ 'f', 1, "mar-file", "string", "Set the mar file name [none]" },
	{ 'i', 1, "info-file", "string", "Set the info file name [none]" },
	{ 'r', 1, "resource-file", "string", "Set the resource fork file name [none]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -d: decode mar files",
		"usage: mar -d [options] [mar [data [resource [info]]]]",
		opts
	);

	fflush (stdout);
}

static
int mar_decode_block (FILE *inp, FILE *out, unsigned long cnt)
{
	unsigned long n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, out, cnt, &crc)) {
		return (1);
	}

	if (fread (par_buf, 1, 4, inp) != 4) {
		return (1);
	}

	if (get_uint32_be (par_buf, 0) != crc) {
		return (1);
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		if (fread (par_buf, 1, n, inp) != n) {
			return (1);
		}
	}

	return (0);
}

static
int mar_decode_mar (unsigned char *mar, FILE *fmar, FILE *fdat, FILE *frsc, FILE *finf)
{
	unsigned long datn, rscn;

	datn = mar_get_data_size (mar);
	rscn = mar_get_rsrc_size (mar);

	if (mar_decode_block (fmar, fdat, datn)) {
		return (1);
	}

	if (mar_decode_block (fmar, frsc, rscn)) {
		return (1);
	}

	if (finf != NULL) {
		macfile_t mf;

		mf_init (&mf);
		mar_get_macfile (mar, &mf);
		mf_save (&mf, finf);
	}

	return (0);
}

static
int mar_decode (const char *inp, const char *dat, const char *rsc, const char *inf)
{
	int           r;
	unsigned long tm;
	FILE          *finp, *fdat, *frsc, *finf;
	unsigned char mar[128];

	finp = mac_open_inp (inp, "rb", 0);

	if (finp == NULL) {
		fprintf (stderr, "%s: error opening mar file (%s)\n", arg0, inp);
		return (1);
	}

	if (mar_read (mar, finp)) {
		return (1);
	}

	if (mar_is_mar (mar) == 0) {
		fprintf (stderr, "%s: not a mar file (%s)\n", arg0, inp);
		return (1);
	}

	fdat = mac_open_out (dat, "wb", 0);

	if ((dat != NULL) && (fdat == NULL)) {
		fprintf (stderr, "%s: error opening data fork file (%s)\n", arg0, dat);
		return (1);
	}

	frsc = mac_open_out (rsc, "wb", 0);

	if ((rsc != NULL) && (frsc == NULL)) {
		fprintf (stderr, "%s: error opening resource fork file (%s)\n", arg0, rsc);
		return (1);
	}

	finf = mac_open_out (inf, "w", 0);

	if ((inf != NULL) && (finf == NULL)) {
		fprintf (stderr, "%s: error opening finf file (%s)\n", arg0, inf);
		return (1);
	}

	r = mar_decode_mar (mar, finp, fdat, frsc, finf);

	if (r) {
		return (1);
	}

	mac_close (finf);
	mac_close (frsc);
	mac_close (fdat);
	mac_close (finp);

	tm = mar_get_mtime (mar);

	uf_set_mtime (dat, tm);
	uf_set_mtime (rsc, tm);

	return (0);
}

int mar_decode_main (int argc, char **argv)
{
	int  r;
	char **optarg;
	const char *inp, *dat, *rsc, *inf;

	arg0 = argv[0];

	inp = NULL;
	dat = NULL;
	rsc = NULL;
	inf = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'd':
			dat = optarg[0];
			break;

		case 'f':
			inp = optarg[0];
			break;

		case 'i':
			inf = optarg[0];
			break;

		case 'r':
			rsc = optarg[0];
			break;

		case 0:
			if (inp == NULL) {
				inp = optarg[0];
			}
			else if (dat == NULL) {
				dat = optarg[0];
			}
			else if (rsc == NULL) {
				rsc = optarg[0];
			}
			else if (inf == NULL) {
				inf = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		default:
			return (1);
		}
	}

	if (mar_decode (inp, dat, rsc, inf)) {
		return (1);
	}

	return (0);
}
