/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_extract.c                                        *
 * Created:     2011-06-26 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2016 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


#define MAR_MAX_PATH 4096


static const char *arg0 = NULL;


static int        par_verbose = 0;
static int        par_data_only = 0;

static const char *par_folder_file = ".folder";


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 0, "data-only", NULL, "Only extract data forks [no]" },
	{ 'f', 1, "mar-file", "string", "Set the input file name [none]" },
	{ 'F', 1, "folder-file", "string", "Set the folder file name [.folder]" },
	{ 'v', 0, "verbose", NULL, "Print file names to stderr [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -x: extract mar archives",
		"usage: mar -x [options]",
		opts
	);

	fflush (stdout);
}

static
int mar_extract_block (FILE *inp, FILE *out, unsigned long cnt, int fill)
{
	unsigned long n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, out, cnt, &crc)) {
		return (1);
	}

	if (fread (par_buf, 1, 4, inp) != 4) {
		return (1);
	}

	if (get_uint32_be (par_buf, 0) != crc) {
		fprintf (stderr, "%s: crc error\n", arg0);
		return (1);
	}

	if ((out != NULL) && fill) {
		if (fwrite (par_buf, 1, 4, out) != 4) {
			return (1);
		}
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		if (fread (par_buf, 1, n, inp) != n) {
			return (1);
		}

		if ((out != NULL) && fill) {
			memset (par_buf, 0, n);

			if (fwrite (par_buf, 1, n, out) != n) {
				return (1);
			}
		}
	}

	return (0);
}

static
int mar_extract_file_fp (FILE *inp, FILE *out, const unsigned char *mar)
{
	if (fwrite (mar, 1, 128, out) != 128) {
		return (1);
	}

	if (mar_extract_block (inp, out, mar_get_data_size (mar), 1)) {
		return (1);
	}

	if (mar_extract_block (inp, out, mar_get_rsrc_size (mar), 1)) {
		return (1);
	}

	if (mar_extract_block (inp, out, mar_get_comment_size (mar), 1)) {
		return (1);
	}

	return (0);
}

static
int mar_extract_data_fp (FILE *inp, FILE *out, const unsigned char *mar)
{
	if (mar_extract_block (inp, out, mar_get_data_size (mar), 0)) {
		return (1);
	}

	if (mar_extract_block (inp, NULL, mar_get_rsrc_size (mar), 0)) {
		return (1);
	}

	if (mar_extract_block (inp, NULL, mar_get_comment_size (mar), 0)) {
		return (1);
	}

	return (0);
}

static
int mar_prefix_add (char *prefix, const char *name, const char *ext)
{
	unsigned long np, nn, ne;

	np = strlen (prefix);
	nn = strlen (name);
	ne = strlen (ext);

	if ((np + nn + ne + 1) >= MAR_MAX_PATH) {
		return (1);
	}

	if (np > 0) {
		prefix[np++] = '/';
	}

	strcpy (prefix + np, name);
	strcpy (prefix + np + nn, ext);

	return (0);
}

static
int mar_get_file_name (const unsigned char *mar, char *prefix, const char *ext)
{
	int  r;
	char *fname;

	fname = mar_get_name_cstr (mar);

	if (fname == NULL) {
		return (1);
	}

	r = mar_prefix_add (prefix, fname, ext);

	free (fname);

	return (r);
}

static
int mar_extract_file (FILE *inp, const unsigned char *mar, char *prefix)
{
	int           r;
	FILE          *out;

	if (mar_get_file_name (mar, prefix, par_data_only ? "" : ".mac")) {
		return (1);
	}

	if (par_verbose) {
		fprintf (stderr, "%s\n", prefix);
	}

	out = fopen (prefix, "wb");

	if (out == NULL) {
		return (1);
	}

	if (par_data_only) {
		r = mar_extract_data_fp (inp, out, mar);
	}
	else {
		r = mar_extract_file_fp (inp, out, mar);
	}

	fclose (out);

	if (r) {
		return (1);
	}

	uf_set_mtime (prefix, mar_get_mtime (mar));

	return (r);
}

static
int mar_extract_folder_file (const unsigned char *mar, char *prefix)
{
	FILE *fp;

	if (par_folder_file == NULL) {
		return (0);
	}

	if (mar_prefix_add (prefix, par_folder_file, "")) {
		return (1);
	}

	fp = fopen (prefix, "wb");

	if (fp == NULL) {
		fprintf (stderr, "%s: error creating folder file (%s)\n",
			arg0, prefix
		);
		return (1);
	}

	if (fwrite (mar, 1, 128, fp) != 128) {
		fclose (fp);
		return (1);
	}

	fclose (fp);

	uf_set_mtime (prefix, mar_get_mtime (mar));

	return (0);
}

static
int mar_extract_folder (FILE *inp, unsigned char *mar, char *prefix)
{
	unsigned long n;
	unsigned long id;
	unsigned long mtime;

	if (mar_get_file_name (mar, prefix, "")) {
		return (1);
	}

	if (par_verbose) {
		fprintf (stderr, "%s/\n", prefix);
	}

#ifdef __MINGW32__
	if (mkdir (prefix) != 0) {
#else
	if (mkdir (prefix, 0777) != 0) {
#endif
		fprintf (stderr, "%s: can't make directory (%s)\n",
			arg0, prefix
		);
		return (1);
	}

	n = strlen (prefix);

	mtime = mar_get_mtime (mar);

	if (par_data_only == 0) {
		if (mar_extract_folder_file (mar, prefix)) {
			return (1);
		}
	}

	prefix[n] = 0;

	while (1) {
		if (fread (mar, 1, 128, inp) != 128) {
			fprintf (stderr, "%s: error reading mar header\n", arg0);
			return (1);
		}

		if (mar_is_mar (mar) == 0) {
			fprintf (stderr, "%s: bad mar header\n", arg0);
			return (1);
		}

		prefix[n] = 0;

		id = mar_get_id (mar);

		if (id == MAR_FILE) {
			if (mar_extract_file (inp, mar, prefix)) {
				return (1);
			}
		}
		else if (id == MAR_FOLDER_START) {
			if (mar_extract_folder (inp, mar, prefix)) {
				return (1);
			}
		}
		else if (id == MAR_FOLDER_END) {
			break;
		}
		else {
			fprintf (stderr,
				"%s: bad mar header (not file or folder)\n",
				arg0
			);
			return (1);
		}
	}

	prefix[n] = 0;

	uf_set_mtime (prefix, mtime);

	return (0);
}

static
int mar_extract_fp (FILE *inp, const char *fname)
{
	unsigned char mar[128];
	char          prefix[MAR_MAX_PATH];
	unsigned long id;

	prefix[0] = 0;

	if (fread (mar, 1, 128, inp) != 128) {
		fprintf (stderr,
			"%s: error reading the archive header (%s)\n",
			arg0, fname
		);
		return (1);
	}

	if (mar_is_mar (mar) == 0) {
		fprintf (stderr, "%s: not a mar archive (%s)\n",
			arg0, fname
		);
		return (1);
	}

	id = mar_get_id (mar);

	if (id == MAR_FILE) {
		if (mar_extract_file (inp, mar, prefix)) {
			return (1);
		}
	}
	else if (id == MAR_FOLDER_START) {
		if (mar_extract_folder (inp, mar, prefix)) {
			return (1);
		}
	}
	else {
		fprintf (stderr,
			"%s: archive header error (not file or folder)\n",
			arg0
		);
		return (1);
	}

	return (0);
}

static
int mar_extract (const char *inp)
{
	int  r;
	FILE *fp;

	fp = mac_open_inp (inp, "rb", 0);

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", arg0, inp);
		return (1);
	}

	r = mar_extract_fp (fp, inp);

	mac_close (fp);

	return (r);
}

int mar_extract_main (int argc, char **argv)
{
	int  r;
	char **optarg;

	arg0 = argv[0];

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'd':
			par_data_only = 1;
			break;

		case 'f':
			if (mar_extract (optarg[0])) {
				return (1);
			}
			break;

		case 'F':
			par_folder_file = optarg[0];

			if (par_folder_file[0] == 0) {
				par_folder_file = NULL;
			}
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (mar_extract (optarg[0])) {
				return (1);
			}
			break;

		default:
			fprintf (stderr, "%s: unknown option (%s)\n", arg0, optarg[0]);
			return (1);
		}
	}

	return (0);
}
