/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_list.c                                           *
 * Created:     2011-06-26 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2019 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


#define MAR_MAX_PATH 4096


static const char *arg0 = NULL;


static char par_verbose = 0;
static char par_list_ctime = 0;
static char par_list_mtime = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'C', 0, "ctime", NULL, "List the creation time [no]" },
	{ 'f', 1, "mar-file", "string", "Set the input file name [none]" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flag filter [none]" },
	{ 'L', 1, "file-flags", "int", "Set the file flag filter [none]" },
	{ 'M', 0, "mtime", NULL, "List the modification time [yes]" },
	{ 'v', 0, "verbose", NULL, "Print file names to stderr [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -l: list the contents of mar archives",
		"usage: mar -l [options] [file]",
		opts
	);

	fflush (stdout);
}

static
int mar_prefix_add (char *prefix, const char *name, const char *ext)
{
	unsigned long np, nn, ne;

	np = strlen (prefix);
	nn = strlen (name);
	ne = strlen (ext);

	if ((np + nn + ne + 1) >= MAR_MAX_PATH) {
		return (1);
	}

	if (np > 0) {
		prefix[np++] = '/';
	}

	strcpy (prefix + np, name);
	strcpy (prefix + np + nn, ext);

	return (0);
}

static
int mar_get_file_name (const unsigned char *mar, char *prefix, const char *ext)
{
	int  r;
	char *fname;

	fname = mar_get_name_cstr (mar);

	if (fname == NULL) {
		return (1);
	}

	r = mar_prefix_add (prefix, fname, ext);

	free (fname);

	return (r);
}

static
int mar_list_block (FILE *inp, unsigned long cnt)
{
	unsigned long n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, NULL, cnt, &crc)) {
		return (1);
	}

	if (fread (par_buf, 1, 4, inp) != 4) {
		return (1);
	}

	if (get_uint32_be (par_buf, 0) != crc) {
		fprintf (stderr, "%s: crc error\n", arg0);
		return (1);
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		if (fread (par_buf, 1, n, inp) != n) {
			return (1);
		}
	}

	return (0);
}

static
int mar_list_mar (const unsigned char *mar, const char *fname)
{
	int           isdir;
	unsigned      yy, mm, dd;
	unsigned      hrs, min, sec;
	unsigned long mtime;
	unsigned long data, rsrc;
	unsigned long type, creator;
	unsigned      fflags, attrib;
	char          stype[16], screator[16];

	fflags = mar_get_finder_flags (mar);

	if ((par_finder_flags_val ^ fflags) & par_finder_flags_msk) {
		return (0);
	}

	attrib = mar_get_file_flags (mar);

	if ((par_file_flags_val ^ attrib) & par_file_flags_msk) {
		return (0);
	}

	isdir = (attrib & MF_FILE_ISDIR) != 0;

	if (isdir) {
		data = 0;
		rsrc = 0;

		type = 0x2d2d2d2d;
		creator = 0x2d2d2d2d;
	}
	else {
		data = mar_get_data_size (mar);
		rsrc = mar_get_rsrc_size (mar);

		type = mar_get_type (mar);
		creator = mar_get_creator (mar);
	}

	str_set_fourcc (stype, 16, type);
	str_set_fourcc (screator, 16, creator);

	fprintf (stdout, "%s %04X %02X %8lu %8lu",
		isdir ? "drwxr-xr-x" : "-rw-r--r--",
		fflags, attrib, data, rsrc
	);

	if (par_list_ctime) {
		mtime = mar_get_ctime (mar);
		mac_time_to_ymd (mtime, &yy, &mm, &dd);
		mac_time_to_hms (mtime, &hrs, &min, &sec);

		fprintf (stdout, "  %04u-%02u-%02u %02u:%02u:%02u",
			yy, mm + 1, dd + 1, hrs, min, sec
		);
	}

	if (par_list_mtime || !par_list_ctime) {
		mtime = mar_get_mtime (mar);
		mac_time_to_ymd (mtime, &yy, &mm, &dd);
		mac_time_to_hms (mtime, &hrs, &min, &sec);

		fprintf (stdout, "  %04u-%02u-%02u %02u:%02u:%02u",
			yy, mm + 1, dd + 1, hrs, min, sec
		);
	}

	fprintf (stdout, "  %-6s %-7s %s%s\n",
		stype, screator, fname, isdir ? "/" : ""
	);

	return (0);
}

static
int mar_list_file (FILE *inp, const unsigned char *mar, char *prefix)
{
	if (mar_get_file_name (mar, prefix, "")) {
		return (1);
	}

	mar_list_mar (mar, prefix);

	if (mar_list_block (inp, mar_get_data_size (mar))) {
		return (1);
	}

	if (mar_list_block (inp, mar_get_rsrc_size (mar))) {
		return (1);
	}

	if (mar_list_block (inp, mar_get_comment_size (mar))) {
		return (1);
	}

	return (0);
}

static
int mar_list_folder (FILE *inp, unsigned char *mar, char *prefix)
{
	unsigned long n;
	unsigned long id;

	if (mar_get_file_name (mar, prefix, "")) {
		return (1);
	}

	mar_list_mar (mar, prefix);

	n = strlen (prefix);

	while (1) {
		if (fread (mar, 1, 128, inp) != 128) {
			return (1);
		}

		if (mar_is_mar (mar) == 0) {
			return (1);
		}

		prefix[n] = 0;

		id = mar_get_id (mar);

		if (id == MAR_FILE) {
			if (mar_list_file (inp, mar, prefix)) {
				return (1);
			}
		}
		else if (id == MAR_FOLDER_START) {
			if (mar_list_folder (inp, mar, prefix)) {
				return (1);
			}
		}
		else if (id == MAR_FOLDER_END) {
			break;
		}
		else {
			fprintf (stderr, "%s: mar header error\n",
				arg0
			);
			return (1);
		}
	}

	prefix[n] = 0;

	return (0);
}

static
int mar_list_fp (FILE *inp)
{
	unsigned char mar[128];
	char          prefix[MAR_MAX_PATH];
	unsigned long id;

	prefix[0] = 0;

	while (1) {
		if (fread (mar, 1, 128, inp) != 128) {
			return (1);
		}

		if (mar_is_mar (mar) == 0) {
			return (1);
		}

		id = mar_get_id (mar);

		if (id == MAR_FILE) {
			if (mar_list_file (inp, mar, prefix)) {
				return (1);
			}
		}
		else if (id == MAR_FOLDER_START) {
			if (mar_list_folder (inp, mar, prefix)) {
				return (1);
			}
		}
		else {
			fprintf (stderr, "%s: mar header error\n",
				arg0
			);
			return (1);
		}
	}

	return (0);
}

static
int mar_list (const char *inp)
{
	int  r;
	FILE *fp;

	fp = mac_open_inp (inp, "rb", 0);

	if (fp == NULL) {
		return (1);
	}

	r = mar_list_fp (fp);

	mac_close (fp);

	return (r);
}

int mar_list_main (int argc, char **argv)
{
	int        r;
	char       **optarg;
	const char *inp;

	arg0 = argv[0];

	inp = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'C':
			par_list_ctime = 1;
			break;

		case 'f':
			inp = optarg[0];
			break;

		case 'l':
			if (mar_set_property ("finder-flags", optarg[0])) {
				fprintf (stderr, "%s: bad finder flags (%s)\n",
					arg0, optarg[0]
				);
				return (1);
			}
			break;

		case 'L':
			if (mar_set_property ("file-flags", optarg[0])) {
				fprintf (stderr, "%s: bad file flags (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		case 'M':
			par_list_mtime = 1;
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (inp == NULL) {
				inp = optarg[0];
			}
			else {
				fprintf (stderr,
					"%s: too many files (%s)\n",
					arg0, optarg[0]
				);
				return (1);
			}
			break;

		default:
			return (1);
		}
	}

	if (mar_list (inp)) {
		return (1);
	}

	return (0);
}
