/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mbp/mbpenc.c                                             *
 * Created:     2007-12-02 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"

#include "mbpenc.h"

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/dir.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/strings.h>


typedef struct {
	FILE          *fp;

	unsigned long ofs;

	unsigned long data_cnt;
	unsigned long rsrc_cnt;
} mbp_enc_t;


static int mbp_encode_entry (void *ext, unsigned op, const mac_stat_t *ms);


static const char    *arg0 = NULL;

static int           par_verbose = 1;

static int           par_reset = 0;

static unsigned long par_type = MAC_FOURCC_UNKN;
static unsigned long par_creator = MAC_FOURCC_UNKN;

static const char    *par_folder_file = ".folder";


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'f', 1, "folder-file", "string", "Set the folder file name [.folder]" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'o', 1, "output", "string", "Set the output file name" },
	{ 'q', 0, "quiet", NULL, "Disable printing of file names [no]" },
	{ 'r', 0, "reset", NULL, "Reset file and finder flags [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mbpenc: encode MacBinary II+ files",
		"usage: mbpenc [options] output [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mbpenc version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2007-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int mbp_copy (FILE *dst, unsigned long dofs, FILE *src, unsigned long sofs, unsigned long cnt)
{
	unsigned long n;
	unsigned char buf[512];

	while (cnt > 0) {
		n = (cnt < 512) ? cnt : 512;

		if (mac_read (src, buf, sofs, n)) {
			return (1);
		}

		if (mac_write (dst, buf, dofs, n)) {
			return (1);
		}

		sofs += n;
		dofs += n;

		cnt -= n;
	}

	return (0);
}

static
void mbp_init (mbp_enc_t *mbp)
{
	mbp->fp = NULL;

	mbp->ofs = 0;

	mbp->data_cnt = 0;
	mbp->rsrc_cnt = 0;
}

static
void mbp_free (mbp_enc_t *mbp)
{
	mac_close (mbp->fp);
}

static
int mbp_set_fname (mbp_enc_t *mbp, const char *fname)
{
	if (mbp->fp != NULL) {
		return (1);
	}

	if (fname == NULL) {
		return (1);
	}

	mbp->fp = mac_open_out (fname, "wb", 0);

	if (mbp->fp == NULL) {
		return (1);
	}

	return (0);
}

static
void mbp_reset (unsigned char *mb, int folder)
{
	mb_set_position_x (mb, 0);
	mb_set_position_y (mb, 0);

	mb_set_scroll_x (mb, 0);
	mb_set_scroll_y (mb, 0);

	mb_set_folder_id (mb, 0);

	if (folder) {
		mb_set_data_size (mb, 0);
		mb_set_rsrc_size (mb, 0);
	}

	mb_set_finder_flags (mb, mb_get_finder_flags (mb) & ~MF_FINDER_INITED);
	mb_set_file_flags (mb, 0);

	mb_set_crc (mb, mb_calc_crc (mb));
}

static
int mbp_encode_file_mb (mbp_enc_t *mbp, const mac_stat_t *ms, FILE *fp, unsigned char *mb)
{
	unsigned long  rsrc, data, info;
	unsigned long  rsrc_r, data_r, info_r;

	if (par_reset) {
		mbp_reset (mb, 0);
	}

	data = mb_get_data_size (mb);
	rsrc = mb_get_rsrc_size (mb);
	info = mb_get_getinfo_size (mb);

	data_r = (data + 127) & ~0x7fUL;
	rsrc_r = (rsrc + 127) & ~0x7fUL;
	info_r = (info + 127) & ~0x7fUL;

	if (mac_write (mbp->fp, mb, mbp->ofs, 128)) {
		return (1);
	}

	mbp->ofs += 128;

	if (mbp_copy (mbp->fp, mbp->ofs, fp, 128, data_r + rsrc_r + info_r)) {
		return (1);
	}

	mbp->ofs += data_r + rsrc_r + info_r;

	mbp->rsrc_cnt += rsrc;
	mbp->data_cnt += data;

	if (par_verbose) {
		fprintf (stdout, "M %8lu %8lu %s\n", data, rsrc, ms->name);
	}

	return (0);
}

static
int mbp_encode_file_data (mbp_enc_t *mbp, const mac_stat_t *ms, FILE *fp)
{
	unsigned char mb[128];
	unsigned long data;
	unsigned long mtime;

	mb_init (mb);

	if (mb_set_name_cstr (mb, mac_get_last_component (ms->name, '/'))) {
		return (1);
	}

	mb_set_type (mb, par_type);
	mb_set_creator (mb, par_creator);

	mtime = mac_time_from_unix (ms->mtime);
	mb_set_mtime (mb, mtime);
	mb_set_ctime (mb, mtime);

	data = ms->size;

	mb_set_data_size (mb, data);
	mb_set_rsrc_size (mb, 0);

	mb_set_crc (mb, mb_calc_crc (mb));

	if (mac_write (mbp->fp, mb, mbp->ofs, 128)) {
		return (1);
	}

	mbp->ofs += 128;

	if (mbp_copy (mbp->fp, mbp->ofs, fp, 0, data)) {
		return (1);
	}

	mbp->ofs += data;

	if (data & 127) {
		unsigned n;

		n = 128 - (data & 127);

		memset (mb, 0, n);

		if (mac_write (mbp->fp, mb, mbp->ofs, n)) {
			return (1);
		}

		mbp->ofs += n;
	}

	mbp->data_cnt += data;

	if (par_verbose) {
		fprintf (stdout, "D %8lu %8lu %s\n", data, 0UL, ms->name);
	}

	return (0);
}

static
int mbp_encode_file (mbp_enc_t *mbp, const mac_stat_t *ms)
{
	int           r;
	FILE          *fp;
	unsigned char mb[128];

	if (par_folder_file != NULL) {
		if (strcmp (mac_get_last_component (ms->name, '/'), par_folder_file) == 0) {
			return (0);
		}
	}

	fp = fopen (ms->name, "rb");
	if (fp == NULL) {
		return (1);
	}

	if (mb_read (mb, fp)) {
		r = mbp_encode_file_data (mbp, ms, fp);
	}
	else {
		if (mb_get_version (mb) != 2) {
			r = mbp_encode_file_data (mbp, ms, fp);
		}
		else {
			r = mbp_encode_file_mb (mbp, ms, fp, mb);
		}
	}

	fclose (fp);

	if (r) {
		fprintf (stderr, "*** error (%s)\n", ms->name);
	}

	return (r);
}

static
int mbp_encode_folder_start_inf (unsigned char *mb, const char *fname)
{
	FILE *fp;

	fp = fopen (fname, "rb");
	if (fp == NULL) {
		return (1);
	}

	memset (mb, 0, 128);

	mac_read (fp, mb, 0, 128);

	if (mb_calc_crc (mb) != mb_get_crc (mb)) {
		macfile_t mf;

		fseek (fp, 0, SEEK_SET);

		if (mf_load (&mf, fp)) {
			fprintf (stderr,
				"%s: folder file error\n",
				arg0
			);

			fclose (fp);

			return (1);
		}

		mb_init (mb);
		mb_set_macfile (mb, &mf);
	}

	fclose (fp);

	if ((mb[0] != 0) && (mb[0] != 1)) {
		fprintf (stderr,
			"%s: folder file error (id)\n",
			arg0
		);
		return (1);
	}

	if (mb_get_type (mb) != MAC_FOURCC_FOLD) {
		fprintf (stderr,
			"%s: folder file error (type)\n",
			arg0
		);
		return (1);
	}

	return (0);
}

static
int mbp_encode_folder_start (mbp_enc_t *mbp, const mac_stat_t *ms, unsigned char *mb)
{
	int           r;
	unsigned long mtime;
	char          *fold;

	if (par_folder_file != NULL) {
		fold = mac_combine_path (ms->name, par_folder_file);
		r = mbp_encode_folder_start_inf (mb, fold);
		free (fold);
	}
	else {
		r = 1;
	}

	if (r) {
		mb_init (mb);

		if (mb_set_name_cstr (mb, mac_get_last_component (ms->name, '/'))) {
			return (1);
		}

		mtime = mac_time_from_unix (ms->mtime);
		mb_set_mtime (mb, mtime);
		mb_set_ctime (mb, mtime);

	}

	if (par_reset) {
		mbp_reset (mb, 1);
	}

	mb[0] = 1;

	mb_set_type (mb, MAC_FOURCC_FOLD);
	mb_set_creator (mb, 0xffffffff);

	mb_set_vers_read (mb, 0x82);
	mb_set_vers_write (mb, 0x82);

	mb_set_crc (mb, mb_calc_crc (mb));

	if (mac_write (mbp->fp, mb, mbp->ofs, 128)) {
		return (1);
	}

	mbp->ofs += 128;

	return (0);
}

static
int mbp_encode_folder_done (mbp_enc_t *mbp, const mac_stat_t *ms, unsigned char *mb)
{
	mb_set_type (mb, MAC_FOURCC_FOLD);
	mb_set_creator (mb, 0xfffffffe);

	mb[0] = 1;

	mb_set_crc (mb, mb_calc_crc (mb));

	if (mac_write (mbp->fp, mb, mbp->ofs, 128)) {
		return (1);
	}

	mbp->ofs += 128;

	return (0);
}

static
int mbp_encode_folder (mbp_enc_t *mbp, const mac_stat_t *ms)
{
	unsigned char mb[128];
	unsigned long data, rsrc;

	rsrc = mbp->rsrc_cnt;
	data = mbp->data_cnt;

	if (mbp_encode_folder_start (mbp, ms, mb)) {
		fprintf (stderr, "*** error: folder start (%s)\n", ms->name);
		return (1);
	}

	if (mac_dir_list_dir (ms->name, 0, mbp, mbp_encode_entry)) {
		return (1);
	}

	rsrc = mbp->rsrc_cnt - rsrc;
	data = mbp->data_cnt - data;

	if (mbp_encode_folder_done (mbp, ms, mb)) {
		fprintf (stderr, "*** error: folder done (%s)\n", ms->name);
		return (1);
	}

	if (par_verbose) {
		fprintf (stdout, "F %8lu %8lu %s\n", data, rsrc, ms->name);
	}

	return (0);
}

static
int mbp_encode_entry (void *ext, unsigned op, const mac_stat_t *ms)
{
	mbp_enc_t *mbp = ext;

	if (op == MAC_LIST_ENTER) {
		if (mbp_encode_folder (mbp, ms)) {
			return (1);
		}

		return (0);
	}

	if (op == MAC_LIST_LEAVE) {
		return (0);
	}

	if (op == MAC_LIST_LIST) {
		if (ms->is_file) {
			if (mbp_encode_file (mbp, ms)) {
				return (1);
			}
		}
		else if (ms->is_dir) {
			return (0);
		}
		else {
			fprintf (stderr, "*** error: bad file type (%s)\n",
				ms->name
			);
			return (1);
		}
	}

	return (0);
}

static
int mbp_encode (mbp_enc_t *mbp, const char *src)
{
	if (mbp->fp == NULL) {
		fprintf (stderr, "%s: need an output file name\n", arg0);
		return (1);
	}

	while ((src[0] == '.') && (src[1] == '/')) {
		src += 2;
	}

	if (mac_dir_list (src, 0, mbp, mbp_encode_entry)) {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int       r;
	char      **optarg;
	mbp_enc_t mbp;

	arg0 = argv[0];

	mbp_init (&mbp);

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'f':
			par_folder_file = optarg[0];

			if (par_folder_file[0] == 0) {
				par_folder_file = NULL;
			}
			break;

		case 'i':
			if (mbp_encode (&mbp, optarg[0])) {
				return (1);
			}
			break;

		case 'o':
			if (mbp_set_fname (&mbp, optarg[0])) {
				fprintf (stderr,
					"%s: can't open output file\n",
					arg0
				);
				return (1);
			}
			break;

		case 'q':
			par_verbose = 0;
			break;

		case 'r':
			par_reset = 1;
			break;

		case 0:
			if (mbp.fp == NULL) {
				if (mbp_set_fname (&mbp, optarg[0])) {
					fprintf (stderr,
						"%s: can't open output file (%s)\n",
						arg0, optarg[0]
					);
					return (1);
				}
			}
			else {
				if (mbp_encode (&mbp, optarg[0])) {
					return (1);
				}
			}
			break;

		default:
			return (1);
		}
	}

	mbp_free (&mbp);

	return (0);
}
