/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mfs/mfsls.c                                              *
 * Created:     2017-11-04 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2017 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/mfs.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static const char    *arg0 = NULL;

static unsigned long par_offset = 0;
static unsigned long par_blocks = 0;

static char          par_ctime = 0;
static char          par_long = 0;
static char          par_numeric = 0;
static char          par_physical = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 0, "ctime", NULL, "Use creation time instead of modification time [no]" },
	{ 'i', 1, "input", "string", "Set an input file name [none]" },
	{ 'l', 0, "long", NULL, "Long format [no]" },
	{ 'n', 0, "numeric", NULL, "Numeric type/creator and mtime [no]" },
	{ 'N', 1, "blocks", "int", "Set the volume size in blocks [auto]" },
	{ 'O', 1, "offset", "int", "Set the volume starting offset [0]" },
	{ 'p', 0, "physical", NULL, "Use physical file sizes [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mfsls: list the directory of MFS volumes",
		"usage: mfsls [options] [file...]",
		opts
	);

	fputs ("\nlong format: FL VV FNUM FOLD FFLG MTIME DATA RSRC TYPE CRTR NAME\n"
		"  FL    = File flags\n"
		"  VV    = Version\n"
		"  FNUM  = File number\n"
		"  FOLD  = Parent folder ID\n"
		"  FFLG  = Finder flags\n"
		"  MTIME = Modification or creation time\n"
		"  DATA  = Data fork size\n"
		"  RSRC  = Resource fork size\n"
		"  TYPE  = File type\n"
		"  CRTR  = File creator\n"
		"  NAME  = File name\n",
		stdout
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mfsls version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2017 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
int get_directory (mfs_vol_t *mfs, unsigned char **buf, unsigned *size)
{
	*size = mfs_get_directory_size (mfs);

	if ((*buf = malloc (*size)) == NULL) {
		return (1);
	}

	if (mfs_read_directory (mfs, *buf, *size) != *size) {
		free (*buf);
		return (1);
	}

	return (0);
}

static
int list_entry_long (FILE *fp, unsigned char *buf, unsigned long max)
{
	unsigned      folder, finder_flags;
	unsigned long type, creator;
	unsigned long data, rsrc;
	unsigned long mtime;
	unsigned long file;
	char          stype[16], screator[16], date[32];
	char          *str;

	str = mac_pstring_decode (buf + 50, "");

	if (str == NULL) {
		return (max);
	}

	if (par_physical) {
		data = get_uint32_be (buf, 28);
		rsrc = get_uint32_be (buf, 38);
	}
	else {
		data = get_uint32_be (buf, 24);
		rsrc = get_uint32_be (buf, 34);
	}

	type = get_uint32_be (buf, 2);
	creator = get_uint32_be (buf, 6);

	if (par_numeric) {
		sprintf (stype, "%04lX", type);
		sprintf (screator, "%04lX", creator);
	}
	else {
		str_set_fourcc (stype, 16, type);
		str_set_fourcc (screator, 16, creator);
	}

	folder = get_uint16_be (buf, 16);
	finder_flags = get_uint16_be (buf, 10);

	if (par_ctime) {
		mtime = get_uint32_be (buf, 42);
	}
	else {
		mtime = get_uint32_be (buf, 46);
	}

	if (par_numeric) {
		sprintf (date, "%10lu", mtime);
	}
	else {
		unsigned yy, mm, dd, hrs, min, sec;

		mac_time_to_ymd (mtime, &yy, &mm, &dd);
		mac_time_to_hms (mtime, &hrs, &min, &sec);

		sprintf (date, "%04u-%02u-%02u %02u:%02u:%02u",
			yy, mm + 1, dd + 1, hrs, min, sec
		);
	}

	file = get_uint32_be (buf, 18);

	fprintf (fp, "%02X %02X %04lX %04X %04X %s %8lu %8lu  %-6s %-6s %s\n",
		buf[0], buf[1], file, folder, finder_flags,
		date,
		data, rsrc, stype, screator, str
	);

	free (str);

	return (0);
}

static
int list_entry_short (FILE *fp, unsigned char *buf, unsigned long max)
{
	char *str;

	str = mac_pstring_decode (buf + 50, "");

	if (str == NULL) {
		return (1);
	}

	fprintf (fp, "%s\n", str);

	free (str);

	return (0);
}

static
unsigned long list_entry (FILE *fp, unsigned char *buf, unsigned long max)
{
	int      r;
	unsigned ncnt;

	if (max < 51) {
		return (max);
	}

	ncnt = buf[50];

	if ((ncnt + 51) > max) {
		return (max);
	}

	if (par_long) {
		r = list_entry_long (fp, buf, max);
	}
	else {
		r = list_entry_short (fp, buf, max);
	}

	if (r) {
		return (max);
	}

	return (ncnt + 51);
}

static
int mfs_list_fp (mfs_vol_t *mfs, FILE *fp)
{
	unsigned      size;
	unsigned char *buf;
	unsigned long idx;

	if (get_directory (mfs, &buf, &size)) {
		fprintf (stderr, "%s: reading directory failed\n", arg0);
		return (1);
	}

	idx = 0;

	while (idx < size) {
		if (buf[idx] & 0x80) {
			idx += list_entry (fp, buf + idx, size - idx);
		}
		else {
			idx += 1;
		}

		idx = (idx + 1) & ~1UL;
	}

	free (buf);

	return (0);
}

static
int mfs_list (const char *fname, unsigned long offset)
{
	int           r;
	unsigned long blocks, mtime;
	FILE          *fp;
	mfs_vol_t     mfs;

	if (uf_stat (fname, &blocks, &mtime)) {
		fprintf (stderr, "%s: stat failed (%s)\n", arg0, fname);
		return (1);
	}

	if (par_blocks > 0) {
		blocks = par_blocks;
	}
	else {
		if (offset >= blocks) {
			fprintf (stderr, "%s: bad offset (%lu > %lu)\n",
				arg0, offset, blocks
			);
			return (1);
		}

		blocks = (blocks - offset) / 512;
	}

	fp = mac_open_inp (fname, "rb", 0);

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			arg0, fname
		);
		return (1);
	}

	if (mfs_open_auto (&mfs, fp, offset, blocks)) {
		fprintf (stderr, "%s: opening MFS volume failed (%s)\n",
			arg0, fname
		);
		mac_close (fp);
		return (1);
	}

	r = mfs_list_fp (&mfs, stdout);

	mfs_close (&mfs, 0);

	mac_close (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r, ret;
	char **optarg;

	arg0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'c':
			par_ctime = 1;
			break;

		case 'i':
			ret |= mfs_list (optarg[0], par_offset);
			break;

		case 'l':
			par_long = 1;
			break;

		case 'n':
			par_numeric = 1;
			break;

		case 'N':
			par_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'O':
			par_offset = strtoul (optarg[0], NULL, 0);
			break;

		case 'p':
			par_physical = 1;
			break;

		case 0:
			ret |= mfs_list (optarg[0], par_offset);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
