/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mfs/mkmfs.c                                              *
 * Created:     2012-04-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2012 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include "mfsinfo.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/mfs.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


#define TAG "%-24s "


static const char    *arg0 = NULL;

static unsigned long par_offset = 0;
static unsigned long par_blocks = 0;
static unsigned long par_ablk_size = 0;
static unsigned      par_dir_blocks = 12;
static unsigned long par_clump_size = 0;
static const char    *par_name = NULL;
static const char    *par_ctime = NULL;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 1, "allocation-block-size", "int", "Set the allocation block size [0]" },
	{ 'c', 1, "clump-size", "int", "Set the clump size [8192]" },
	{ 'C', 1, "creation-time", "int", "Set the volume creation time [now]" },
	{ 'd', 1, "directory-blocks", "int", "Set the number of directory blocks [12]" },
	{ 'n', 1, "blocks", "int", "Set the number of blocks in the volume [0]" },
	{ 'O', 1, "offset", "int", "Set the starting offset [0]" },
	{ 's', 1, "size", "int", "Set the volume size in KiB [0]" },
	{ 'v', 1, "volume-name", "string", "Set the volume name [Noname]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mkmfs: create MFS file systems",
		"usage: mkmfs [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mkmfs version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2012 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
int mkmfs_fp (FILE *fp, unsigned long blk, unsigned long ofs)
{
	unsigned      i;
	unsigned      blk_boot, blk_bmap, blk_dir, blk_data;
	unsigned      ablk_size, ablk_cnt;
	unsigned long clump_size;
	const char    *name;
	unsigned long ctime;
	time_t        now;
	unsigned char buf[512];

	if (blk < 16) {
		return (1);
	}

	name = (par_name != NULL) ? par_name : "Noname";

	if (par_ctime != NULL) {
		ctime = mac_time_from_string (par_ctime);
	}
	else {
		now = time (NULL);
		ctime = mac_time_from_unix (now);
	}

	blk_boot = 2;
	blk_bmap = 2;
	blk_dir = par_dir_blocks;

	if (blk <= (blk_boot + blk_bmap + blk_dir + 2)) {
		return (1);
	}

	blk_data = blk - blk_boot - blk_bmap - blk_dir - 2;

	if (par_ablk_size == 0) {
		ablk_size = (512UL * blk_data) / 640;
	}
	else {
		ablk_size = par_ablk_size;
	}

	ablk_size = (ablk_size + 511) & ~511UL;
	ablk_cnt = (512UL * blk_data) / ablk_size;

	blk_bmap = (64 + (3 * ablk_cnt + 1) / 2 + 511) / 512;

	if (blk_bmap > 2) {
		blk_data -= blk_bmap - 2;
		ablk_cnt = (512UL * blk_data) / ablk_size;
		blk_bmap = (64 + (3 * ablk_cnt + 1) / 2 + 511) / 512;
	}

	clump_size = (par_clump_size > 0) ? par_clump_size : 8192;

	fprintf (stderr, "volume name:           %s\n", name);
	fprintf (stderr, "blocks:                %lu\n", blk);
	fprintf (stderr, "boot blocks:           %u\n", blk_boot);
	fprintf (stderr, "block map blocks:      %u\n", blk_bmap);
	fprintf (stderr, "directory start:       %u\n", blk_boot + blk_bmap);
	fprintf (stderr, "directory blocks:      %u\n", blk_dir);
	fprintf (stderr, "data start:            %u\n", blk_boot + blk_bmap + blk_dir);
	fprintf (stderr, "data blocks:           %u\n", ablk_cnt * (ablk_size / 512));
	fprintf (stderr, "allocation block size: %u\n", ablk_size);
	fprintf (stderr, "allocation blocks:     %u\n", ablk_cnt);
	fprintf (stderr, "clump size:            %lu\n", clump_size);

	memset (buf, 0, 512);

	for (i = 0; i < (blk_boot + blk_bmap + blk_dir); i++) {
		if (mac_write (fp, buf, ofs + 512UL * i, 512)) {
			return (1);
		}
	}

	for (i = 0; i < 2; i++) {
		if (mac_write (fp, buf, ofs + 512UL * (blk - i - 1), 512)) {
			return (1);
		}
	}

	set_uint16_be (buf, 0, MFS_ID);
	set_uint32_be (buf, 2, ctime); /* ctime */
	set_uint32_be (buf, 6, ctime); /* btime */
	set_uint16_be (buf, 10, 0); /* attr */
	set_uint16_be (buf, 12, 0); /* number of files */
	set_uint16_be (buf, 14, blk_boot + blk_bmap); /* directory start */
	set_uint16_be (buf, 16, blk_dir); /* directory blocks */
	set_uint16_be (buf, 18, ablk_cnt); /* allocation blocks */
	set_uint32_be (buf, 20, ablk_size); /* allocation block size */
	set_uint32_be (buf, 24, clump_size); /* clump size */
	set_uint16_be (buf, 28, blk_boot + blk_bmap + blk_dir); /* first allocation block */
	set_uint32_be (buf, 30, 1); /* next file number */
	set_uint16_be (buf, 34, ablk_cnt); /* free allocation blocks */

	mac_pstring_encode_buf (buf + 36, 28, name);

	if (mac_write (fp, buf, ofs + 1024, 64)) {
		return (1);
	}

	if (mac_write (fp, buf, ofs + 512UL * blk - 1024, 64)) {
		return (1);
	}

	buf[0] = 0;

	if (mac_write (fp, buf, ofs + 512UL * blk - 1, 1)) {
		return (1);
	}

	return (0);
}

static
int mkmfs (const char *fname, unsigned long blk, unsigned long ofs)
{
	int  r;
	FILE *fp;

	fp = mac_open_out (fname, "wb", 0);

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			arg0, fname
		);
		return (1);
	}

	r = mkmfs_fp (fp, blk, ofs);

	mac_close (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r, ret;
	char **optarg;

	arg0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'a':
			par_ablk_size = strtoul (optarg[0], NULL, 0);
			break;

		case 'c':
			par_clump_size = strtoul (optarg[0], NULL, 0);
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'd':
			par_dir_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'n':
			par_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'O':
			par_offset = strtoul (optarg[0], NULL, 0);
			break;

		case 's':
			par_blocks = 2 * strtoul (optarg[0], NULL, 0);
			break;

		case 'v':
			par_name = optarg[0];
			break;

		case 0:
			ret |= mkmfs (optarg[0], par_blocks, par_offset);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
