/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/sit/sitenc.c                                             *
 * Created:     2007-12-01 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"

#include "sitenc.h"

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/dir.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/strings.h>


#define SIT_MAGIC1 0x53495421
#define SIT_MAGIC2 0x724c6175


typedef struct {
	FILE          *fp;

	unsigned long ofs;

	unsigned      entry_cnt;

	unsigned long data_cnt;
	unsigned long rsrc_cnt;
} sit_enc_t;


static int sit_encode_entry (void *ext, unsigned op, const mac_stat_t *ms);


static const char    *arg0 = NULL;

static int           par_verbose = 1;

static unsigned      par_ff_mask = 0xffff;

static unsigned long par_type = MAC_FOURCC_UNKN;
static unsigned long par_creator = MAC_FOURCC_UNKN;

static const char    *par_folder_file = ".folder";


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'f', 1, "folder-file", "string", "Set the folder file name [.folder]" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'o', 1, "output", "string", "Set the output file name" },
	{ 'q', 0, "quiet", NULL, "Disable printing of file names [no]" },
	{ 'r', 0, "reset", NULL, "Reset the finder flags [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"sitenc: encode uncompressed StuffIt 1.5 files",
		"usage: sitenc [options] output [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"sitenc version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2007-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int sit_copy (FILE *dst, unsigned long dofs, FILE *src, unsigned long sofs,
	unsigned long cnt, unsigned short *crc)
{
	unsigned long n;
	crc16_t       c;
	unsigned char buf[512];

	crc16_init (&c, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);

	while (cnt > 0) {
		n = (cnt < 512) ? cnt : 512;

		if (mac_read (src, buf, sofs, n)) {
			return (1);
		}

		if (mac_write (dst, buf, dofs, n)) {
			return (1);
		}

		crc16_calc (&c, buf, n);

		sofs += n;
		dofs += n;

		cnt -= n;
	}

	crc16_done (&c);

	*crc = crc16_get_crc (&c);

	return (0);
}

static
int sit_copy_mar (FILE *dst, unsigned long dofs, FILE *src, unsigned long sofs,
	unsigned long cnt, unsigned short *crc)
{
	unsigned long n;
	crc16_t       c;
	unsigned long mar_crc;
	unsigned char buf[512];

	if (cnt == 0) {
		*crc = 0;
		return (0);
	}

	crc16_init (&c, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);

	mar_crc = 0;

	while (cnt > 0) {
		n = (cnt < 512) ? cnt : 512;

		if (mac_read (src, buf, sofs, n)) {
			return (1);
		}

		if (mac_write (dst, buf, dofs, n)) {
			return (1);
		}

		mar_crc = mar_calc_crc (buf, n, mar_crc);

		crc16_calc (&c, buf, n);

		sofs += n;
		dofs += n;

		cnt -= n;
	}

	crc16_done (&c);

	*crc = crc16_get_crc (&c);

	if (mac_read (src, buf, sofs, 4)) {
		return (1);
	}

	if (get_uint32_be (buf, 0) != mar_crc) {
		return (1);
	}

	return (0);
}

static
void sit_init (sit_enc_t *sit)
{
	sit->fp = NULL;

	sit->ofs = 22;

	sit->entry_cnt = 0;

	sit->data_cnt = 0;
	sit->rsrc_cnt = 0;
}

static
void sit_free (sit_enc_t *sit)
{
	mac_close (sit->fp);
}

static
int sit_set_fname (sit_enc_t *sit, const char *fname)
{
	if (sit->fp != NULL) {
		return (1);
	}

	sit->fp = mac_open_out (fname, "wb", 0);

	if (sit->fp == NULL) {
		return (1);
	}

	return (0);
}

static
int sit_write_header (sit_enc_t *sit)
{
	unsigned char buf[22];

	set_uint32_be (buf, 0, SIT_MAGIC1);
	set_uint16_be (buf, 4, sit->entry_cnt);
	set_uint32_be (buf, 6, sit->ofs);
	set_uint32_be (buf, 10, SIT_MAGIC2);
	set_uint8_be (buf, 14, 1);
	set_uint8_be (buf, 15, 0);
	set_uint16_be (buf, 16, 0);
	set_uint32_be (buf, 18, 0);

	if (mac_write (sit->fp, buf, 0, 22)) {
		return (1);
	}

	return (0);
}

static
int sit_encode_file_mb (sit_enc_t *sit, const mac_stat_t *ms, FILE *fp, unsigned char *mb)
{
	unsigned       i, n;
	unsigned char  buf[112];
	unsigned long  ofs;
	unsigned long  rsrc, data;
	unsigned long  rsrc_ofs;
	unsigned short crc;

	memset (buf, 0, 112);

	set_uint8_be (buf, 0, 0);
	set_uint8_be (buf, 1, 0);

	n = mb[1];
	if ((n < 1) || (n > 63)) {
		return (1);
	}

	buf[2] = n;

	for (i = 0; i < n; i++) {
		buf[3 + i] = mb[2 + i];
	}

	set_uint32_be (buf, 66, mb_get_type (mb));
	set_uint32_be (buf, 70, mb_get_creator (mb));
	set_uint16_be (buf, 74, mb_get_finder_flags (mb) & par_ff_mask);
	set_uint32_be (buf, 76, mb_get_ctime (mb));
	set_uint32_be (buf, 80, mb_get_mtime (mb));

	rsrc = mb_get_rsrc_size (mb);
	data = mb_get_data_size (mb);

	sit->rsrc_cnt += rsrc;
	sit->data_cnt += data;

	rsrc_ofs = (data + 127) & ~0x7fUL;

	set_uint32_be (buf, 84, rsrc);
	set_uint32_be (buf, 88, data);
	set_uint32_be (buf, 92, rsrc);
	set_uint32_be (buf, 96, data);

	ofs = sit->ofs;

	if (sit_copy (sit->fp, ofs + 112, fp, 128 + rsrc_ofs, rsrc, &crc)) {
		return (1);
	}

	set_uint16_be (buf, 100, crc);

	if (sit_copy (sit->fp, ofs + 112 + rsrc, fp, 128, data, &crc)) {
		return (1);
	}

	set_uint16_be (buf, 102, crc);

	crc = crc16_crc (buf, 110, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);
	set_uint16_be (buf, 110, crc);

	if (mac_write (sit->fp, buf, ofs, 112)) {
		return (1);
	}

	sit->ofs += 112 + rsrc + data;
	sit->entry_cnt += 1;

	if (par_verbose) {
		fprintf (stdout, "M %8lu %8lu %s\n", data, rsrc, ms->name);
	}

	return (0);
}

static
int sit_encode_file_mar (sit_enc_t *sit, const mac_stat_t *ms, FILE *fp, unsigned char *mar)
{
	unsigned            i, n;
	unsigned char       buf[112];
	unsigned long       ofs;
	const unsigned char *name;
	unsigned long       rsrc, data;
	unsigned long       rsrc_ofs;
	unsigned short      crc;

	memset (buf, 0, 112);

	set_uint8_be (buf, 0, 0);
	set_uint8_be (buf, 1, 0);

	name = mar_get_name_pstr (mar);

	n = name[0];

	if ((n < 1) || (n > 63)) {
		return (1);
	}

	buf[2] = n;

	for (i = 0; i < n; i++) {
		buf[3 + i] = name[i + 1];
	}

	set_uint32_be (buf, 66, mar_get_type (mar));
	set_uint32_be (buf, 70, mar_get_creator (mar));
	set_uint16_be (buf, 74, mar_get_finder_flags (mar) & par_ff_mask);
	set_uint32_be (buf, 76, mar_get_ctime (mar));
	set_uint32_be (buf, 80, mar_get_mtime (mar));

	rsrc = mar_get_rsrc_size (mar);
	data = mar_get_data_size (mar);

	sit->rsrc_cnt += rsrc;
	sit->data_cnt += data;

	rsrc_ofs = (data == 0) ? 0 : ((data + 4 + 127) & ~127UL);

	set_uint32_be (buf, 84, rsrc);
	set_uint32_be (buf, 88, data);
	set_uint32_be (buf, 92, rsrc);
	set_uint32_be (buf, 96, data);

	ofs = sit->ofs;

	if (sit_copy_mar (sit->fp, ofs + 112, fp, 128 + rsrc_ofs, rsrc, &crc)) {
		return (1);
	}

	set_uint16_be (buf, 100, crc);

	if (sit_copy_mar (sit->fp, ofs + 112 + rsrc, fp, 128, data, &crc)) {
		return (1);
	}

	set_uint16_be (buf, 102, crc);

	crc = crc16_crc (buf, 110, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);
	set_uint16_be (buf, 110, crc);

	if (mac_write (sit->fp, buf, ofs, 112)) {
		return (1);
	}

	sit->ofs += 112 + rsrc + data;
	sit->entry_cnt += 1;

	if (par_verbose) {
		fprintf (stdout, "M %8lu %8lu %s\n", data, rsrc, ms->name);
	}

	return (0);
}

static
int sit_encode_file_data (sit_enc_t *sit, const mac_stat_t *ms, FILE *fp)
{
	unsigned char  buf[112];
	unsigned long  ofs;
	unsigned long  data;
	unsigned long  val;
	unsigned short crc;

	memset (buf, 0, 112);

	set_uint8_be (buf, 0, 0);
	set_uint8_be (buf, 1, 0);

	if (mac_pstring_encode_buf (buf + 2, 64, mac_get_last_component (ms->name, '/'))) {
		return (1);
	}

	set_uint32_be (buf, 66, par_type);
	set_uint32_be (buf, 70, par_creator);
	set_uint16_be (buf, 74, 0);

	val = mac_time_from_unix (ms->mtime);
	set_uint32_be (buf, 76, val);
	set_uint32_be (buf, 80, val);

	data = ms->size;

	sit->data_cnt += data;

	set_uint32_be (buf, 84, 0);
	set_uint32_be (buf, 88, data);
	set_uint32_be (buf, 92, 0);
	set_uint32_be (buf, 96, data);

	ofs = sit->ofs;

	set_uint16_be (buf, 100, 0);

	if (sit_copy (sit->fp, ofs + 112, fp, 0, data, &crc)) {
		return (1);
	}

	set_uint16_be (buf, 102, crc);

	crc = crc16_crc (buf, 110, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);
	set_uint16_be (buf, 110, crc);

	if (mac_write (sit->fp, buf, ofs, 112)) {
		return (1);
	}

	sit->ofs += 112 + data;
	sit->entry_cnt += 1;

	if (par_verbose) {
		fprintf (stdout, "D %8lu %8lu %s\n", data, 0UL, ms->name);
	}

	return (0);
}

static
int sit_encode_file (sit_enc_t *sit, const mac_stat_t *ms)
{
	int           r;
	FILE          *fp;
	unsigned char hdr[128];

	if (par_folder_file != NULL) {
		if (strcmp (mac_get_last_component (ms->name, '/'), par_folder_file) == 0) {
			return (0);
		}
	}

	fp = fopen (ms->name, "rb");

	if (fp == NULL) {
		return (1);
	}

	if (mac_read (fp, hdr, 0, 128)) {
		r = sit_encode_file_data (sit, ms, fp);
	}
	else {
		if (mar_is_mar (hdr) && (mar_get_id (hdr) == MAR_FILE)) {
			r = sit_encode_file_mar (sit, ms, fp, hdr);
		}
		else if (mb_get_version (hdr) == 2) {
			r = sit_encode_file_mb (sit, ms, fp, hdr);
		}
		else {
			r = sit_encode_file_data (sit, ms, fp);
		}
	}

	fclose (fp);

	if (r) {
		fprintf (stderr, "*** error (%s)\n", ms->name);
	}

	return (r);
}

static
int sit_encode_folder_done_inf (unsigned char *buf, const mac_stat_t *ms)
{
	unsigned            i, n;
	char                *path;
	FILE                *fp;
	macfile_t           mf;
	unsigned char       hdr[128];
	const unsigned char *name;

	if (par_folder_file == NULL) {
		return (1);
	}

	path = mac_combine_path (ms->name, par_folder_file);

	fp = fopen (path, "rb");

	if (fp == NULL) {
		free (path);
		return (1);
	}

	free (path);

	memset (hdr, 0, 128);

	mac_read (fp, hdr, 0, 128);

	mf_init (&mf);

	if (mar_is_mar (hdr)) {
		if (mar_get_id (hdr) != MAR_FOLDER_START) {
			fprintf (stderr,
				"%s: folder file error (mar id)\n",
				arg0
			);

			fclose (fp);

			return (1);
		}

		mar_get_macfile (hdr, &mf);
	}
	else if (mb_calc_crc (hdr) == mb_get_crc (hdr)) {
		if (mb_get_type (hdr) != MAC_FOURCC_FOLD) {
			fprintf (stderr,
				"%s: folder file error (mb type)\n",
				arg0
			);

			fclose (fp);

			return (1);
		}

		mb_get_macfile (hdr, &mf);
	}
	else {
		fseek (fp, 0, SEEK_SET);

		if (mf_load (&mf, fp)) {
			fprintf (stderr,
				"%s: folder file error (info)\n",
				arg0
			);

			fclose (fp);

			return (1);
		}
	}

	fclose (fp);

	name = mf_get_name_pstr (&mf);

	n = name[0];

	if ((n < 1) || (n > 63)) {
		return (1);
	}

	buf[2] = n;

	for (i = 0; i < n; i++) {
		buf[3 + i] = name[i + 1];
	}

	set_uint16_be (buf, 74, mf_get_finder_flags (&mf) & par_ff_mask);
	set_uint32_be (buf, 76, mf_get_ctime (&mf));
	set_uint32_be (buf, 80, mf_get_mtime (&mf));

	return (0);
}

static
int sit_encode_folder_done (sit_enc_t *sit, const mac_stat_t *ms,
	unsigned long ofs1, unsigned long ofs2,
	unsigned long rsrc, unsigned long data)
{
	unsigned char  buf[112];
	unsigned long  val;
	unsigned short crc;

	memset (buf, 0, 112);

	if (sit_encode_folder_done_inf (buf, ms)) {
		if (mac_pstring_encode_buf (buf + 2, 64, mac_get_last_component (ms->name, '/'))) {
			return (1);
		}

		set_uint16_be (buf, 74, 0);

		val = mac_time_from_unix (ms->mtime);
		set_uint32_be (buf, 76, val);
		set_uint32_be (buf, 80, val);
	}

	set_uint8_be (buf, 0, 0x20);
	set_uint8_be (buf, 1, 0x20);

	set_uint32_be (buf, 66, 0x00000000);
	set_uint32_be (buf, 70, 0x00000000);

	set_uint32_be (buf, 84, rsrc);
	set_uint32_be (buf, 88, data);
	set_uint32_be (buf, 92, 0);
	set_uint32_be (buf, 96, ofs2 - ofs1);

	set_uint16_be (buf, 100, 0);
	set_uint16_be (buf, 102, 0);

	crc = crc16_crc (buf, 110, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);
	set_uint16_be (buf, 110, crc);

	if (mac_write (sit->fp, buf, ofs1, 112)) {
		return (1);
	}

	set_uint8_be (buf, 0, 0x21);
	set_uint8_be (buf, 1, 0x21);

	set_uint32_be (buf, 84, 0);
	set_uint32_be (buf, 88, 0);
	set_uint32_be (buf, 92, 0);
	set_uint32_be (buf, 96, 0);

	crc = crc16_crc (buf, 110, 0x8005, CRC_REVERSE_INP | CRC_REVERSE_OUT);
	set_uint16_be (buf, 110, crc);

	if (mac_write (sit->fp, buf, ofs2, 112)) {
		return (1);
	}

	if (par_verbose) {
		fprintf (stdout, "F %8lu %8lu %s\n", data, rsrc, ms->name);
	}

	return (0);
}

static
int sit_encode_folder (sit_enc_t *sit, const mac_stat_t *ms)
{
	unsigned      entry;
	unsigned long ofs1, ofs2;
	unsigned long data, rsrc;

	ofs1 = sit->ofs;
	rsrc = sit->rsrc_cnt;
	data = sit->data_cnt;
	entry = sit->entry_cnt;

	sit->ofs += 112;

	if (mac_dir_list_dir (ms->name, 0, sit, sit_encode_entry)) {
		return (1);
	}

	ofs2 = sit->ofs;
	rsrc = sit->rsrc_cnt - rsrc;
	data = sit->data_cnt - data;

	if (sit_encode_folder_done (sit, ms, ofs1, ofs2, rsrc, data)) {
		fprintf (stderr, "*** error: folder (%s)\n", ms->name);
		return (1);
	}

	sit->entry_cnt = entry + 1;

	sit->ofs += 112;

	return (0);
}

static
int sit_encode_entry (void *ext, unsigned op, const mac_stat_t *ms)
{
	sit_enc_t *sit = ext;

	if (op == MAC_LIST_ENTER) {
		if (sit_encode_folder (sit, ms)) {
			return (1);
		}

		return (0);
	}

	if (op == MAC_LIST_LEAVE) {
		return (0);
	}

	if (op == MAC_LIST_LIST) {
		if (ms->is_file) {
			if (sit_encode_file (sit, ms)) {
				return (1);
			}
		}
		else if (ms->is_dir) {
			return (0);
		}
		else {
			fprintf (stderr, "*** error: bad file type (%s)\n",
				ms->name
			);
			return (1);
		}
	}

	return (0);
}

static
int sit_encode (sit_enc_t *sit, const char *src)
{
	if (sit->fp == NULL) {
		fprintf (stderr, "%s: need an output file\n", arg0);
		return (1);
	}

	while ((src[0] == '.') && (src[1] == '/')) {
		src += 2;
	}

	if (mac_dir_list (src, 0, sit, sit_encode_entry)) {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int       r;
	char      **optarg;
	sit_enc_t sit;

	arg0 = argv[0];

	sit_init (&sit);

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'f':
			par_folder_file = optarg[0];

			if (par_folder_file[0] == 0) {
				par_folder_file = NULL;
			}
			break;

		case 'i':
			if (sit_encode (&sit, optarg[0])) {
				return (1);
			}
			break;

		case 'o':
			if (sit_set_fname (&sit, optarg[0])) {
				fprintf (stderr,
					"%s: can't open output file\n",
					arg0
				);
				return (1);
			}
			break;

		case 'q':
			par_verbose = 0;
			break;

		case 'r':
			par_ff_mask = ~MF_FINDER_INITED;
			break;

		case 0:
			if (sit.fp == NULL) {
				if (sit_set_fname (&sit, optarg[0])) {
					fprintf (stderr,
						"%s: can't open output file (%s)\n",
						arg0, optarg[0]
					);
					return (1);
				}
			}
			else {
				if (sit_encode (&sit, optarg[0])) {
					return (1);
				}
			}
			break;

		default:
			return (1);
		}
	}

	if (sit.fp != NULL) {
		if (sit_write_header (&sit)) {
			return (1);
		}
	}

	sit_free (&sit);

	return (0);
}
